// ---------------------------------------------------------------------------------------------------------------------
//	PhotoChrome v6.xx - Atari ST interlaced image reduction framework
//	1991-2016 Douglas Little (d.m.l)
// ---------------------------------------------------------------------------------------------------------------------
//	special thanks to:
//		evil@DHS, RA/PDX, Troed/SYNC for help testing and input on improvements
// ---------------------------------------------------------------------------------------------------------------------

// ====================================================================
//	reusable image processing stuff
// ====================================================================

#include "common.h"

// --------------------------------------------------------------------
// colour space conversion
// --------------------------------------------------------------------

namespace cie
{
	
// --------------------------------------------------------------------
// CIE.XYZ <-> RGB
//
// The range for valid R, G, B values is [0,1]. Note, this matrix has negative 
// coefficients. Some XYZ color may be transformed to RGB values that are 
// negative or greater than one. This means that not all visible colors can 
// be produced using the RGB system.

// --------------------------------------------------------------------
// CIE.XYZ -> RGB / Observer. = 2, Illuminant = D65
//
//  [ R ]   [  3.240479 -1.537150 -0.498535 ]   [ X ]
//  [ G ] = [ -0.969256  1.875992  0.041556 ] * [ Y ]
//  [ B ]   [  0.055648 -0.204043  1.057311 ]   [ Z ].
//
// --------------------------------------------------------------------
// RGB -> CIE.XYZ / Observer. = 2, Illuminant = D65
//
//  [ X ]   [  0.412453  0.357580  0.180423 ]   [ R ]
//  [ Y ] = [  0.212671  0.715160  0.072169 ] * [ G ]
//  [ Z ]   [  0.019334  0.119193  0.950227 ]   [ B ].

// fixed point version
/*
void m_rgb2xyz_d65
(
	signed short r, signed short g, signed short b,
	signed short& x, signed short& y, signed short& z
)
{
	signed long lx = (((signed long)( 0.412453 * 65536.0) * r) + ((signed long)( 0.357580 * 65536.0) * g) + ((signed long)( 0.180423 * 65536.0) * b)) >> 16;
	signed long ly = (((signed long)( 0.212671 * 65536.0) * r) + ((signed long)( 0.715160 * 65536.0) * g) + ((signed long)( 0.072169 * 65536.0) * b)) >> 16;
	signed long lz = (((signed long)( 0.019334 * 65536.0) * r) + ((signed long)( 0.119193 * 65536.0) * g) + ((signed long)( 0.950227 * 65536.0) * b)) >> 16;
	x = lx;
	y = ly;
	z = lz;
}

void m_xyz2rgb_d65
(
	signed short x, signed short y, signed short z,
	signed short& r, signed short& g, signed short& b
)
{
	signed long lr = (((signed long)( 3.240479 * 65536.0) * x) + ((signed long)(-1.537150 * 65536.0) * y) + ((signed long)(-0.498535 * 65536.0) * z)) >> 16;
	signed long lg = (((signed long)(-0.969256 * 65536.0) * x) + ((signed long)( 1.875992 * 65536.0) * y) + ((signed long)( 0.041556 * 65536.0) * z)) >> 16;
	signed long lb = (((signed long)( 0.055648 * 65536.0) * x) + ((signed long)(-0.204043 * 65536.0) * y) + ((signed long)( 1.057311 * 65536.0) * z)) >> 16;
	r = lr;
	g = lg;
	b = lb;
}
*/
// --------------------------------------------------------------------

 // float (double) version. if we're going to use floats for precision purposes, might as well use doubles
void rgb2xyz
(
	double r, double g, double b,
	double& x, double& y, double& z
)
{
	static const double r255 = 1.0 / 255.0;
	static const double r1055 = 1.0 / 1.055;
	static const double r1292 = 1.0 / 12.92;
	
	double vr = r * r255;
	double vg = g * r255;
	double vb = b * r255;
	
	if (vr > 0.04045f) 
		vr = pow((vr + 0.055f) * r1055, 2.4f);
	else
		vr *= r1292;

	if (vg > 0.04045f)
		vg = pow((vg + 0.055f) * r1055, 2.4f);
	else          
		vg *= r1292;

	if (vb > 0.04045f) 
		vb = pow((vb + 0.055f) * r1055, 2.4f);
	else          
		vb *= r1292;

	// retain space 0-255 for orthogonal colour processing
	vr *= 100.0f;
	vg *= 100.0f;
	vb *= 100.0f;
	
	//Observer. = 2, Illuminant = D65
	//x = ( 0.412453 * vr) + ( 0.357580 * vg) + ( 0.180423 * vb);
	//y = ( 0.212671 * vr) + ( 0.715160 * vg) + ( 0.072169 * vb);
	//z = ( 0.019334 * vr) + ( 0.119193 * vg) + ( 0.950227 * vb);

	double vx,vy,vz;
	m_rgb2xyz_d65(vr,vg,vb,vx,vy,vz);
	//x = xclamp(vx,0.0,255.0);
	//y = xclamp(vy,0.0,255.0);
	//z = xclamp(vz,0.0,255.0);
	x = vx;
	y = vy;
	z = vz;
}

// float (double) version. if we're going to use floats for precision purposes, might as well use doubles
void xyz2rgb
(
	double x, double y, double z,
	double& r, double& g, double& b
)
{
	static const double r100 = 1.0 / 100.0;
	static const double r24 = 1.0 / 2.4;
	
	// from space 0-255
	double vx = x * r100;
	double vy = y * r100;
	double vz = z * r100;

	//Observer. = 2, Illuminant = D65
	//double vr = ( 3.240479 * vx) + (-1.537150 * vy) + (-0.498535 * vz);
	//double vg = (-0.969256 * vx) + ( 1.875992 * vy) + ( 0.041556 * vz);
	//double vb = ( 0.055648 * vx) + (-0.204043 * vy) + ( 1.057311 * vz);

	double vr,vg,vb;
	m_xyz2rgb_d65(vx,vy,vz,vr,vg,vb);
	
	if (vr > 0.0031308)
		vr = (1.055f * pow(vr, r24)) - 0.055f;
	else                     
		vr *= 12.92f;

	if (vg > 0.0031308)
		vg = (1.055f * pow(vg, r24)) - 0.055f;
	else                     
		vg *= 12.92f;
	
	if (vb > 0.0031308)
		vb = (1.055f * pow(vb, r24)) - 0.055f;
	else                     
		vb *= 12.92f;
	
	// retain space 0-255 for orthogonal colour processing
	r = xclamp(vr * 255.0f,0.0,255.0);
	g = xclamp(vg * 255.0f,0.0,255.0);
	b = xclamp(vb * 255.0f,0.0,255.0);
}
	
// --------------------------------------------------------------------

void xyz2cielab
(
	double x, double y, double z,
	double& l, double& a, double& b
) 
{	
	static const double r3 = 1.0 / 3.0;
	static const double c16d116 = 16.0 / 116.0;
	
	// Observer= 2, Illuminant= D65
	static const double r_refwhite_d65_x = 1.0 /  95.047;
	static const double r_refwhite_d65_y = 1.0 / 100.000;
	static const double r_refwhite_d65_z = 1.0 / 108.883;
	//static const double r8_refwhite_d65_x = 1.0 / (( 95.047 * 255.0) /  95.047);
	//static const double r8_refwhite_d65_y = 1.0 / ((100.000 * 255.0) / 100.000);
	//static const double r8_refwhite_d65_z = 1.0 / ((108.883 * 255.0) / 108.883);
	
	double vx = x * r_refwhite_d65_x;
	double vy = y * r_refwhite_d65_y;
	double vz = z * r_refwhite_d65_z;

	if (vx > 0.008856)
		vx = pow(vx, r3);
	else                    
		vx = (7.787 * vx) + c16d116;

	if (vy > 0.008856)
		vy = pow(vy, r3);
	else
		vy = (7.787 * vy) + c16d116;

	if (vz > 0.008856) 
		vz = pow(vz, r3);
	else                  
		vz = (7.787 * vz) + c16d116;

	l = (116.0f * vy) - 16.0f;
	a = 500.0f * (vx - vy);
	b = 200.0f * (vy - vz);
}

void cielab2xyz
(
	double l, double a, double b,
	double& x, double& y, double& z
)
{
	static const double c16d116 = 16.0 / 116.0;
	static const double r7787 = 1.0 / 7.787;
	
	double vy = (l + 16.0f) / 116.0f;
	double vx = (a / 500.0f) + vy;
	double vz = vy - (b / 200.0f);

	double vx3 = vx * vx * vx;
	double vy3 = vy * vy * vy;
	double vz3 = vz * vz * vz;
	
	if (vy3 > 0.008856f) 
		vy = vy3;
	else                      
		vy = (vy - c16d116) * r7787;

	if (vx3 > 0.008856f) 
		vx = vx3;
	else                      
		vx = (vx - c16d116) * r7787;

	if (vz3 > 0.008856f) 
		vz = vz3;
	else                      
		vz = (vz - c16d116) * r7787;

	// Observer= 2, Illuminant= D65
	static const double c_refwhite_d65_x =  95.047;
	static const double c_refwhite_d65_y = 100.000;
	static const double c_refwhite_d65_z = 108.883;
	
	x = vx * c_refwhite_d65_x;
	y = vy * c_refwhite_d65_y;
	z = vz * c_refwhite_d65_z;
	
	//static const double c8_refwhite_d65_x = ( 95.047 * 255.0) /  95.047;
	//static const double c8_refwhite_d65_y = (100.000 * 255.0) / 100.000;
	//static const double c8_refwhite_d65_z = (108.883 * 255.0) / 108.883;
	//x = xclamp(vx * c8_refwhite_d65_x,0.0,255.0);
	//y = xclamp(vy * c8_refwhite_d65_y,0.0,255.0);
	//z = xclamp(vz * c8_refwhite_d65_z,0.0,255.0);
}
 
// --------------------------------------------------------------------
 
void rgb2cielab
(
	double r, double g, double b,
	double& ciel, double& ciea, double& cieb
)
{
	double x,y,z;
	rgb2xyz(r,g,b,x,y,z);
	xyz2cielab(x,y,z,ciel,ciea,cieb);
}

void cielab2rgb
(
	double ciel, double ciea, double cieb,
	double& r, double& g, double& b
)
{
	double x,y,z;
	cielab2xyz(ciel,ciea,cieb,x,y,z);
	xyz2rgb(x,y,z,r,g,b);
}

// --------------------------------------------------------------------

extern "C" 
{
#if defined(__ATARI__)

word_t htons(word_t hostshort)
{ 
	return hostshort;
}
long_t htonl(long_t hostlong)
{
	return hostlong;
}
word_t ntohs(word_t netshort)
{
	return netshort;
}
long_t ntohl(long_t netlong)
{
	return netlong;
}

#else

__stdcall word_t htons(word_t hostshort)
{ 
	word_t v = hostshort;

	word_t vs = 
		((v & 0x00FF) << 8) | 
		((v & 0xFF00) >> 8) 
	;

	return v;
}
__stdcall long_t htonl(long_t hostlong)
{ 
	long_t v = hostlong;

	long_t vs = 
		((v & 0x000000FF) << 24) | 
		((v & 0x0000FF00) << 8) | 
		((v & 0x00FF0000) >> 8) | 
		((v & 0xFF000000) >> 24)
	;

	return v;
}
__stdcall word_t ntohs(word_t netshort)
{ 
	word_t v = netshort;

	word_t vs = 
		((v & 0x00FF) << 8) | 
		((v & 0xFF00) >> 8) 
	;

	return v;
}
__stdcall long_t ntohl(long_t netlong)
{ 
	long_t v = netlong;

	long_t vs = 
		((v & 0x000000FF) << 24) | 
		((v & 0x0000FF00) << 8) | 
		((v & 0x00FF0000) >> 8) | 
		((v & 0xFF000000) >> 24)
	;

	return v;
}

#endif
};

// --------------------------------------------------------------------

};

// --------------------------------------------------------------------
