// ---------------------------------------------------------------------------------------------------------------------
//	PhotoChrome v6.xx - Atari ST interlaced image reduction framework
//	1991-2016 Douglas Little (d.m.l)
// ---------------------------------------------------------------------------------------------------------------------
//	special thanks to:
//		evil@DHS, RA/PDX, Troed/SYNC for help testing and input on improvements
// ---------------------------------------------------------------------------------------------------------------------

#ifndef common_h_
#define common_h_

// ====================================================================
// 	common includes, types, helper functions and other misc. stuff
// ====================================================================

#include <string>
#include <vector>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <memory.h>
#include <math.h>

// --------------------------------------------------------------------
//	binary literal template/macro
// --------------------------------------------------------------------

// dml: bit-pattern input argument for binary(n) is OCTAL format, and that means fewer bits can be passed than you might
// expect before the type range is exceeded and bits get clipped. longest_t was configured to 32bits, which imposes a 
// limit of 11 octal 'bits'. longest_t has been changed to 64bit which allows 22 octal 'bits' to be passed without 
// clipping - enough to handle 16bit words, which is helpful in PCS for palette swizzles etc. It can't be used for
// 24 or 32bit patterns unless a 128bit type is used - and I don't trust GCC that far :)
typedef unsigned long long longest_t;

#include "binarylit.h"

// --------------------------------------------------------------------

#include "rnd.h"

// --------------------------------------------------------------------
//	postability stuff

#if defined(__ATARI__) || defined(__MINT__)

// this should probably be fixed by editing c++config.h in the GCC headers to define '_GLIBCXX_HAVE__XXXF' etc. since they are present
// in the GCC 4.6.x libstdc++ but this will suffice for now!
#if !defined(_GLIBCXX_HAVE__SINF)
extern float cosf(float a);
extern float sinf(float a);
extern float tanf(float a);
extern float acosf(float a);
extern float asinf(float a);
extern float atanf(float a);
extern float atan2f(float a, float b);
extern float coshf(float a);
extern float sinhf(float a);
extern float tanhf(float a);
extern float expf(float a);
extern float logf(float a);
extern float log10f(float a);
extern float modff(float a, float *b);
extern float powf(float a, float b);
extern float sqrtf(float a);
extern float ceilf(float a);
extern float floorf(float a);
extern float fabsf(float a);
extern float fmodf(float a, float b);
extern float hypotf(float a, float b);
#endif

// looks like a similar deal - missing from m68k gcc headers, yet are exported from the default libraries...
extern wchar_t *wcschr (__const wchar_t *__wcs, wchar_t __wc);
extern wchar_t *wcschr (wchar_t *__wcs, wchar_t __wc);
extern wchar_t *wcsrchr (__const wchar_t *__wcs, wchar_t __wc);
extern wchar_t *wcsrchr (wchar_t *__wcs, wchar_t __wc);

#endif

// --------------------------------------------------------------------

typedef unsigned char byte_t;
typedef unsigned short word_t;
typedef unsigned long long_t;
typedef signed long datasize_t;

typedef float s_real;
#ifdef __ATARI__
typedef float d_real;	// save memory on atari systems
#else
typedef double d_real;
#endif

#ifdef USE_HIGH_PRECISION
typedef d_real calc_real;
typedef d_real arg_real;
typedef d_real storage_real;
#else
typedef d_real calc_real;
typedef s_real arg_real;
typedef s_real storage_real;
#endif

// --------------------------------------------------------------------

// just in case tolower is not in the host env headers...
inline char mytolower(char in)
{
  if(in<='Z' && in>='A')
    return in-('Z'-'z');
  return in;
} 

inline char sanitise_csv_arg(char ch)
{
	unsigned char uc = (unsigned char)ch;
	if ((uc >= '0' && uc <= '9') ||
		(uc == '-') ||
		(uc == 'x'))
		return ch;

	return ' ';
}

// --------------------------------------------------------------------

template<class T>
inline T xlerp(T t, T a, T b) { return a + (t * (b - a)); }

template<class T>
inline T xmax(T a, T b) { return a > b ? a : b; }
template<class T>
inline T xmin(T a, T b) { return a < b ? a : b; }

template<class T>
inline T xclamp(T v, T a, T b) { return xmin(xmax(v, a), b); }

template<class T>
inline T xabs(T x) { return x > T(0) ? x : -x; }

template <class T>
inline T xsign(const T& _v) { return (_v > T(0)) ? T(1) : ((_v < T(0)) ? T(-1) : T(0)); }

template <class T>
inline T xsqr(const T& _v) { return _v * _v; }

inline int xsmod(const int _a, const int _b)
{ 
	if (_a >= 0) 
		return (_a % _b); 
	return (_b - ((_b - _a) % _b));
}

//------------------------------------------------------------------------*
// float to int with explicit rounding schemes

inline float xftoi_floor(float f) { return floor(f); }
inline float xftoi_near(float f) { return floor(f + 0.5f); }
inline float xftoi_ceil(float f) { return ceil(f); }

inline double xftoi_floor(double f) { return floor(f); }
inline double xftoi_near(double f) { return floor(f + 0.5f); }
inline double xftoi_ceil(double f) { return ceil(f); }

#if !(COLOUR_ROUNDING_SCHEME-0)
inline float xftoi(float f) { return xftoi_floor(f); }
inline double xftoi(double f) { return xftoi_floor(f); }
#elif !(COLOUR_ROUNDING_SCHEME-1)
inline float xftoi(float f) { return xftoi_near(f); }
inline double xftoi(double f) { return xftoi_near(f); }
#elif !(COLOUR_ROUNDING_SCHEME-2)
inline float xftoi(float f) { return xftoi_ceil(f); }
inline double xftoi(double f) { return xftoi_ceil(f); }
#endif

//------------------------------------------------------------------------*
// convert between float <-> integer colour ranges

#define MAX_PALETTE_SHIFT 4
#define MAX_PALETTE_SIZE (1<<MAX_PALETTE_SHIFT)

inline unsigned short f2col(float f) { return (unsigned short)(int)xftoi(xclamp<float>(f * 255.0f, 0.0f, 255.0f)); }
inline unsigned short f2col(double f) { return (unsigned short)(int)xftoi(xclamp<float>(f * 255.0f, 0.0f, 255.0f)); }
inline float col2f(signed short c) { return xclamp<float>((float)c / 255.0f, 0.0f, 1.0f); }

inline unsigned short f2range(float f, float r) { return (unsigned short)(int)xftoi(xclamp<float>(f * r, 0.0f, r)); }
inline unsigned short f2range(double f, double r) { return (unsigned short)(int)xftoi(xclamp<float>(f * r, 0.0f, r)); }
inline float range2f(signed short c, float r) { return xclamp<float>((float)c / r, 0.0f, 1.0f); }

inline int pal_crange(signed short _colourdepth) { return ((1 << _colourdepth) - 1); }

//3 = n(0-7) = v(0-14)
//4 = n(0-15) = v(0-30)
//inline int nvi_crange(signed short _colourdepth) { return ((1 << (_colourdepth + 1)) - 2); }
//inline int nvi_crange(signed short _colourdepth) { return (((1 << _colourdepth) - 1) << 1); }
inline int nvi_crange(signed short _colourdepth, signed short _fields) { return (((1 << _colourdepth) - 1) * _fields); }

inline int src_crange() { return (256 - 1); }

static const int max_fields = 4;

inline float nvi_quantize(float u, signed short _colourdepth, signed short _fields)
{
	int p = f2range(u, (float)nvi_crange(_colourdepth, _fields));
	return range2f(p, (float)nvi_crange(_colourdepth, _fields));
}

inline float pal_quantize(float u, signed short _colourdepth)
{
	int p = f2range(u, (float)pal_crange(_colourdepth));
	return range2f(p, (float)pal_crange(_colourdepth));
}

//------------------------------------------------------------------------*

template<class T>
inline T inverse_product(T _a, T _b)
{
	return T(1.0) - ((T(1.0) - _a) * (T(1.0) - _b));
}

template<class T>
inline T inverse_attenuation(T _a, T _i)
{
	return T(1.0) - ((T(1.0) - _a) * _i);
}

// --------------------------------------------------------------------

inline int msb(unsigned long v) 
{
	static const int pos[32] = 
	{
		0,  1,  28, 2,  29, 14, 24, 3,
		30, 22, 20, 15, 25, 17, 4,  8, 
		31, 27, 13, 23, 21, 19, 16, 7, 
		26, 12, 18, 6,  11, 5,  10, 9
	};
	
	v |= v >> 1;
	v |= v >> 2;
	v |= v >> 4;
	v |= v >> 8;
	v |= v >> 16;
	v = (v >> 1) + 1;
	
	return pos[(v * 0x077CB531UL) >> 27];
}

inline unsigned long nlpo2(unsigned long x)
{
        x |= (x >> 1);
        x |= (x >> 2);
        x |= (x >> 4);
        x |= (x >> 8);
        x |= (x >> 16);
        return (x+1);
}

inline unsigned long reverse(unsigned long x)
{
    x = ((x >> 1)  & 0x55555555u) | ((x & 0x55555555u) << 1);
    x = ((x >> 2)  & 0x33333333u) | ((x & 0x33333333u) << 2);
    x = ((x >> 4)  & 0x0f0f0f0fu) | ((x & 0x0f0f0f0fu) << 4);
    x = ((x >> 8)  & 0x00ff00ffu) | ((x & 0x00ff00ffu) << 8);
    x = ((x >> 16) & 0x0000ffffu) | ((x & 0x0000ffffu) << 16);
    return x;
}

//------------------------------------------------------------------------*
// read or write a bigendian word in a host- and alignment-insensitive way
// meant to be safe and portable, not necessarily quick.

inline word_t read_be_word(void* _addr)
{
	return ((word_t)(((byte_t*)_addr)[0]) << 8) | 
		   ((word_t)(((byte_t*)_addr)[1]));
}	

inline void write_be_word(void* _addr, word_t _val)
{
	((byte_t*)_addr)[0] = (_val >> 8);
	((byte_t*)_addr)[1] = (_val & 0xFF);
}	

// byte & word versions redundant, but make for less confusing source code

inline byte_t read_byte(void* _addr)
{
	return *((byte_t*)_addr);
}	

inline void write_byte(void* _addr, byte_t _val)
{
	*((byte_t*)_addr) = _val;
}

inline word_t read_word(void* _addr)
{
	return *((word_t*)_addr);
}	

inline void write_word(void* _addr, word_t _val)
{
	*((word_t*)_addr) = _val;
}	

// copies native words in source buffer to BE words in destination buffer
// note: inefficient, but it's only used for emitting raw image data to file
inline void copy_be_words(word_t* _dst, word_t* _src, datasize_t _size)
{
	for (int i = 0; i < _size; ++i)
		write_be_word(_dst++, read_word(_src++));
}	

inline void endianswap32(long_t* _data, datasize_t _num)
{
	for (datasize_t i = 0; i < _num; ++i)
	{
		long_t v = _data[i];

		long_t vs = 
			((v & 0x000000FF) << 24) | 
			((v & 0x0000FF00) << 8) | 
			((v & 0x00FF0000) >> 8) | 
			((v & 0xFF000000) >> 24)
		;
		
		_data[i] = vs;
	}
}

inline void endianswap16(word_t* _data, datasize_t _num)
{
	for (datasize_t i = 0; i < _num; ++i)
	{
		word_t v = _data[i];

		word_t vs = 
			((v & 0x00FF) << 8) | 
			((v & 0xFF00) >> 8)
		;
		
		_data[i] = vs;
	}
}

// --------------------------------------------------------------------

namespace cie
{
	// --------------------------------------------------------------------
	// RGB <-> CIEXYZ conversion matrices - fixed point version
/*	
	void m_rgb2xyz_d65
	(
		signed short r, signed short g, signed short b,
		signed short& x, signed short& y, signed short& z
	);

	void m_xyz2rgb_d65
	(
		signed short x, signed short y, signed short z,
		signed short& r, signed short& g, signed short& b
	);
*/	
	// --------------------------------------------------------------------
	// RGB <-> CIEXYZ conversion matrices - float version
	
	inline void m_rgb2xyz_d65
	(
		double r, double g, double b,
		double& x, double& y, double& z
	)
	{
		x = ( 0.412453 * r) + ( 0.357580 * g) + ( 0.180423 * b);
		y = ( 0.212671 * r) + ( 0.715160 * g) + ( 0.072169 * b);
		z = ( 0.019334 * r) + ( 0.119193 * g) + ( 0.950227 * b);
	} 

	inline void m_xyz2rgb_d65
	(
		double x, double y, double z,
		double& r, double& g, double& b
	)
	{	
		r = ( 3.240479 * x) + (-1.537150 * y) + (-0.498535 * z);
		g = (-0.969256 * x) + ( 1.875992 * y) + ( 0.041556 * z);
		b = ( 0.055648 * x) + (-0.204043 * y) + ( 1.057311 * z);
	}	
	
	// --------------------------------------------------------------------
	// RGB <-> CIEXYZ conversion functions - float version
	
	void rgb2xyz
	(
		double r, double g, double b,
		double& x, double& y, double& z
	);

	void xyz2rgb
	(
		double x, double y, double z,
		double& r, double& g, double& b
	);	

	void xyz2cielab
	(
		double x, double y, double z,
		double& l, double& a, double& b
	);
	
	void cielab2xyz
	(
		double l, double a, double b,
		double& x, double& y, double& z
	);
	
	void rgb2cielab
	(
		double r, double g, double b,
		double& ciel, double& ciea, double& cieb
	);

	void cielab2rgb
	(
		double ciel, double ciea, double cieb,
		double& r, double& g, double& b
	);

};			

// --------------------------------------------------------------------

template <class T>
inline void saturationchange(T &R, T &G, T &B, T _delta)
{
	
	static const T Pr = 0.299;
	static const T Pg = 0.587;
	static const T Pb = 0.114;

	T P = sqrt((xsqr(R) * Pr) + (xsqr(G) * Pg) + (xsqr(B) * Pb));
	
	R = P + (R - P) * _delta;
	G = P + (G - P) * _delta;
	B = P + (B - P) * _delta;
}

// --------------------------------------------------------------------

extern "C"
{
word_t htons(word_t hostshort);
long_t htonl(long_t hostlong);
word_t ntohs(word_t netshort);
long_t ntohl(long_t netlong);
};

// --------------------------------------------------------------------

#endif // common_h_
