// ---------------------------------------------------------------------------------------------------------------------
//	PhotoChrome v6.xx - Atari ST interlaced image reduction framework
//	1991-2016 Douglas Little (d.m.l)
// ---------------------------------------------------------------------------------------------------------------------
//	special thanks to:
//		evil@DHS, RA/PDX, Troed/SYNC for help testing and input on improvements
// ---------------------------------------------------------------------------------------------------------------------

// ====================================================================
// 	harness to process/convert a single file
// ====================================================================
//	note: this bit is volatile - likely to be refactored
// ====================================================================

#include <iostream>
#include <algorithm>
#include <string>

#include "common.h"
#include "config.h"

#if defined(__ATARI__) || defined(__MINGW32__)
#include "FreeImage.3.15.3.h"
#else
#include "FreeImage.3.15.4.h"
#endif

#include "imagepkg.h"
#include "pcsimage_std_sgl.h"
#include "pcsimage_std_dual.h"
#include "pcsimage_std_tri.h"
#include "pcsimage_std_quad.h"

extern Arguments g_allargs;

// --------------------------------------------------------------------

bool diagnostics_ = false;

bool process_file(const char* const _name, const char* const _outname)
{
	// nudge image package into emitting version info before any other messages
	ImagePkg::instance();
	
	// create source image, and perform any preparation processing (i.e. bitdepth, size)
	
	bool success = true;


	std::string sname(_name);
	std::transform(sname.begin(), sname.end(), sname.begin(), mytolower);
	
	int type = -1;
	if      (sname.find(".jpg") != std::string::npos)
		type = 0;
	else if (sname.find(".tga") != std::string::npos)
		type = 1;
	else if (sname.find(".png") != std::string::npos)
		type = 2;
	else if (sname.find(".bmp") != std::string::npos)
		type = 3;
	else
	{
	    if (sname.find("_test") == 0 && 
	    	sname.find(".") == std::string::npos)
		{
			std::string num = sname.substr(5, std::string::npos);
			type = atoi(num.c_str());
			printf("selected built-in test source (%d)...\n", type);
			type = -(type+1);
		}
		else
		{			
			success = false;
			printf("error - unknown source image file extension!\n");		
			//printf("press any key...\n"); getchar();
		}
	}

	FIBITMAP *fib = 0;

	// read normal (non-testcard) image sources before creating image class so we can auto-size the convertor where relevant
	if (success && (type >= 0))
	{	
		if (!g_allargs.quiet)
			printf("> reading source image [%s]...\n", _name);		

		fib = PCSImage::read_image(_name, type);
		if (fib)
		{
			if (g_allargs.verbose)
				printf("> source image read ok.\n");
		}
		else
		{
			printf("error reading source image!\n");	
			success = false;
		}
	}

	PCSImage* img = 0;

	if (success)
	{
		switch (g_allargs.ccmode)
		{
			case CCMode_NONE:
			case CCMode_DEGAS:
				// 'image' mode is really 320x200 PI1s so these will get resized to default
				break;
			case CCMode_PLAYFIELD:
			case CCMode_SUPERPALETTE:
			case CCMode_PALETTE:
			case CCMode_RAWIMAGE:
				{
					// auto-size the source image, to bypass the need to specify output dimensions on commandline
					int w = 0;
					int h = 0;
					ImagePkg::instance().get_size(fib, w, h);	
					g_allargs.xsize = w;
					g_allargs.ysize = h;
				}
				break;
		};


		switch (g_allargs.ccmode)
		{
			case CCMode_NONE:
				printf("PhotoChrome mode\n");
				break;
			case CCMode_DEGAS:
				printf("CryptoChrome mode: 320x200 Degas PI1\n");
				break;
			case CCMode_PLAYFIELD:
				printf("CryptoChrome mode: game playfield\n");
				break;
			case CCMode_SUPERPALETTE:
				printf("CryptoChrome mode: game superpalette (multiple images)\n");
				break;
			case CCMode_PALETTE:
				printf("CryptoChrome mode: palette\n");
				break;
			case CCMode_RAWIMAGE:
				printf("CryptoChrome mode: raw planar image\n");
				break;
		};

		switch (g_allargs.fields)
		{
		case 1:
			img = new PCSImage_std_sgl(_name, g_allargs.xsize, g_allargs.ysize, g_allargs.colourdepth, g_allargs.colours, g_allargs.planes, g_allargs.palettes, g_allargs.palette_skip);
			break;
		case 2:
			img = new PCSImage_std_dual(_name, g_allargs.xsize, g_allargs.ysize, g_allargs.colourdepth, g_allargs.colours, g_allargs.planes, g_allargs.palettes, g_allargs.palette_skip);
			break;
		case 3:
			img = new PCSImage_std_tri(_name, g_allargs.xsize, g_allargs.ysize, g_allargs.colourdepth, g_allargs.colours, g_allargs.planes, g_allargs.palettes, g_allargs.palette_skip);
			break;
		case 4:
			img = new PCSImage_std_quad(_name, g_allargs.xsize, g_allargs.ysize, g_allargs.colourdepth, g_allargs.colours, g_allargs.planes, g_allargs.palettes, g_allargs.palette_skip);
			break;
		default:
			{ printf("critial error - invalid number of fields specified!\n"); getchar(); exit(1); }
		}
				
		if (!img) { printf("critical error - could not allocate memory for image class!\n"); getchar(); exit(1); }
	}

	if (1 && success)
	{	
		if (type >= 0)
		{
			// adopt read image as source
			img->assign_source(fib);
		}
		else
		{
			if (!g_allargs.quiet)
				printf("> reading source image [%s]...\n", _name);		

			if (!img->read_source(type))
			{
				success = false;
				printf("error reading source image!\n");	
				//printf("press any key...\n"); getchar();							
			}
			else
			{
				if (g_allargs.verbose)
					printf("> source image read ok.\n");
			}
		}
	}

	if (1 && success)
	{	
		if (g_allargs.verbose)
			printf("> sanitising source image for conversion...\n");		

		if (!img->sanitise_source())
		{
			success = false;
			printf("error reading source image!\n");		
			//printf("press any key...\n"); getchar();							
		}
		else
		{
			if (g_allargs.verbose)
				printf("> image sanitised ok.\n");
		}
	}
			
	if (1 && success)
	{	
		// diagnostics: write reference image as PNG
		if (g_allargs.diagnostics)
		{
			if (g_allargs.verbose)
				printf("> diagnostics: writing reference image...\n");
				
			//std::string refname(_name);
			//size_t p = refname.find_last_of(".");
			//if (p != std::string::npos)
			//	refname = refname.substr(0, p);
			//refname.append("R.png");
			std::string refname("srcref.png");
			
			if (!img->write_reference(refname.c_str()))
			{
				success = false;
				printf("error writing reference PNG!\n");
				//printf("press any key...\n"); getchar();							
			}
			else
			{
				if (g_allargs.verbose)				
					printf("> diagnostics: reference PNG written ok.\n");
			}
		}
	}

	if (g_allargs.ccmode > CCMode_NONE)
	{
#ifdef CRYPTOCHROME
		if (1 && success)
		{
			if (!g_allargs.quiet)
				printf("> converting [cryptochrome] image...\n");
				
			img->convert_cc();
		}
#else
		printf("> error - CryptoChrome support not compiled into this build!\n");
#endif
	}
	else
	{

		// convert image...			
		if (1 && success)
		{
			if (!g_allargs.quiet)
				printf("> converting image...\n");
				
			if (!img->convert())
			{
				success = false;
				printf("> error converting image!\n");					
				//printf("press any key...\n"); getchar();							
			}
			else
			{
				if (g_allargs.verbose)
					printf("> image converted ok.\n");
			}
		}
		
		if (1 && success)
		{
			if (!g_allargs.quiet)
				printf("> creating output file...\n");
				
			// todo: extension rename never happens because outname is always valid?
			std::string outname(_outname);
			if (outname.length() == 0)
			{
				outname = _name;
				size_t p = outname.find_last_of(".");
				if (p != std::string::npos)
					outname = outname.substr(0, p);
					
				if (g_allargs.storage == StorageFmt_RAW_DISPLAY || g_allargs.storage == StorageFmt_RAW_ILEAVED_DISPLAY)
					outname.append(".pcr");
				else
					outname.append(".pcs");
			}
							
			if (!img->generate_pcs_file(outname.c_str()))
			{
				success = false;
				printf("> error generating file format!\n");					
				//printf("press any key...\n"); getchar();							
			}
			else
			{			
				if (g_allargs.verbose)
					printf("> conversion completed ok.\n");
			}
		}
	}

	delete img; img = 0;
	
	// clean up image package
	ImagePkg::destroy();
		
	return success;
}

std::vector<unsigned int> *pccgenpool = 0;

bool ccgen_process_multiple_sources(const char* const _outname)
{
	// nudge image package into emitting version info before any other messages
	ImagePkg::instance();
	
	// create source image, and perform any preparation processing (i.e. bitdepth, size)
	
	bool success = true;

	static const int poolwidth = 1024;

	int s = (pccgenpool->size() + (poolwidth-1)) & -poolwidth;
	int w = poolwidth;
	int h = s >> 10;

	// set PCSImage output dimensions to pixel pool image dimensions
	g_allargs.xsize = w;
	g_allargs.ysize = h;

	PCSImage* img = 0;
	img = new PCSImage_std_sgl(_outname, g_allargs.xsize, g_allargs.ysize, g_allargs.colourdepth, g_allargs.colours, g_allargs.planes, g_allargs.palettes, g_allargs.palette_skip);
			
	if (!img) { printf("error - could not allocate memory for image class!\n"); getchar(); exit(1); }

	if (1 && success)
	{	
		if (!g_allargs.quiet)
			printf("> generating source image (%dx%d) from pixel pool...\n", w, h);		

		if (!img->read_blank(w, h))
		{
			success = false;
			printf("error generating source image!\n");	
			//printf("press any key...\n"); getchar();							
		}
		else
		{
			FIBITMAP* poolimg = img->get_active_source();

			int i = 0;
			int imax = pccgenpool->size();
			for (int y = 0; y < h; ++y)
			{
				for (int x = 0; x < w; ++x)
				{
					if (i < imax)
					{
						unsigned int rgb = pccgenpool->at(i++);

						byte_t r,g,b;
						r = (rgb >> 16) & 255;
						g = (rgb >> 8) & 255;
						b = (rgb >> 0) & 255;

						ImagePkg::instance().put_rgb(poolimg,x,y,h,r,g,b);
					}
				}
			}

			if (g_allargs.verbose)
				printf("> source image generated ok.\n");
		}
	}

#ifdef CRYPTOCHROME
	if (1 && success)
	{
		if (!g_allargs.quiet)
			printf("> converting [cryptochrome] texture set...\n");
			
		img->convert_cc();
	}
#else
	printf("> error - CryptoChrome support not compiled into this build!\n");
#endif
	
	delete img; img = 0;
	
	// clean up image package
	ImagePkg::destroy();
		
	return success;
}


bool ccgen_append_source(const char* const _name)
{
	if (!pccgenpool)
		pccgenpool = new std::vector<unsigned int>;

	// nudge image package into emitting version info before any other messages
	ImagePkg::instance();
	
	// create source image, and perform any preparation processing (i.e. bitdepth, size)
	
	bool success = true;

	std::string sname(_name);
	std::transform(sname.begin(), sname.end(), sname.begin(), mytolower);

	int type = -1;
	if (sname.find(".jpg") != std::string::npos)
		type = 0;
	else if (sname.find(".tga") != std::string::npos)
		type = 1;
	else if (sname.find(".png") != std::string::npos)
		type = 2;
	else if (sname.find(".bmp") != std::string::npos)
		type = 3;
	else
	{
		success = false;
		printf("error - unknown source image file extension!\n");		
	}

	if (1 && success)
	{	
		if (!g_allargs.quiet)
			printf("> adding source image [%s]...\n", _name);		

		FIBITMAP *fib = 0;
		{
			if (type == 0)
				fib = ImagePkg::instance().read_jpg(sname.c_str());
			else if (type == 1)
				fib = ImagePkg::instance().read_tga(sname.c_str());
			else if (type == 2)
				fib = ImagePkg::instance().read_png(sname.c_str());
			else if (type == 3)
				fib = ImagePkg::instance().read_bmp(sname.c_str());
			else
				{ printf("error - unknown image type (%d)\n", type); getchar(); exit(1); }

			if (g_allargs.verbose)
				printf("> source image read ok.\n");
		}

		{
			int d = ImagePkg::instance().get_depth(fib);
			if (g_allargs.verbose)
				printf("check colour depth: %d ", d);
			if (d != 24)
			{
				if (g_allargs.verbose)
					printf("...converting to 24bit RGB... ");
				FIBITMAP *tmp = ImagePkg::instance().reformat(fib);
				if (!tmp)
				{
					if (g_allargs.verbose)
						printf("...failed!\n");				
					return false;
				}
				ImagePkg::instance().unload(fib);
				fib = tmp;
			}
			if (g_allargs.verbose)
				printf("...ok.\n");
		}

		// add pixels to colour pool
		{
			int w = 0;
			int h = 0;
			ImagePkg::instance().get_size(fib, w, h);

			// make some space
			pccgenpool->reserve(pccgenpool->size() + (w*h));

			// append pixels
			for (int y = 0; y < h; ++y)
			{
				for (int x = 0; x < w; ++x)
				{
					byte_t r,g,b;
					ImagePkg::instance().get_rgb(fib,x,y,h,r,g,b);
					unsigned int rgb = (unsigned int)r << 16 | (unsigned int)g << 8 | (unsigned int)b << 0;
					pccgenpool->push_back(rgb);
				}
			}

			if (g_allargs.verbose)
				printf("> added %d pixels to colour pool (%d total).\n", (w*h), pccgenpool->size());
		}
/*
		// create pool image
		{
			static const int poolwidth = 1024;

			int s = (pccgenpool->size() + (poolwidth-1)) & -poolwidth;
			int w = poolwidth;
			int h = s >> 10;

			FIBITMAP *poolimg = ImagePkg::instance().read_blank(w, h);
			for (int y = 0; y < h; ++y)
			{
				for (int x = 0; x < w; ++x)
				{
					u32 rgb = pccgenpool->at(i);

					byte_t r,g,b;
					r = (rgb >> 16) & 255;
					g = (rgb >> 8) & 255;
					b = (rgb >> 0) & 255;

					ImagePkg::instance().put_rgb(poolimg,x,y,h,r,g,b);
				}
			}

			//ImagePkg::instance().write_png(active_source_, _name);

			if (g_allargs.verbose)
				printf("> created %d x %d pool image.\n", w, h);
		}
*/
		if (fib)
			ImagePkg::instance().unload(fib);
	}
			
	return success;
}