// ---------------------------------------------------------------------------------------------------------------------
//	PhotoChrome v6.xx - Atari ST interlaced image reduction framework
//	1991-2016 Douglas Little (d.m.l)
// ---------------------------------------------------------------------------------------------------------------------
//	special thanks to:
//		evil@DHS, RA/PDX, Troed/SYNC for help testing and input on improvements
// ---------------------------------------------------------------------------------------------------------------------

// ====================================================================
//	error measurement unit
// ====================================================================

#include "common.h"
#include "config.h"

#if defined(__ATARI__) || defined(__MINGW32__)
#include "FreeImage.3.15.3.h"
#else
#include "FreeImage.3.15.4.h"
#endif

#include "imagepkg.h"
#include "pcsimage.h"

extern Arguments g_allargs;

// --------------------------------------------------------------------

// --------------------------------------------------------------------

DiagnosticUnit::DiagnosticUnit
(	
	signed short _w, signed short _h, 
	signed short _colourdepth, 
	signed short _fields, 
	DiagnosticSpace _space, 
	const char* const _ename
)
	: w_(_w), h_(_h), 
	  colourdepth_(_colourdepth), 
	  fields_(_fields),
	  space_(_space), 
	  err_name_(_ename),
	  err_(0),
	  dots_(0),
	  imgpkg_(ImagePkg::instance())
{
}

DiagnosticUnit::~DiagnosticUnit()
{
}

void DiagnosticUnit::reset()
{
	err_ = 0;
	dots_ = 0;
}

// --------------------------------------------------------------------

class DiagnosticUnit_Bias : public DiagnosticUnit
{
	calc_real& bias_;
	
public:

	DiagnosticUnit_Bias
	(
		signed short _w, signed short _h, 
		signed short _colourdepth, 
		signed short _fields,
		DiagnosticSpace _space, 
		const char* const _ename, 
		calc_real& _bias
	)
		: DiagnosticUnit(_w, _h, _colourdepth, _fields, _space, _ename),
		  bias_(_bias)
	{
	}
		
	virtual void sample
	(
		FIBITMAP* _result, signed short _x, signed short _y, 
		calc_real _fpx, calc_real _fpy, calc_real _fpz,
		calc_real _fsx, calc_real _fsy, calc_real _fsz
	)
	{
		calc_real fex = _fpx - _fsx;
		calc_real fey = _fpy - _fsy;
		calc_real fez = _fpz - _fsz;
		
		calc_real b = (fex + fey + fez);
		
		err_ += b;
						
		dots_ += 3.0f;
		
		unsigned short er = 255, eg = 255, eb = 255;		
		
		if ((xabs(fex) < (1.0 / 255.0)) &&
			(xabs(fey) < (1.0 / 255.0)) &&
			(xabs(fez) < (1.0 / 255.0)))
		{
			// white = exact colour match
		}
		else if (b > 1e-6)
		{
			// green = too high
			er = 0;
			eg = 255;
			eb = 0;
		}
		else if (b < -1e-6)
		{
			// blue = too low
			er = 0;
			eg = 0;
			eb = 255;
		}
		else
		{
			// yellow - pixel bias is ~zero but not exact colour match
			er = 255;
			eg = 255;
			eb = 0;
		}
					
		imgpkg_.put_rgb(_result, (int)_x, (int)_y, h_, (byte_t)er, (byte_t)eg, (byte_t)eb);					
	}
		
	virtual void evaluate(FIBITMAP* _result)
	{ 
		imgpkg_.write_png(_result, err_name_.c_str());		
		
		// output bias level
		bias_ = err_ / dots_;
		
		calc_real err = bias_ * 255.0f;
		
		printf("Bias   = %.6f / 255 (%.4f%%)\n", (double)err, (double)(err / 255.0f) * 100.0f);		
	}
	
};

// --------------------------------------------------------------------

class DiagnosticUnit_Range : public DiagnosticUnit
{
	
	calc_real& bias_;
	
public:

	DiagnosticUnit_Range
	(
		signed short _w, signed short _h, 
		signed short _colourdepth, 
		signed short _fields,
		DiagnosticSpace _space, 
		calc_real& _bias
	)
		: DiagnosticUnit(_w, _h, _colourdepth, _fields, _space),
		  bias_(_bias)
	{
	}
		
	virtual void sample
	(
		FIBITMAP* _result, signed short _x, signed short _y, 
		calc_real _fpx, calc_real _fpy, calc_real _fpz,
		calc_real _fsx, calc_real _fsy, calc_real _fsz
	)
	{
		calc_real fex = (_fpx - _fsx) - bias_;
		calc_real fey = (_fpy - _fsy) - bias_;
		calc_real fez = (_fpz - _fsz) - bias_;
		
		calc_real b = (fex + fey + fez);
		
		calc_real l = (_fsx + _fsy + _fsz) / 3.0f;
		
		if (l >= 0.5f)
		{
			uerr_ += b;
			udots_ += 3.0;
		}
		else
		{
			lerr_ -= b;
			ldots_ += 3.0;
		}
		
		err_ += b;			
		dots_ += 3.0f;
	}
		
	virtual void evaluate(FIBITMAP* _result)
	{ 		
		calc_real uerr = (udots_ > 0.0f) ? ((uerr_ * 255.0f) / udots_) : 0.0f;
		calc_real lerr = (ldots_ > 0.0f) ? ((lerr_ * 255.0f) / ldots_) : 0.0f;
		
		printf("uRange = %.6f / 255 (%.4f%%)\n", (double)uerr, (double)((uerr / 255.0f) * 100.0f));		
		printf("lRange = %.6f / 255 (%.4f%%)\n", (double)lerr, (double)((lerr / 255.0f) * 100.0f));

		if ((xabs(uerr_) > 1e-6f) && 
			(xabs(lerr_) > 1e-6f))				
			printf("nRange = %.6f\n", (double)(uerr / lerr));		
	}
	
	void reset()
	{
		DiagnosticUnit::reset();
		
		uerr_ = 0;
		lerr_ = 0;	
		udots_ = 0;
		ldots_ = 0;
	}
	
	calc_real uerr_;
	calc_real lerr_;
	calc_real udots_;
	calc_real ldots_;
	
};

// --------------------------------------------------------------------

class DiagnosticUnit_LinERR : public DiagnosticUnit
{
	
public:

	DiagnosticUnit_LinERR
	(
		signed short _w, signed short _h, 
		signed short _colourdepth, 
		signed short _fields,
		DiagnosticSpace _space, 
		const char* const _ename
	)
		: DiagnosticUnit(_w, _h, _colourdepth, _fields, _space, _ename)
	{
	}
		
	virtual void sample
	(
		FIBITMAP* _result, signed short _x, signed short _y, 
		calc_real _fpx, calc_real _fpy, calc_real _fpz,
		calc_real _fsx, calc_real _fsy, calc_real _fsz
	)
	{
		calc_real fex = _fpx - _fsx;
		calc_real fey = _fpy - _fsy;
		calc_real fez = _fpz - _fsz;
		
		calc_real fax = xabs(fex);
		calc_real fay = xabs(fey);
		calc_real faz = xabs(fez);
		
		err_ += (fax + fay + faz) / 3.0f;
		dots_ += 1.0f;
		
		// back into 8-bit space				
		unsigned short ex = f2col(fax * 1.0f);
		unsigned short ey = f2col(fay * 1.0f);
		unsigned short ez = f2col(faz * 1.0f);
		
		imgpkg_.put_rgb(_result, (int)_x, (int)_y, h_, (byte_t)ex, (byte_t)ey, (byte_t)ez);					
	}
		
	virtual void evaluate(FIBITMAP* _result)
	{ 
		imgpkg_.write_png(_result, err_name_.c_str());		
		
		calc_real pcterr = (err_ * 100.0f) / dots_;
		
		printf("LinERR(%s) = %.6f%%\n", 
			(space_ == DiagnosticSpace_8bit) ? "8bit" : 
			(space_ == DiagnosticSpace_native_virtual) ? "virt" : 
			(space_ == DiagnosticSpace_field) ? "field" :
			"unknown", 
			pcterr);			
	}
	
};

// --------------------------------------------------------------------

class DiagnosticUnit_VecERR : public DiagnosticUnit
{
	
public:

	DiagnosticUnit_VecERR
	(
		signed short _w, signed short _h, 
		signed short _colourdepth, 
		signed short _fields,
		DiagnosticSpace _space, 
		const char* const _ename
	)
		: DiagnosticUnit(_w, _h, _colourdepth, _fields, _space, _ename)
	{
	}
		
	virtual void sample
	(
		FIBITMAP* _result, signed short _x, signed short _y, 
		calc_real _fpx, calc_real _fpy, calc_real _fpz,
		calc_real _fsx, calc_real _fsy, calc_real _fsz
	)
	{
		calc_real fex = _fpx - _fsx;
		calc_real fey = _fpy - _fsy;
		calc_real fez = _fpz - _fsz;
		
		fex *= fex;
		fey *= fey;
		fez *= fez;
		
		err_ += sqrt(fex + fey + fez);
		dots_ += 1.0f;
		
		// back into 8-bit space				
		unsigned short ex = f2col(fex * 128.0f);
		unsigned short ey = f2col(fey * 128.0f);
		unsigned short ez = f2col(fez * 128.0f);
		
		imgpkg_.put_rgb(_result, (int)_x, (int)_y, h_, (byte_t)ex, (byte_t)ey, (byte_t)ez);					
	}
		
	virtual void evaluate(FIBITMAP* _result)
	{ 
		imgpkg_.write_png(_result, err_name_.c_str());		
		
		calc_real pcterr = (err_ * 100.0f) / dots_;

		printf("VecERR(%s) = %.6f%%\n", 
			(space_ == DiagnosticSpace_8bit) ? "8bit" : 
			(space_ == DiagnosticSpace_native_virtual) ? "virt" : 
			(space_ == DiagnosticSpace_field) ? "field" :
			"unknown", 
			(double)pcterr);			
	}
	
};

// --------------------------------------------------------------------

class DiagnosticUnit_SquERR : public DiagnosticUnit
{
	
public:

	DiagnosticUnit_SquERR
	(
		signed short _w, signed short _h, 
		signed short _colourdepth, 
		signed short _fields,
		DiagnosticSpace _space
	)
		: DiagnosticUnit(_w, _h, _colourdepth, _fields, _space)
	{
	}
		
	virtual void sample
	(
		FIBITMAP* _result, signed short _x, signed short _y, 
		calc_real _fpx, calc_real _fpy, calc_real _fpz,
		calc_real _fsx, calc_real _fsy, calc_real _fsz
	)
	{
		calc_real fex = _fpx - _fsx;
		calc_real fey = _fpy - _fsy;
		calc_real fez = _fpz - _fsz;
		
		fex *= fex;
		fey *= fey;
		fez *= fez;
		
		err_ += fex + fey + fez;
		dots_ += 1.0f;
	}
		
	virtual void evaluate(FIBITMAP* _result)
	{ 		
		calc_real err = sqrt(err_) / dots_;
		
		printf("SquERR(%s) = %.8f\n", 
			(space_ == DiagnosticSpace_8bit) ? "8bit" : 
			(space_ == DiagnosticSpace_native_virtual) ? "virt" : 
			(space_ == DiagnosticSpace_field) ? "field" :
			"unknown", 
			(double)err);			
	}
	
};

// --------------------------------------------------------------------
// create all diagnostic units, to be executed in sequence

void PCSImage::init_diagnostic_units()
{
	// caution: persistent var - hack to get bias from one pass to another
	static calc_real bias = 0;	
	diagnosticunits_.clear();
	
	// 
	diagnosticunits_.push_back(new DiagnosticUnit_Bias  (display_width_, display_height_, colourdepth_, fields_, DiagnosticSpace_8bit, "bias.png",	bias));
	diagnosticunits_.push_back(new DiagnosticUnit_Range (display_width_, display_height_, colourdepth_, fields_, DiagnosticSpace_8bit, 				bias));

	// error measurements against native virtual source
	diagnosticunits_.push_back(new DiagnosticUnit_LinERR(display_width_, display_height_, colourdepth_, fields_, DiagnosticSpace_native_virtual, "linerrnv.png"));
	diagnosticunits_.push_back(new DiagnosticUnit_VecERR(display_width_, display_height_, colourdepth_, fields_, DiagnosticSpace_native_virtual, "vecerrnv.png"));
	diagnosticunits_.push_back(new DiagnosticUnit_SquERR(display_width_, display_height_, colourdepth_, fields_, DiagnosticSpace_native_virtual));

	// error measurements against 8-bit source
	diagnosticunits_.push_back(new DiagnosticUnit_LinERR(display_width_, display_height_, colourdepth_, fields_, DiagnosticSpace_8bit, "linerr8b.png"));
	diagnosticunits_.push_back(new DiagnosticUnit_VecERR(display_width_, display_height_, colourdepth_, fields_, DiagnosticSpace_8bit, "vecerr8b.png"));
	diagnosticunits_.push_back(new DiagnosticUnit_SquERR(display_width_, display_height_, colourdepth_, fields_, DiagnosticSpace_8bit));
}

// --------------------------------------------------------------------
