// ---------------------------------------------------------------------------------------------------------------------
//	PhotoChrome v6.xx - Atari ST interlaced image reduction framework
//	1991-2016 Douglas Little (d.m.l)
// ---------------------------------------------------------------------------------------------------------------------
//	special thanks to:
//		evil@DHS, RA/PDX, Troed/SYNC for help testing and input on improvements
// ---------------------------------------------------------------------------------------------------------------------

// ====================================================================
//	reusable image processing stuff
// ====================================================================

#include "common.h"
#include "config.h"

#if defined(__ATARI__) || defined(__MINGW32__)
#include "FreeImage.3.15.3.h"
#else
#include "FreeImage.3.15.4.h"
#endif

#include "imagepkg.h"
#include "pcsimage.h"

extern Arguments g_allargs;

// --------------------------------------------------------------------

DitherUnit::DitherUnit(signed short _field, signed short _colourdepth, signed short _fields, float _dither_level)
	: LCG(0xC1894B75 + (1317 * _field)),
	  colourdepth_(_colourdepth),
	  fields_(_fields),
	  field_(_field),
	  dither_level_(_dither_level)
{
}

void DitherUnit::reset()
{
	LCG::reset(0xC1894B75 + (1317 * field_));
	onReset();
}

// --------------------------------------------------------------------

class DitherUnit_PRNG : public DitherUnit
{
public:

	DitherUnit_PRNG(signed short _field, signed short _colourdepth, signed short _fields, float _dither_level)
		: DitherUnit(_field, _colourdepth, _fields, _dither_level)
	{
	}
	
	//virtual signed short sample(signed short _x, signed short _y)
	//{ 
	//	return ((lcg_sample() & ((dither_magnitude_<<1)-1))-dither_magnitude_);
	//}

	virtual void fsample(signed short _x, signed short _y, float& _xadj, float& _yadj, float& _zadj) 
	{
		float sample = ((float)lcg_sample() / 255.0f) - 0.5f;
		
		float dither = (dither_level_ * sample) / (float)(nvi_crange(colourdepth_, fields_));

		_xadj = dither;	 
		_yadj = dither;	 
		_zadj = dither;	 		 
	}
};
	
class DitherUnit_STIPPLE : public DitherUnit
{
public:

	DitherUnit_STIPPLE(signed short _field, signed short _colourdepth, signed short _fields, float _dither_level)
		: DitherUnit(_field, _colourdepth, _fields, _dither_level)
	{
	}
	
	//virtual signed short sample(signed short _x, signed short _y)
	//{ 
	//	signed short interlace = dither_magnitude_;
	//	signed short eval = (_x + (_y & 1)) & 1;
	//	if (eval)
	//		interlace -= (dither_magnitude_ << 1);	
	//	return interlace;
	//}

	virtual void fsample(signed short _x, signed short _y, float& _xadj, float& _yadj, float& _zadj) 
	{
		signed short eval = (_x + (_y & 1)) & 1;

		float dither = (dither_level_ * 0.25f) / (float)(nvi_crange(colourdepth_, fields_));

		if (eval)
			dither = -dither;
					
		_xadj = dither;	 
		_yadj = dither;	 
		_zadj = dither;	 
	}
};
	
class DitherUnit_CPRNG : public DitherUnit
{
public:

	DitherUnit_CPRNG(signed short _field, signed short _colourdepth, signed short _fields, float _dither_level)
		: DitherUnit(_field, _colourdepth, _fields, _dither_level)
	{
	}
	
	//virtual signed short sample(signed short _x, signed short _y)
	//{ 
	//	return ((lcg_sample() & ((dither_magnitude_<<1)-1))-dither_magnitude_);
	//}

	virtual void fsample(signed short _x, signed short _y, float& _xadj, float& _yadj, float& _zadj) 
	{
		float sample_x = ((float)lcg_sample() / 255.0f) - 0.5f;
		float sample_y = ((float)lcg_sample() / 255.0f) - 0.5f;
		float sample_z = ((float)lcg_sample() / 255.0f) - 0.5f;
		
		_xadj = (dither_level_ * sample_x) / (float)(nvi_crange(colourdepth_, fields_));
		_yadj = (dither_level_ * sample_y) / (float)(nvi_crange(colourdepth_, fields_));
		_zadj = (dither_level_ * sample_z) / (float)(nvi_crange(colourdepth_, fields_));
	}
};

class DitherUnit_MATRIX : public DitherUnit
{
private:

	int size_;			
	const float* m_;
	float bias_;
	float scale_;
	
	void init()
	{
		int n = size_ * size_;
		float h = 0.0;
		float l = 100.0;
		float m = 0;
		float mc = 0;
		for (int i = 0; i < n; ++i)
		{
			float s = m_[i];
			h = xmax(h, s);
			l = xmin(l, s);
			m += s;
			mc += 1.0;
		}
		m /= mc;
		bias_ = m;
		scale_ = 1.0 / ((h-m)+0.5);
	}
		
public:

	DitherUnit_MATRIX(int _size, const float* _m, signed short _field, signed short _colourdepth, signed short _fields, float _dither_level)
		: DitherUnit(_field, _colourdepth, _fields, _dither_level),
		  m_(_m),
		  size_(_size)
	{
		init();
	}
	
	virtual void fsample(signed short _x, signed short _y, float& _xadj, float& _yadj, float& _zadj) 
	{
		signed short index = ((_x % size_) + ((_y % size_) * size_));

		float dither = ((m_[index] - bias_) * scale_ * dither_level_ * 0.5f) / (float)(nvi_crange(colourdepth_, fields_));
					
		_xadj = dither;	 
		_yadj = dither;	 
		_zadj = dither;	 
	}
};

// --------------------------------------------------------------------

//	dc.b		04,00,02,00
//	dc.b		00,01,00,01
//	dc.b		02,00,03,00
//	dc.b		00,01,00,01

// --------------------------------------------------------------------

class DitherUnit_TINY : public DitherUnit_MATRIX 
{
private:

	static const int matrix_size_;			
	static const float matrix_[];
		
public:

	DitherUnit_TINY(signed short _field, signed short _colourdepth, signed short _fields, float _dither_level)
		: DitherUnit_MATRIX(matrix_size_, matrix_, _field, _colourdepth, _fields, _dither_level)
	{
	}
};

const int DitherUnit_TINY::matrix_size_ = 3;			
const float DitherUnit_TINY::matrix_[] =
{
	 2, 6, 4,
	 5, 0, 1,
	 8, 3, 7
};

// --------------------------------------------------------------------

class DitherUnit_BAYER4 : public DitherUnit_MATRIX 
{
private:

	static const int matrix_size_;			
	static const float matrix_[];
		
public:

	DitherUnit_BAYER4(signed short _field, signed short _colourdepth, signed short _fields, float _dither_level)
		: DitherUnit_MATRIX(matrix_size_, matrix_, _field, _colourdepth, _fields, _dither_level)
	{
	}
};

const int DitherUnit_BAYER4::matrix_size_ = 4;	
const float DitherUnit_BAYER4::matrix_[] =
{
	  0,  8,  2, 10,
	 12,  4, 14,  6,
	  3, 11,  1,  9,
	 15,  7, 13,  5
};

// --------------------------------------------------------------------

class DitherUnit_MESHED : public DitherUnit_MATRIX 
{
private:

	static const int matrix_size_;			
	static const float matrix_[];
		
public:

	DitherUnit_MESHED(signed short _field, signed short _colourdepth, signed short _fields, float _dither_level)
		: DitherUnit_MATRIX(matrix_size_, matrix_, _field, _colourdepth, _fields, _dither_level)
	{
	}
};

const int DitherUnit_MESHED::matrix_size_ = 4;	
const float DitherUnit_MESHED::matrix_[] =
{
	 11,  4,  6,  9,
	 12,  0,  2, 14,
	  7,  8, 10,  5,
	  3, 15, 13,  1
};

// --------------------------------------------------------------------

class DitherUnit_PANMAGIC : public DitherUnit_MATRIX 
{
private:

	static const int matrix_size_;			
	static const float matrix_[];
		
public:

	DitherUnit_PANMAGIC(signed short _field, signed short _colourdepth, signed short _fields, float _dither_level)
		: DitherUnit_MATRIX(matrix_size_, matrix_, _field, _colourdepth, _fields, _dither_level)
	{
	}
};

const int DitherUnit_PANMAGIC::matrix_size_ = 4;	
const float DitherUnit_PANMAGIC::matrix_[] =
{
	  7, 10, 13,  0,
	 12,  1,  6, 11,
	  2, 15,  8,  5,
	  9,  3,  3, 14
};
	
// --------------------------------------------------------------------

class DitherUnit_BAYER8 : public DitherUnit_MATRIX 
{
private:

	static const int matrix_size_;			
	static const float matrix_[];
		
public:

	DitherUnit_BAYER8(signed short _field, signed short _colourdepth, signed short _fields, float _dither_level)
		: DitherUnit_MATRIX(matrix_size_, matrix_, _field, _colourdepth, _fields, _dither_level)
	{
	}
};

const int DitherUnit_BAYER8::matrix_size_ = 8;	
const float DitherUnit_BAYER8::matrix_[] =
{
	  1, 33,  9, 41,  3, 35, 11, 43,
	 49, 17, 57, 25, 51, 19, 59, 27,
	 13, 45,  5, 37, 15, 47,  7, 39,
	 61, 29, 53, 21, 63, 31, 55, 23,
	  4, 36, 12, 44,  2, 34, 10, 42,
	 52, 20, 60, 28, 50, 18, 58, 26,
	 16, 48,  8, 40, 14, 46,  6, 38,
	 64, 32, 56, 24, 62, 30, 54, 22
};

// --------------------------------------------------------------------

void PCSImage::init_dither_unit(int _field)
{
	delete ditherunit_; ditherunit_ = 0;
	
	DitherUnit* newunit = 0;	
	switch (g_allargs.dithertype)
	{
	default:
	case DitherType_OFF:
		break;
	case DitherType_PRNG:
		newunit = new DitherUnit_PRNG(_field, colourdepth_, fields_, g_allargs.ditherlevel);
		break;
	case DitherType_STIPPLE:
		newunit = new DitherUnit_STIPPLE(_field, colourdepth_, fields_, g_allargs.ditherlevel);
		break;
	case DitherType_CPRNG:
		newunit = new DitherUnit_CPRNG(_field, colourdepth_, fields_, g_allargs.ditherlevel);
		break;
	case DitherType_TINY:
		newunit = new DitherUnit_TINY(_field, colourdepth_, fields_, g_allargs.ditherlevel);
		break;
	case DitherType_BAYER4:
		newunit = new DitherUnit_BAYER4(_field, colourdepth_, fields_, g_allargs.ditherlevel);
		break;
	case DitherType_MESHED:
		newunit = new DitherUnit_MESHED(_field, colourdepth_, fields_, g_allargs.ditherlevel);
		break;
	case DitherType_PANMAGIC:
		newunit = new DitherUnit_PANMAGIC(_field, colourdepth_, fields_, g_allargs.ditherlevel);
		break;
	case DitherType_BAYER8:
		newunit = new DitherUnit_BAYER8(_field, colourdepth_, fields_, g_allargs.ditherlevel);
		break;
	}
	
	ditherunit_ = newunit;
}

// --------------------------------------------------------------------
