// ---------------------------------------------------------------------------------------------------------------------
//	PhotoChrome v6.xx - Atari ST interlaced image reduction framework
//	1991-2016 Douglas Little (d.m.l)
// ---------------------------------------------------------------------------------------------------------------------
//	special thanks to:
//		evil@DHS, RA/PDX, Troed/SYNC for help testing and input on improvements
// ---------------------------------------------------------------------------------------------------------------------

// ====================================================================
//	error diffusion unit
// ====================================================================

#include "common.h"
#include "config.h"

#if defined(__ATARI__) || defined(__MINGW32__)
#include "FreeImage.3.15.3.h"
#else
#include "FreeImage.3.15.4.h"
#endif

#include "imagepkg.h"
#include "pcsimage.h"

extern Arguments g_allargs;

// --------------------------------------------------------------------

// --------------------------------------------------------------------

ErrorUnit::ErrorUnit(signed short _w, signed short _h, signed short _colourdepth, signed short _fields, float _error_level)
	: LCG(0),
	  w_(_w), h_(_h),
	  fields_(_fields),
	  direction_(1),
	  field_(0),
	  iteration_(0), max_iterations_(1),
	  field_magnitude_(1.0f),
	  error_scale_(_error_level)
{
	// field error buffers
	for (int i = 0; i < max_fields; ++i)
		fieldbuffers_[i] = 0;

	spatialbuffer_ = 0;	// FS (spatial) error
}

ErrorUnit::~ErrorUnit()
{
	for (int i = 0; i < max_fields; ++i)
		delete[] fieldbuffers_[i];

	delete spatialbuffer_;
}

void ErrorUnit::set_field(signed short _field)
{
	field_ = _field;
	LCG::reset(0xA2136D37 + (1529 * _field));

	int field_iteration = (iteration_ * fields_) + _field;
	int max_field_iterations = (max_iterations_ * fields_);

	calc_real t = 1.0f;
	if (max_field_iterations > 1)
		t = calc_real(field_iteration) / calc_real(max_field_iterations-1);
		
	field_magnitude_ = error_scale_ * t;

	if (g_allargs.verbose)	
		printf("field error magnitude: %.4f\n", (double)field_magnitude_);
		
	// important - spatial error diffusion is cumulative so each field needs re-initialised to 0's
	if (spatialbuffer_)
	{
		datasize_t pixels = w_ * h_;
		spatialbuffer_ = new xyzerror[pixels];				
	}		
}

void ErrorUnit::set_iteration(signed short _i, signed short _mi)
{
	iteration_ = _i;	
	max_iterations_ = _mi;
}

void ErrorUnit::set_sweep(signed short _direction)
{
	direction_ = _direction;
}

// --------------------------------------------------------------------

class ErrorUnit_FIELD : public ErrorUnit
{
	
public:

	ErrorUnit_FIELD(signed short _w, signed short _h, signed short _colourdepth, signed short _fields, float _error_level)
		: ErrorUnit(_w, _h, _colourdepth, _fields, _error_level)
	{
		datasize_t pixels = _w * _h;
		for (int f = 0; f < fields_; ++f)
			fieldbuffers_[f] = new xyzerror[pixels];
	}
	
// i		d		s1	a1	e1		s2	a2	e2
// 0		128		0	120	-8		-	-	-
// 1		128		0	120	-8		-8	126	-2
// 2		128		-10	130	+2		0	<stop>	

// i		d		s1	a1	e1		s2	a2	e2
// 0		128		0	120	-8		-	-	-
// 1		128		0	120	-8		-8	134	+6
// 2		128		+6	122	-6		-8	134	+6		
		

	virtual void sample(signed short _x, signed short _y, float& _ex, float& _ey, float& _ez)
	{ 
		//const xyzerror& et = errorbuffers_[field_ & 1];
		xyzerror* peo = fieldbuffers_[(field_ + 1) % fields_];
		const xyzerror& eo = peo[(_y * w_) + _x];
		
		// scale error from feedback, to encourage convergeance
		_ex = (eo.x * field_magnitude_);
		_ey = (eo.y * field_magnitude_);
		_ez = (eo.z * field_magnitude_);
	}
	
	// feedback error from this frame
	virtual void feedback(signed short _x, signed short _y, float _ex, float _ey, float _ez)
	{
		xyzerror* pet = fieldbuffers_[field_ % fields_];
		xyzerror& et = pet[(_y * w_) + _x];

		et.x = _ex;
		et.y = _ey;
		et.z = _ez;		
	}
	
	// field-only interface just reflects combined interface, since only field is implemented anyway
	virtual void sample_field(signed short _x, signed short _y, float& _ex, float& _ey, float& _ez)
	{ 
		ErrorUnit_FIELD::sample(_x, _y, _ex, _ey, _ez);
	}
	
	// feedback error from this frame
	virtual void feedback_field(signed short _x, signed short _y, float _ex, float _ey, float _ez)
	{
		ErrorUnit_FIELD::feedback(_x, _y, _ex, _ey, _ez);
	}		
};

// --------------------------------------------------------------------

class ErrorUnit_SPATIAL : public ErrorUnit
{
protected:
	
	int mw_;
	int mh_;
	const float* m_;
	float smag_;
	
public:

	ErrorUnit_SPATIAL(signed short _w, signed short _h, signed short _colourdepth, signed short _fields, float _error_level, int _mw, int _mh, const float* _m)
		: ErrorUnit(_w, _h, _colourdepth, _fields, _error_level),		  
		  mw_(_mw),
		  mh_(_mh),
		  m_(_m)
	{
		datasize_t pixels = _w * _h;
		spatialbuffer_ = new xyzerror[pixels];				
		smag_ = _error_level / (float)fields_;
	}
	
	virtual void sample(signed short _x, signed short _y, float& _ex, float& _ey, float& _ez)
	{ 
		xyzerror* peo = spatialbuffer_;
		const xyzerror& eo = peo[(_y * w_) + _x];
		
		// scale error from feedback
		_ex = (eo.x * smag_);
		_ey = (eo.y * smag_);
		_ez = (eo.z * smag_);
	}
	
	// feedback error from this frame
	virtual void feedback(signed short _x, signed short _y, float _ex, float _ey, float _ez)
	{
		xyzerror* pet = spatialbuffer_;

		const int mhwidth = mw_ >> 1;
		
		int my = 0;
		int ye = _y + mh_;
		int yy = _y;
		for (; yy < ye; ++yy, ++my)
		{
			if (yy >= 0 && yy < h_)
			{
				xyzerror* et_y = &pet[(yy * w_)];
						
				int mx = 0;								
				int xe = _x + ((mhwidth+1) * direction_);
				int xx = _x - (mhwidth * direction_);
				for (; xx != xe; xx += direction_, ++mx)
				{
					if (xx >= 0 && xx < w_)
					{
						
						xyzerror& et = et_y[xx];
						float c = m_[(my * mw_) + mx];
						
						et.x += _ex * c;
						et.y += _ey * c;
						et.z += _ez * c;	
					}
				}
			}
		}

/*
		xyzerror* pet = spatialbuffer_;

		xyzerror* et_y0 = &pet[((_y + 0) * w_) + _x];
		xyzerror* et_y1 = &pet[((_y + 1) * w_) + _x];
		
		// handle alternating x-sweep		
		signed short xbeg = 0;
		signed short xend = (w_- 1);
		if (direction_ < 0)
		{
			xend = 0;
			xbeg = (w_- 1);
		}

		static const float c7 = 7.0 / 16.0;
		static const float c1 = 1.0 / 16.0;
		static const float c5 = 5.0 / 16.0;
		static const float c3 = 3.0 / 16.0;
				
		// floyd-steinberg matrix with clipping
		if (_x != xend)
		{
			xyzerror& et = et_y0[direction_];
			et.x += _ex * c7;
			et.y += _ey * c7;
			et.z += _ez * c7;		

			if (_y != (h_ - 1))
			{
				xyzerror& et = et_y1[direction_];
				et.x += _ex * c1;
				et.y += _ey * c1;
				et.z += _ez * c1;
			}
		}

		if (_y != (h_ - 1))
		{
			xyzerror& et = et_y1[0];
			et.x += _ex * c5;
			et.y += _ey * c5;
			et.z += _ez * c5;		
	
			if (_x != xbeg)
			{
				xyzerror& et = et_y1[-direction_];
				et.x += _ex * c3;
				et.y += _ey * c3;
				et.z += _ez * c3;		
			}
		}	
*/
	}
	
	// spatial-only interface just reflects combined interface, since only spatial is implemented anyway
	virtual void sample_spatial(signed short _x, signed short _y, float& _ex, float& _ey, float& _ez)
	{ 
		ErrorUnit_SPATIAL::sample(_x, _y, _ex, _ey, _ez);
	}
	
	// feedback error from this frame
	virtual void feedback_spatial(signed short _x, signed short _y, float _ex, float _ey, float _ez)
	{
		ErrorUnit_SPATIAL::feedback(_x, _y, _ex, _ey, _ez);
	}	
};

// --------------------------------------------------------------------

class ErrorUnit_SPATIAL_FLOYD : public ErrorUnit_SPATIAL
{

	static const float c7 = 7.0 / 16.0;
	static const float c1 = 1.0 / 16.0;
	static const float c5 = 5.0 / 16.0;
	static const float c3 = 3.0 / 16.0;
		
	static const int matrixw_ = 3;
	static const int matrixh_ = 2;

	static const float matrix_[];
	
public:

	ErrorUnit_SPATIAL_FLOYD(signed short _w, signed short _h, signed short _colourdepth, signed short _fields, float _error_level)
		: ErrorUnit_SPATIAL(_w, _h, _colourdepth, _fields, _error_level, matrixw_, matrixh_, matrix_)
	{
	}
};

const float ErrorUnit_SPATIAL_FLOYD::matrix_[] =
{
	 0,  0, c7,
	c3, c5, c1
};

// --------------------------------------------------------------------
	
class ErrorUnit_SPATIAL_JAJUNI : public ErrorUnit_SPATIAL
{

	static const float c7 = 7.0 / 48.0;
	static const float c5 = 5.0 / 48.0;
	static const float c3 = 3.0 / 48.0;
	static const float c1 = 1.0 / 48.0;
	
	static const int matrixw_ = 5;
	static const int matrixh_ = 3;

	static const float matrix_[];
			
public:

	ErrorUnit_SPATIAL_JAJUNI(signed short _w, signed short _h, signed short _colourdepth, signed short _fields, float _error_level)
		: ErrorUnit_SPATIAL(_w, _h, _colourdepth, _fields, _error_level, matrixw_, matrixh_, matrix_)
	{
	}
};

const float ErrorUnit_SPATIAL_JAJUNI::matrix_[] =
{
	 0,  0,  0, c7, c5,
	c3, c5, c7, c5, c3,
	c1, c3, c5, c3, c1
};

// --------------------------------------------------------------------
	
class ErrorUnit_SPATIAL_ZHIGANG : public ErrorUnit_SPATIAL
{

	static const float c7 = 7.0 / 16.0;
	static const float c5 = 5.0 / 16.0;
	static const float c3 = 3.0 / 16.0;
	static const float c1 = 1.0 / 16.0;
	
	static const int matrixw_ = 5;
	static const int matrixh_ = 2;

	static const float matrix_[];
			
public:

	ErrorUnit_SPATIAL_ZHIGANG(signed short _w, signed short _h, signed short _colourdepth, signed short _fields, float _error_level)
		: ErrorUnit_SPATIAL(_w, _h, _colourdepth, _fields, _error_level, matrixw_, matrixh_, matrix_)
	{
	}
};

const float ErrorUnit_SPATIAL_ZHIGANG::matrix_[] =
{
	 0,  0,  0, c7,  0,
	c1, c3, c5,  0,  0
};

// --------------------------------------------------------------------
	
class ErrorUnit_SPATIAL_SHIAU1 : public ErrorUnit_SPATIAL
{

	static const float c4 = 4.0 / 8.0;
	static const float c2 = 2.0 / 8.0;
	static const float c1 = 1.0 / 8.0;
	
	static const int matrixw_ = 5;
	static const int matrixh_ = 2;

	static const float matrix_[];
			
public:

	ErrorUnit_SPATIAL_SHIAU1(signed short _w, signed short _h, signed short _colourdepth, signed short _fields, float _error_level)
		: ErrorUnit_SPATIAL(_w, _h, _colourdepth, _fields, _error_level, matrixw_, matrixh_, matrix_)
	{
	}
};

const float ErrorUnit_SPATIAL_SHIAU1::matrix_[] =
{
	 0,  0,  0, c4,  0,
	c1, c1, c2,  0,  0
};

// --------------------------------------------------------------------
	
class ErrorUnit_SPATIAL_SHIAU2 : public ErrorUnit_SPATIAL
{

	static const float c8 = 8.0 / 16.0;
	static const float c4 = 4.0 / 16.0;
	static const float c2 = 2.0 / 16.0;
	static const float c1 = 1.0 / 16.0;
	
	static const int matrixw_ = 7;
	static const int matrixh_ = 2;

	static const float matrix_[];
			
public:

	ErrorUnit_SPATIAL_SHIAU2(signed short _w, signed short _h, signed short _colourdepth, signed short _fields, float _error_level)
		: ErrorUnit_SPATIAL(_w, _h, _colourdepth, _fields, _error_level, matrixw_, matrixh_, matrix_)
	{
	}
};

const float ErrorUnit_SPATIAL_SHIAU2::matrix_[] =
{
	 0,  0,  0,  0, c8,  0,  0,
	c1, c1, c2, c4,  0,  0,  0
};

// --------------------------------------------------------------------
	
class ErrorUnit_SPATIAL_STUCKI : public ErrorUnit_SPATIAL
{

	static const float c8 = 8.0 / 42.0;
	static const float c4 = 4.0 / 42.0;
	static const float c2 = 2.0 / 42.0;
	static const float c1 = 1.0 / 42.0;
	
	static const int matrixw_ = 5;
	static const int matrixh_ = 3;

	static const float matrix_[];
			
public:

	ErrorUnit_SPATIAL_STUCKI(signed short _w, signed short _h, signed short _colourdepth, signed short _fields, float _error_level)
		: ErrorUnit_SPATIAL(_w, _h, _colourdepth, _fields, _error_level, matrixw_, matrixh_, matrix_)
	{
	}
};

const float ErrorUnit_SPATIAL_STUCKI::matrix_[] =
{
	 0,  0,  0, c8, c4,
	c2, c4, c8, c4, c2,
	c1, c2, c4, c2, c1
};

// --------------------------------------------------------------------
	
class ErrorUnit_SPATIAL_BURKES : public ErrorUnit_SPATIAL
{

	static const float c4 = 4.0 / 16.0;
	static const float c2 = 2.0 / 16.0;
	static const float c1 = 1.0 / 16.0;
	
	static const int matrixw_ = 5;
	static const int matrixh_ = 2;

	static const float matrix_[];
			
public:

	ErrorUnit_SPATIAL_BURKES(signed short _w, signed short _h, signed short _colourdepth, signed short _fields, float _error_level)
		: ErrorUnit_SPATIAL(_w, _h, _colourdepth, _fields, _error_level, matrixw_, matrixh_, matrix_)
	{
	}
};

const float ErrorUnit_SPATIAL_BURKES::matrix_[] =
{
	 0,  0,  0, c4, c2,
	c1, c2, c4, c2, c1
};

// --------------------------------------------------------------------
	
class ErrorUnit_SPATIAL_SIERRA : public ErrorUnit_SPATIAL
{

	static const float c5 = 5.0 / 32.0;
	static const float c4 = 4.0 / 32.0;
	static const float c3 = 3.0 / 32.0;
	static const float c2 = 2.0 / 32.0;
	
	static const int matrixw_ = 5;
	static const int matrixh_ = 3;

	static const float matrix_[];
			
public:

	ErrorUnit_SPATIAL_SIERRA(signed short _w, signed short _h, signed short _colourdepth, signed short _fields, float _error_level)
		: ErrorUnit_SPATIAL(_w, _h, _colourdepth, _fields, _error_level, matrixw_, matrixh_, matrix_)
	{
	}
};

const float ErrorUnit_SPATIAL_SIERRA::matrix_[] =
{
	 0,  0,  0, c5, c3,
	c2, c4, c5, c4, c2,
	 0, c2, c3, c2,  0
};

// --------------------------------------------------------------------
	
class ErrorUnit_SPATIAL_ATKINSON : public ErrorUnit_SPATIAL
{

	static const float cx = 1.0 / 8.0;
	
	static const int matrixw_ = 5;
	static const int matrixh_ = 3;

	static const float matrix_[];
			
public:

	ErrorUnit_SPATIAL_ATKINSON(signed short _w, signed short _h, signed short _colourdepth, signed short _fields, float _error_level)
		: ErrorUnit_SPATIAL(_w, _h, _colourdepth, _fields, _error_level, matrixw_, matrixh_, matrix_)
	{
	}
};

const float ErrorUnit_SPATIAL_ATKINSON::matrix_[] =
{
	 0,  0,  0, cx, cx,
	 0, cx, cx, cx,  0,
	 0,  0, cx,  0,  0
};

// --------------------------------------------------------------------

//
/*
class ErrorUnit_MULTIMODE : public ErrorUnit
{
	
public:

	ErrorUnit_MULTIMODE(signed short _w, signed short _h, signed short _colourdepth, signed short _fields, float _error_level)
		: ErrorUnit(_w, _h, _colourdepth, _fields, _error_level)
	{
		datasize_t pixels = _w * _h;
		for (int f = 0; f < fields_; ++f)
			fieldbuffers_[f] = new xyzerror[pixels];
		spatialbuffer_ = new xyzerror[pixels];	
	}
	
	virtual void sample(signed short _x, signed short _y, float& _ex, float& _ey, float& _ez)
	{ 
		xyzerror* peo = fieldbuffers_[field_ % fields_];
		const xyzerror& eo = peo[(_y * w_) + _x];

		xyzerror* pef = spatialbuffer_;
		const xyzerror& ef = pef[(_y * w_) + _x];
		
		// scale error from feedback
		_ex = (eo.x * field_magnitude_) + (ef.x * (1.0f - error_scale_));
		_ey = (eo.y * field_magnitude_) + (ef.y * (1.0f - error_scale_));
		_ez = (eo.z * field_magnitude_) + (ef.z * (1.0f - error_scale_));
	}
	
	virtual void sample_field(signed short _x, signed short _y, float& _ex, float& _ey, float& _ez)
	{ 
		xyzerror* peo = fieldbuffers_[field_ % fields_];
		const xyzerror& eo = peo[(_y * w_) + _x];
		
		// scale error from feedback
		_ex = (eo.x * field_magnitude_);
		_ey = (eo.y * field_magnitude_);
		_ez = (eo.z * field_magnitude_);
	}	

	virtual void sample_spatial(signed short _x, signed short _y, float& _ex, float& _ey, float& _ez)
	{ 
		xyzerror* pef = spatialbuffer_;
		const xyzerror& ef = pef[(_y * w_) + _x];
		
		// scale error from feedback
		_ex = (ef.x * (1.0f - error_scale_));
		_ey = (ef.y * (1.0f - error_scale_));
		_ez = (ef.z * (1.0f - error_scale_));
	}
		
	virtual void feedback_field(signed short _x, signed short _y, float _ex, float _ey, float _ez)
	{
		// field error
		
		xyzerror* pet = fieldbuffers_[field_ % fields_];
		xyzerror& et = pet[(_y * w_) + _x];

		et.x = _ex;
		et.y = _ey;
		et.z = _ez;			
	}	
	
	virtual void feedback_spatial(signed short _x, signed short _y, float _ex, float _ey, float _ez)
	{
		xyzerror* pet = spatialbuffer_;
		xyzerror* et_y0 = &pet[((_y + 0) * w_) + _x];
		xyzerror* et_y1 = &pet[((_y + 1) * w_) + _x];

		// handle alternating x-sweep		
		signed short xbeg = 0;
		signed short xend = (w_- 1);
		if (direction_ < 0)
		{
			xend = 0;
			xbeg = (w_- 1);
		}
		
		static const float c0 = 7.0f / 16.0f;
		static const float c1 = 1.0f / 16.0f;
		static const float c2 = 5.0f / 16.0f;
		static const float c3 = 3.0f / 16.0f;
				
		// floyd-steinberg matrix with clipping
		if (_x != xend)
		{
			xyzerror& et = et_y0[direction_];
			et.x = _ex * c0;
			et.y = _ey * c0;
			et.z = _ez * c0;		

			if (_y != (h_ - 1))
			{
				xyzerror& et = et_y1[direction_];
				et.x = _ex * c1;
				et.y = _ey * c1;
				et.z = _ez * c1;
			}
		}

		if (_y != (h_ - 1))
		{
			xyzerror& et = et_y1[0];
			et.x = _ex * c2;
			et.y = _ey * c2;
			et.z = _ez * c2;		
	
			if (_x != xbeg)
			{
				xyzerror& et = et_y1[-direction_];
				et.x = _ex * c3;
				et.y = _ey * c3;
				et.z = _ez * c3;		
			}
		}				
	}
	
	// feedback all error from this frame
	virtual void feedback(signed short _x, signed short _y, float _ex, float _ey, float _ez)
	{
		feedback_field(_x, _y, _ex, _ey, _ez);
		feedback_spatial(_x, _y, _ex, _ey, _ez);
	}		
};
*/

// --------------------------------------------------------------------

void PCSImage::init_error_unit(int _colourdepth)
{
	delete errorunit_; errorunit_ = 0;
	
	ErrorUnit* newunit = 0;	
	switch (g_allargs.errortype)
	{
	default:
	case ErrorType_OFF:
		//allocating a dummy impl' is expensive vs just leaving it unallocated, since this module has 2 stages
		//newunit = new ErrorUnit_OFF(display_width_, display_height_, _colourdepth, _error_level);
		if (g_allargs.verbose)	
			printf("error feedback disabled...\n");		
		break;
	case ErrorType_FIELD:
		newunit = new ErrorUnit_FIELD(display_width_, display_height_, _colourdepth, fields_, g_allargs.errorlevel);
		if (g_allargs.verbose)	
			printf(
				"using FIELD error feedback unit with %.2f%% feedback level, %d iterations...\n", 
				(g_allargs.errorlevel * 100.0f),
				g_allargs.erroriters
			);		
		break;
	case ErrorType_SPATIAL:
		switch (g_allargs.spatialtype)
		{
		default:
		case SpatialType_FLOYD:
			newunit = new ErrorUnit_SPATIAL_FLOYD(display_width_, display_height_, _colourdepth, fields_, g_allargs.errorlevel);
			if (g_allargs.verbose)	
				printf(
					"using SPATIAL [Floyd-Steinberg] error feedback with %.2f%% feedback level...\n",
					(g_allargs.errorlevel * 100.0f)
				);		
			break;
		case SpatialType_JAJUNI:
			newunit = new ErrorUnit_SPATIAL_JAJUNI(display_width_, display_height_, _colourdepth, fields_, g_allargs.errorlevel);
			if (g_allargs.verbose)	
				printf(
					"using SPATIAL [JaJuNi] error feedback with %.2f%% feedback level...\n",
					(g_allargs.errorlevel * 100.0f)
				);		
			break;
		case SpatialType_ZHIGANG:
			newunit = new ErrorUnit_SPATIAL_ZHIGANG(display_width_, display_height_, _colourdepth, fields_, g_allargs.errorlevel);
			if (g_allargs.verbose)	
				printf(
					"using SPATIAL [Zhigang-Fan] error feedback with %.2f%% feedback level...\n",
					(g_allargs.errorlevel * 100.0f)
				);				
			break;
		case SpatialType_SHIAU1:
			newunit = new ErrorUnit_SPATIAL_SHIAU1(display_width_, display_height_, _colourdepth, fields_, g_allargs.errorlevel);
			if (g_allargs.verbose)	
				printf(
					"using SPATIAL [Shiau-Fan #1] error feedback with %.2f%% feedback level...\n",
					(g_allargs.errorlevel * 100.0f)
				);				
			break;
		case SpatialType_SHIAU2:
			newunit = new ErrorUnit_SPATIAL_SHIAU2(display_width_, display_height_, _colourdepth, fields_, g_allargs.errorlevel);
			if (g_allargs.verbose)	
				printf(
					"using SPATIAL [Shiau-Fan #2] error feedback with %.2f%% feedback level...\n",
					(g_allargs.errorlevel * 100.0f)
				);
			break;
		case SpatialType_STUCKI:
			newunit = new ErrorUnit_SPATIAL_STUCKI(display_width_, display_height_, _colourdepth, fields_, g_allargs.errorlevel);
			if (g_allargs.verbose)	
				printf(
					"using SPATIAL [Stucki] error feedback with %.2f%% feedback level...\n",
					(g_allargs.errorlevel * 100.0f)
				);		
			break;
		case SpatialType_BURKES:
			newunit = new ErrorUnit_SPATIAL_BURKES(display_width_, display_height_, _colourdepth, fields_, g_allargs.errorlevel);
			if (g_allargs.verbose)	
				printf(
					"using SPATIAL [Burkes] error feedback with %.2f%% feedback level...\n",
					(g_allargs.errorlevel * 100.0f)
				);				
			break;
		case SpatialType_SIERRA:
			newunit = new ErrorUnit_SPATIAL_SIERRA(display_width_, display_height_, _colourdepth, fields_, g_allargs.errorlevel);
			if (g_allargs.verbose)	
				printf(
					"using SPATIAL [Sierra] error feedback with %.2f%% feedback level...\n",
					(g_allargs.errorlevel * 100.0f)
				);				
			break;
		case SpatialType_ATKINSON:
			newunit = new ErrorUnit_SPATIAL_ATKINSON(display_width_, display_height_, _colourdepth, fields_, g_allargs.errorlevel);
			if (g_allargs.verbose)	
				printf(
					"using SPATIAL [Atkinson] error feedback with %.2f%% feedback level...\n",
					(g_allargs.errorlevel * 100.0f)
				);						
			break;
		}
		
/*		
	case ErrorType_MULTIMODE:
		newunit = new ErrorUnit_MULTIMODE(display_width_, display_height_, _colourdepth, fields_, g_allargs.errorlevel);
		if (g_allargs.verbose)	
			printf(
				"using MULTIMODE error feedback unit: %.2f%% field : %.2f%% spatial ratio, %d iterations...\n", 
				(g_allargs.errorlevel * 100.0f),
				((1.0f - g_allargs.errorlevel) * 100.0f),
				g_allargs.erroriters
			);		
		break;
*/		
	}
	
	errorunit_ = newunit;
}

// --------------------------------------------------------------------
