// ---------------------------------------------------------------------------------------------------------------------
//	PhotoChrome v6.xx - Atari ST interlaced image reduction framework
//	1991-2016 Douglas Little (d.m.l)
// ---------------------------------------------------------------------------------------------------------------------
//	special thanks to:
//		evil@DHS, RA/PDX, Troed/SYNC for help testing and input on improvements
// ---------------------------------------------------------------------------------------------------------------------

// ====================================================================
//	reusable image processing stuff
// ====================================================================

#include "common.h"
#include "config.h"

#if defined(__ATARI__) || defined(__MINGW32__)
#include "FreeImage.3.15.3.h"
#else
#include "FreeImage.3.15.4.h"
#endif

#include "imagepkg.h"
#include "pcsimage.h"

extern Arguments g_allargs;

extern "C" short PCT_SYNC[];

// --------------------------------------------------------------------

static const unsigned long LCGSEED = 0xA1794529 + (1723 * 1);

InterlaceUnit::InterlaceUnit(signed short _field, signed short _colourdepth, signed short _fields)
	: LCG(LCGSEED),	// interlace unit should be field-stable!
	  field_(_field),
	  colourdepth_(_colourdepth),
	  fields_(_fields)
{
	// begin with 1/2 of the native colour LSB
	int n = (8 - _colourdepth) - 2;
	interlace_offset_ = 1 << n;
}

void InterlaceUnit::reset()
{
	LCG::reset(LCGSEED);
	onReset();
}

// --------------------------------------------------------------------

class InterlaceUnit_NONE : public InterlaceUnit
{
public:

	InterlaceUnit_NONE(signed short _field, signed short _colourdepth, signed short _fields)
		: InterlaceUnit(_field, _colourdepth, _fields)
	{
	}

	virtual float fsample(signed short _x, signed short _y)
	{
/*
;		0	-0.5
;		1	0.5
;
;		0	-1.25
;		1	-0.25
;		2	0.25
;		3	1.25
;		
;		n-1/2
*/
		
		float feval = (float)field_ - ((float)(fields_-1) / 2.0f);

		float interlace = feval / (float)nvi_crange(colourdepth_, fields_);
		
		//signed short eval = (0 + ((0 + field_) & 1)) & 1;
//		float interlace = 0.5f / (float)nvi_crange(colourdepth_, fields_);
//		if (eval)
//			interlace = -interlace;
			
		return interlace;		 
	}

};



class InterlaceUnit_HORIZONTAL : public InterlaceUnit
{
public:

	InterlaceUnit_HORIZONTAL(signed short _field, signed short _colourdepth, signed short _fields)
		: InterlaceUnit(_field, _colourdepth, _fields)
	{
	}

	virtual float fsample(signed short _x, signed short _y)
	{

//		signed short eval = (0 + ((_y + field_) & 1)) & 1;
//
//		float interlace = 0.5f / (float)nvi_crange(colourdepth_, fields_);
//
//		if (eval)
//			interlace = -interlace;


		signed short eval = (_y + field_) % fields_;

		// in 4-field mode, it's possible to interleave the lace offset order
		// if desired - but this doesn't make any sense for 2,3 fields...
		if ((fields_ == 4) && (g_allargs.multifieldadj > 0))
		{
			static const short reorder4[] =
			{
				0,3,1,2
			};

			eval = reorder4[eval];
		}

		float feval = (float)eval - ((float)(fields_-1) / 2.0f);

		float interlace = feval / (float)nvi_crange(colourdepth_, fields_);

			
		return interlace;		 
	}
};

class InterlaceUnit_VERTICAL : public InterlaceUnit
{
public:

	InterlaceUnit_VERTICAL(signed short _field, signed short _colourdepth, signed short _fields)
		: InterlaceUnit(_field, _colourdepth, _fields)
	{
	}
	
	virtual float fsample(signed short _x, signed short _y)
	{
//		signed short eval = (_x + ((0 + field_) & 1)) & 1;
//
//		float interlace = 0.5f / (float)nvi_crange(colourdepth_, fields_);
//
//		if (eval)
//			interlace = -interlace;

		short xm = g_allargs.lacesync ? PCT_SYNC[_x] : _x;

		signed short eval = (xm + field_) % fields_;

		// in 4-field mode, it's possible to interleave the lace offset order
		// if desired - but this doesn't make any sense for 2,3 fields...
		if ((fields_ == 4) && (g_allargs.multifieldadj > 0))
		{
			static const short reorder4[] =
			{
				0,3,1,2
			};

			eval = reorder4[eval];
		}

		float feval = (float)eval - ((float)(fields_-1) / 2.0f);

		float interlace = feval / (float)nvi_crange(colourdepth_, fields_);
			
		return interlace;		 
	}	
};

class InterlaceUnit_CHECKER : public InterlaceUnit
{
public:

	InterlaceUnit_CHECKER(signed short _field, signed short _colourdepth, signed short _fields)
		: InterlaceUnit(_field, _colourdepth, _fields)
	{
	}
	
	virtual float fsample(signed short _x, signed short _y)
	{
//		signed short eval = (_x + ((_y + field_) & 1)) & 1;
//
//		float interlace = 0.5f / (float)nvi_crange(colourdepth_, fields_);
//
//		if (eval)
//			interlace = -interlace;


		signed short eval;

		// CHECKER lace pattern has alternative permutations in 3,4 field modes
		// which can be controlled via 'multifieldadj', because there are many
		// different ways to cycle through N>2 states in a fixed pattern.
		// (it is always a fixed 2-state checker pattern in 2-field mode)
		if (fields_ == 3 || fields_ == 4)
		{
			static const short directions_1d[] = 
			{
				1,-1
			};
			
			static const short directions_2d[2][2] = 
			{
				{  1,-1 },
				{ -1, 1 }
			};			

			if (fields_ == 3)
			{
				short xm = g_allargs.lacesync ? PCT_SYNC[_x] : _x;

				switch (g_allargs.multifieldadj)
				{
				default:
				// cycling 3x3 tile
				case 0:
					{
						static const short offsets[3][3] = 
						{
							{ 0, 1, 2 },
							{ 2, 1, 0 },
							{ 1, 0, 2 }
						};
									
						short offset = offsets[_y % 3][xm % 3];
						
						short direction = 1;//directions_1d[_y % 2];

						eval = xsmod(offset + (direction * field_), fields_);			
					}
					break;
				// 1D countercycling 3x3 tile
				case 1:
					{
						static const short offsets[3][3] = 
						{
							{ 0, 1, 2 },
							{ 2, 1, 0 },
							{ 1, 0, 2 }
						};
									
						short offset = offsets[_y % 3][xm % 3];
						
						short direction = directions_1d[_y % 2];

						eval = xsmod(offset + (direction * field_), fields_);			
					}
					break;
				// 2D countercycling 3x3 tile
				case 2:
					{
						static const short offsets[3][3] = 
						{
							{ 0, 1, 2 },
							{ 2, 1, 0 },
							{ 1, 0, 2 }
						};
									
						short offset = offsets[_y % 3][xm % 3];
						
						short direction = directions_2d[_y % 2][xm % 2];

						eval = xsmod(offset + (direction * field_), fields_);			
					}
					break;
				// cycling 4x4 tile
				case 3:
					{
						static const short offsets[4][4] = 
						{
							{ 1, 0, 2, 0 },
							{ 0, 2, 1, 2 },
							{ 2, 1, 0, 1 },
							{ 0, 2, 1, 2 }
						};
						
						short offset = offsets[_y % 4][xm % 4];

						short direction = 1;//directions_2d[_y % 2][xm % 2];

						eval = xsmod(offset + (direction * field_), fields_);
					}
					break;
				// 1D countercycling 4x4 tile
				case 4:
					{
						static const short offsets[4][4] = 
						{
							{ 1, 0, 2, 0 },
							{ 0, 2, 1, 2 },
							{ 2, 1, 0, 1 },
							{ 0, 2, 1, 2 }
						};
						
						short offset = offsets[_y % 4][xm % 4];

						short direction = directions_1d[_y % 2];

						eval = xsmod(offset + (direction * field_), fields_);
					}
					break;
				// 2D countercycling 4x4 tile
				case 5:
					{
						static const short offsets[4][4] = 
						{
							{ 1, 0, 2, 0 },
							{ 0, 2, 1, 2 },
							{ 2, 1, 0, 1 },
							{ 0, 2, 1, 2 }
						};
						
						short offset = offsets[_y % 4][xm % 4];

						short direction = directions_2d[_y % 2][xm % 2];

						eval = xsmod(offset + (direction * field_), fields_);
					}
					break;
				// 1D countercycling columns of reordered rows
				case 6:
					{
						static const short offsets[] = 
						{
							0, 1, 2, 1, 0, 2
						};
									
						short offset = offsets[_y % 6];
						
						short direction = directions_1d[xm % 2];

						eval = xsmod(offset + (direction * field_), fields_);
					}
				}
			}
			else if (fields_ == 4)
			{
				short xm = g_allargs.lacesync ? PCT_SYNC[_x] : _x;

				switch (g_allargs.multifieldadj)
				{
				default:
				// cycling 4x4 tile
				case 0:
					{
						static const short offsets[4][4] = 
						{
							{ 0, 3, 1, 2 },
							{ 1, 2, 0, 3 },
							{ 2, 0, 3, 1 },
							{ 3, 1, 2, 0 }
						};
						
						short offset = offsets[_y % 4][xm % 4];

						short direction = 1;//directions_2d[_y % 2][xm % 2];

						eval = xsmod(offset + (direction * field_), fields_);
					}
					break;
				// 1D countercycling 4x4 tile
				case 1:
					{
						static const short offsets[4][4] = 
						{
							{ 0, 3, 1, 2 },
							{ 1, 2, 0, 3 },
							{ 2, 0, 3, 1 },
							{ 3, 1, 2, 0 }
						};
						
						short offset = offsets[_y % 4][xm % 4];

						short direction = directions_1d[_y % 2];

						eval = xsmod(offset + (direction * field_), fields_);
					}
					break;
				// 2D countercycling 4x4 tile
				// DECENT RESULT WITH LACESYNC - MINIMAL STREAKING, SOME SHIMMER
				case 2:
					{
						static const short offsets[4][4] = 
						{
							{ 0, 3, 1, 2 },
							{ 1, 2, 0, 3 },
							{ 2, 0, 3, 1 },
							{ 3, 1, 2, 0 }
						};
						
						short offset = offsets[_y % 4][xm % 4];

						short direction = directions_2d[_y % 2][xm % 2];

						eval = xsmod(offset + (direction * field_), fields_);
					}
					break;
				}
			}
			else
			{
				// fault
				assert(false);
			}
						
			// 1) directionless tile

			//if (0)
			//{			
			//	short xm = _x;
	
			//	static const short offsets[4][4] = 
			//	{
			//		{ 1, 0, 2, 0 },
			//		{ 0, 2, 1, 2 },
			//		{ 2, 1, 0, 1 },
			//		{ 0, 2, 1, 2 }
			//	};
			//				
			//	short offset = offsets[xm % 4][_y % 4];
	
			//	eval = (offset + field_) % fields_;
			//}

			// 2) directionless tile with sync
			// TODO: move to SYNC laceunit
			//if (0)
			//{			
			//	short xm = PCT_SYNC[_x];
	
			//	static const short offsets[4][4] = 
			//	{
			//		{ 1, 0, 2, 0 },
			//		{ 0, 2, 1, 2 },
			//		{ 2, 1, 0, 1 },
			//		{ 0, 2, 1, 2 }
			//	};
			//				
			//	short offset = offsets[xm % 4][_y % 4];
	
			//	eval = (offset + field_) % fields_;
			//}






			// 1b) directionless countercycling tile
/*			if (0)
			{
				short xm = _x;

				short offset;
				
				if (fields_ == 3)
				{
					
					static const short offsets[4][4] = 
					{
						{ 1, 0, 2, 0 },
						{ 0, 2, 1, 2 },
						{ 2, 1, 0, 1 },
						{ 0, 2, 1, 2 }
					};
					
					offset = offsets[_y % 4][xm % 4];
											
				}
				else if (fields_ == 4)
				{

					static const short offsets[4][4] = 
					{
						{ 0, 3, 1, 2 },
						{ 1, 2, 0, 3 },
						{ 2, 0, 3, 1 },
						{ 3, 1, 2, 0 }
					};
					
					offset = offsets[_y % 4][xm % 4];

				}
				else
				{
					// fault
					assert(false);
				}


				short direction = 1;//directions_2d[_y % 2][xm % 2];

				eval = xsmod(offset + (direction * field_), fields_);
			}	
*/

			
			// 3 countercycling tile #1
/*			if (0)
			{			
				short xm = _x;//PCT_SYNC[_x];
	
				static const short offsets[2][4] = 
				{
					{ 0, 1, 2, 1 },
					{ 2, 1, 0, 1 }
				};
							
				short offset = offsets[_y % 2][xm % 4];
				
				short direction = directions_1d[_y & 1];

				eval = xsmod(offset + (direction * field_), fields_);			
	
			}			
*/

/*			// 3 countercycling tile #2
			if (0)
			{			
				short xm = _x;//PCT_SYNC[_x];
	
				static const short offsets[2][4] = 
				{
					{ 0, 1, 2, 1 },
					{ 1, 2, 1, 0 }
				};
							
				short offset = offsets[_y % 2][xm % 4];
				
				short direction = directions_1d[_y % 2];

				eval = xsmod(offset + (direction * field_), fields_);			
	
			}	
*/

/*			// cycling 3x3 tile
			// GOOD RESULT! but causes streaking
			if (1)
			{			
				short xm = _x;//PCT_SYNC[_x];
	
				static const short offsets[3][3] = 
				{
					{ 0, 1, 2 },
					{ 2, 1, 0 },
					{ 1, 0, 2 }
				};
							
				short offset = offsets[_y % 3][xm % 3];
				
				short direction = 1;//directions_1d[_y % 2];

				eval = xsmod(offset + (direction * field_), fields_);			
			}
*/
			// cycling 3x3 tile with sync
			// BAD - FLICKERY, VERTICAL BANDS
			// TODO: move to SYNC laceunit
/*			if (0)
			{			
				short xm = PCT_SYNC[_x];
	
				static const short offsets[3][3] = 
				{
					{ 0, 1, 2 },
					{ 2, 1, 0 },
					{ 1, 0, 2 }
				};
							
				short offset = offsets[_y % 3][xm % 3];
				
				short direction = 1;//directions_1d[_y % 2];

				eval = xsmod(offset + (direction * field_), fields_);			
	
			}
*/			
			// PRNG
			// REASONABLE RESULT, CAUSES STREAKING AND SOME FUZZ/FLICKER
			// TODO: move to PRNG laceunit
			//if (0)
			//{			
			//	short xm = _x;
	
			//	unsigned short offset = lcg_sample() >> 6;
			//			
			//	short direction = (lcg_sample() >> 6) & 1 ? -1 : 1;//	1;//directions_2d[_y % 2][xm % 2];

			//	eval = xsmod(offset + (direction * field_), fields_);			
			//}

			// PRNG with sync
			// FUZZY CYCLING, NOT GOOD
			// TODO: move to PRNG laceunit
			//if (0)
			//{			
			//	short xm = PCT_SYNC[_x];
			//	
			//	int xp = (((xm + 19743) * 21927)) >> 6;
			//	int yp = (((_y + 11529) * 18475)) >> 6;
	
			//	unsigned short offset = (xp ^ yp) % 3;
			//			
			//	short direction = ((((((xp ^ yp)) * 11529) + 31475) >> 6) & 1) ? -1 : 1;//	1;//directions_2d[_y % 2][xm % 2];

			//	eval = xsmod(offset + (direction * field_), fields_);			
			//}

			// directionless scanline
			// DECENT RESULT - MINIMAL STREAKING, SOME SHIMMER
/*			if (0)
			{			
				short xm = PCT_SYNC[_x];
	
				static const short offsets[] = 
				{
					0, 1, 2, 1, 0, 2
				};
							
				short offset = offsets[_y % 6];
				
				short direction = directions_1d[xm % 2];

				eval = xsmod(offset + (direction * field_), fields_);			
	
			}	*/		
			
/*
			static const short directions[] = 
			{
				1,-1
			};
						
			short direction = directions[_y & 1];			
			direction = 1;


			static const short offsets[2][3] = 
			{
				{ 0, 1, 2 },
				{ 1, 2, 0 }
			};
									
			short offset = offsets[_y % 2][xm % 3];

			eval = xsmod(direction * (field_+offset), fields_);
*/

/*

			static const short directions[2][2] = 
			{
				{  1,-1 },
				{ -1, 1 }
			};
						
			short direction = directions[xm & 1][_y & 1];			
*/

/*			
			static const short offsets[4][4] = 
			{
//				{ 0, 2, 0 },
//				{ 2, 0, 1 },
//				{ 1, 1, 2 }
				{ 2, 0, 1, 0 },
				{ 0, 1, 2, 1 },
				{ 2, 0, 1, 0 },
				{ 1, 2, 0, 2 }
			};
			
			short offset = offsets[_x % 4][_y % 4];
*/

/*
			static const short offsets[5][5] = 
			{
				{ 1, 0, 2, 0, 1 },
				{ 0, 2, 1, 2, 0 },
				{ 2, 1, 0, 1, 2 },
				{ 0, 2, 1, 2, 0 },
				{ 1, 0, 2, 0, 1 }
			};
*/


					

//			eval = xsmod(((direction * field_)+offset), fields_);
			
//			eval = xsmod(
//					ordering[_x % fields_][_y % fields_] + (field_ * cycling[_x % fields_][_y % fields_]), 
//					fields_
//				);
		}
		else
		{
			short xm = g_allargs.lacesync ? PCT_SYNC[_x] : _x;

			// 2-field mode can't be influenced
			eval = (xm + ((_y + field_) & 1)) & 1;
		}

		float feval = (float)eval - ((float)(fields_-1) / 2.0f);

		float interlace = feval / (float)nvi_crange(colourdepth_, fields_);
						
		return interlace;		 
	}		
};

class InterlaceUnit_PRNG : public InterlaceUnit
{
public:

	InterlaceUnit_PRNG(signed short _field, signed short _colourdepth, signed short _fields)
		: InterlaceUnit(_field, _colourdepth, _fields)
	{
	}

	virtual float fsample(signed short _x, signed short _y)
	{
//		unsigned short eval = ((lcg_sample() >> 6) + field_) & 1;
//
//		float interlace = 0.5f / (float)nvi_crange(colourdepth_, fields_);
//
//		if (eval)
//			interlace = -interlace;

		unsigned short eval;

		if ((fields_ < 3) || (g_allargs.multifieldadj == 0))
		{
			eval = ((lcg_sample() >> 6) + field_) % fields_;
		}
		else
		{
//			short xm = _x;

			unsigned short offset = lcg_sample() >> 6;
					
			short direction = (lcg_sample() >> 6) & 1 ? -1 : 1;//	1;//directions_2d[_y % 2][xm % 2];

			eval = xsmod(offset + (direction * field_), fields_);			
		}


		float feval = (float)eval - ((float)(fields_-1) / 2.0f);

		float interlace = feval / (float)nvi_crange(colourdepth_, fields_);
			
		return interlace;		 
	}	
};
	
// --------------------------------------------------------------------

void PCSImage::init_interlace_unit(int _field)
{
	delete laceunit_; laceunit_ = 0;
	
	InterlaceUnit* newunit = 0;	
	switch (g_allargs.lacetype)
	{
	case LaceType_OFF:
		break;

	default:
//	case LaceType_SYNCHRONIZED:	// todo
	case LaceType_FIELD:
		newunit = new InterlaceUnit_NONE(_field, colourdepth_, fields_);
		break;
	case LaceType_HORIZONTAL:
		newunit = new InterlaceUnit_HORIZONTAL(_field, colourdepth_, fields_);
		break;
	case LaceType_VERTICAL:
		newunit = new InterlaceUnit_VERTICAL(_field, colourdepth_, fields_);
		break;
	case LaceType_CHECKER:
		newunit = new InterlaceUnit_CHECKER(_field, colourdepth_, fields_);
		break;
	case LaceType_PRNG:
		newunit = new InterlaceUnit_PRNG(_field, colourdepth_, fields_);
		break;
	}
	
	laceunit_ = newunit;	
}

// --------------------------------------------------------------------
