// ---------------------------------------------------------------------------------------------------------------------
//	PhotoChrome v6.xx - Atari ST interlaced image reduction framework
//	1991-2016 Douglas Little (d.m.l)
// ---------------------------------------------------------------------------------------------------------------------
//	special thanks to:
//		evil@DHS, RA/PDX, Troed/SYNC for help testing and input on improvements
// ---------------------------------------------------------------------------------------------------------------------

// ====================================================================
// read or write PCS formatted files
// ====================================================================

#include "common.h"
#include "config.h"

#if defined(__ATARI__) || defined(__MINGW32__)
#include "FreeImage.3.15.3.h"
#else
#include "FreeImage.3.15.4.h"
#endif

#include "imagepkg.h"
#include "pcsimage.h"

extern Arguments g_allargs;

// --------------------------------------------------------------------

bool PCSImage::generate_pcs_file(const char* const _name)
{
	bool success = false;

	// in 'shared bitmap' mode we have one image but two palettes generated from a high
	// bitdepth palette. we want to export both palettes so we force the export module to
	// think this is a dual-field image. it will notice the second bitmap is invalid and
	// export the shared bitmap twice.
//	bool flick = (fields_ > 1) || 
//				 (g_allargs.fieldmode == FieldMode_SharedBitmap);
	
	// note: this allocates memory for / assigns the pcs file buffer
	if (g_allargs.verbose)
		printf("formatting data for output [%s]\n", _name);						

//	datasize_t towrite = create_pcs_format(/*flick=*/flick, /*ste=*/true, /*hz60=*/false);
	datasize_t towrite = 0;

	if (g_allargs.storage == StorageFmt_RAW_DISPLAY || 
		g_allargs.storage == StorageFmt_RAW_ILEAVED_DISPLAY)
	{
		towrite = create_pcr_format
		(
			/*bitmapfields=*/fields_, 
			/*sharedbitmap=*/(g_allargs.fieldmode == FieldMode_SharedBitmap), 
			/*interleaved=*/(g_allargs.storage == StorageFmt_RAW_ILEAVED_DISPLAY) && (fields_ == 2)
		);
	}
	else
	{
		towrite = create_pcs_format
		(
			/*bitmapfields=*/fields_, 
			/*sharedbitmap=*/(g_allargs.fieldmode == FieldMode_SharedBitmap), 
			/*ste=*/(g_allargs.colourdepth == 4) || (g_allargs.colourdepth == 5), 
			/*hz60=*/false
		);
	}

	if (!g_allargs.quiet)
		printf("creating output file [%s]\n", _name);
								
	if (towrite > 0)
	{
		FILE *hfile = fopen(_name, "wb");
		
		if (hfile != 0)
		{
			if (g_allargs.verbose)	
				printf("writing output file [%s]\n", _name);						
			datasize_t written = fwrite(pcs_filebuf_, 1, pcs_filesize_, hfile);
			if (written == towrite)
			{
				success = true;
				//printf("written PCS file [%s]\n", _name);						
			}
			else
			{
				printf("error writing output file! check disk space... [%s]\n", _name);						
			}
			fclose(hfile);
		}
		else
		{
			printf("error creating output file! [%s]\n", _name);						
		}
	}
	else
	{
		printf("error formatting output file!\n");						
	}

	// release pcs file buffer
	delete[] pcs_filebuf_; pcs_filebuf_ = 0; pcs_filesize_ = 0;

//	printf("completed writing!\n");						
		
	return success;
}

//------------------------------------------------------------------------*
// header is a bit nasty, being a bag of endian-sensitive words and bits

enum PCS_format_flags
{
	mPCS_format_multiframe = 		binary( 10000011 ),		// image has more than one display frame
		oPCS_format_multiframe = 	0,

	mPCS_format_extensions = 		binary( 01111100 ),		// mask for all extension flags (00000, 11111, 10000 codes are reserved ...for compatibility & extension)
		oPCS_format_extensions = 	2,
	mPCS_format_type =				binary( 00011100 ),		// 3-bit code indicates format type
		oPCS_format_type =			2,
	mPCS_format_fields = 			binary( 01100000 ),		// 2-bits encode unique field count
		oPCS_format_fields = 		5,
		
	mPCS_format_xor_palette = 		binary( 00000010 ),		// ((this_bit ^ multiframe_bit) != 0) = palette >=1 stored as deltas
		oPCS_format_xor_palette = 	1,
	mPCS_format_xor_bitmap = 		binary( 00000001 ), 	// ((this_bit ^ multiframe_bit) != 0) = bitmap >=1 stored as deltas
		oPCS_format_xor_bitmap = 	0,
};

enum PCS_display_flags
{
	mPCS_display_colourbits = 		binary( 11000000 ),		// 11 = 4bit(STE), 00 = 3bit(STF), 01,10 = reserved
		oPCS_display_colourbits = 	6,
//	mPCS_display_devicecode = 		binary( 00111100 ),		// 4bit device code (0=backcompat/STF, 15=backcompat/STE)
//		oPCS_display_devicecode = 	2,
//	mPCS_display_blitter = 			binary( 00000010 ),		// 1 = mode *requires* blitter
//		oPCS_display_blitter = 		1,
	mPCS_display_freqcode = 		binary( 00000001 ),		// 1 = mode *requires* 50hz, otherwise *requires* 60hz
		oPCS_display_freqcode = 	0,
};

struct PCS_header
{
	unsigned short PCS_width;		// warning - endian-sensitive on m68k<->x86!
	unsigned short PCS_height;		//

	// warning - GCC 4.6.3 on m68k/mint does not seem to allow structs of 1 byte... they get padded
	// so we express the bitfields contiguously. 

	// storage flags	
	unsigned char PCS_format_flags;

	// hardware flags
	unsigned char PCS_display_flags;
		
} __attribute__((packed)); 

//------------------------------------------------------------------------*
// this is all done for compatibility with existing PCS, otherwise I would
// use something better. horizontal RLE on words = pretty useless.

datasize_t PCSImage::create_pcs_format(int _bitmapfields, bool _sharedbitmap, bool _ste_palette, bool _60hz)
{
	bool success = true;

	int exportfields = _sharedbitmap ? 2 : _bitmapfields;
	
	bool flick = exportfields > 1;
	
	//------------------------------------------------------------------------*
	// some constants
	
	datasize_t framebuffer_bytes = framebuffer_words_ * sizeof(word_t);

	datasize_t palettefile_skipwords = palette_skip_ * g_allargs.colours;

	datasize_t palettefile_words = palettebuffer_words_ - palettefile_skipwords;
	datasize_t palettefile_bytes = palettefile_words * sizeof(word_t);

	//------------------------------------------------------------------------*
	// allocate space for output file format (assume worst case for compression)

//	pcs_filebuf_ = new byte_t[((framebuffer_size * _bitmapfields) + (palettefile_size * (_sharedbitmap ? 2 : _bitmapfields))) * sizeof(word_t)];

	//printf("exportfields = %d\n", exportfields);
	//printf("framebuffer_words_ = %d\n", framebuffer_words_);
	//printf("palettebuffer_words_ = %d\n", palettebuffer_words_);

	pcs_filebuf_ = 
		new byte_t[sizeof(PCS_header) +
			32000 + // space for compression overflow
			(
				(framebuffer_words_ * exportfields) + 
				(palettebuffer_words_ * exportfields)
			) * sizeof(word_t)
		];
	if (!pcs_filebuf_) { printf("error - could not allocate memory for file buffer!\n"); getchar(); exit(1); }

	word_t* bitmap_delta_ = new word_t[framebuffer_words_ + 4];
	if (!bitmap_delta_) { printf("error - could not allocate memory for bitmap delta!\n"); getchar(); exit(1); }
	word_t* palette_delta_ = new word_t[palettebuffer_words_ + 4];
	if (!palette_delta_) { printf("error - could not allocate memory for palette delta!\n"); getchar(); exit(1); }
	
	//------------------------------------------------------------------------*
	// create header area
	
	PCS_header* pheader = (PCS_header*)pcs_filebuf_;
	
	write_be_word(&(pheader->PCS_width), display_width_);
	write_be_word(&(pheader->PCS_height), display_height_);

	// bitfield: *invert* bit [0],[1] for bitmap, palette deltas
	{
		pheader->PCS_format_flags = 0;
		pheader->PCS_format_flags |= flick ? mPCS_format_multiframe : 0;
		pheader->PCS_format_flags |= (g_allargs.typeident << oPCS_format_type) & mPCS_format_type;
		// in sharedbitmap mode, field count will be 1, but multifield flag will be set
		pheader->PCS_format_flags |= ((_bitmapfields-1) << oPCS_format_fields) & mPCS_format_fields;
	}

	// bitfield: *set* bit [0] for frequency=50hz
	{
		pheader->PCS_display_flags = 0;
		pheader->PCS_display_flags |= _ste_palette ? mPCS_display_colourbits : 0;
		pheader->PCS_display_flags |= _60hz ? 0 : mPCS_display_freqcode;
	}

	//------------------------------------------------------------------------*
	// create data area

	byte_t* pdatastart = pcs_filebuf_ + sizeof(PCS_header);

	// if we're operating in RAW mode, don't bother emitting the header, just emit the data
	if (g_allargs.storage == StorageFmt_RAW_DISPLAY)
		pdatastart = pcs_filebuf_;
	
	//------------------------------------------------------------------------*
	// compress data
	
	byte_t* pdatarun = pdatastart;
	datasize_t compressed_size = 0;
	datasize_t compressed_size_delta = 0;
	datasize_t stored_size = 0;
	
	//------------------------------------------------------------------------*
	// compress frame #1

	// image first
	if (success)
	{			
		switch (g_allargs.storage)
		{			
		case StorageFmt_PPRLE:
		case StorageFmt_UNCOMPRESSED:
			{
				if (g_allargs.verbose)	
					printf("compressing bitmap 1...\n");
				compressed_size = compress_bitmap(bitmaps_[0], pdatarun, verify_);
				pdatarun += compressed_size;
				if (compressed_size == 0)
					success - false;
			}
			break;
		case StorageFmt_DIRECT_DISPLAY:
//		case StorageFmt_RAW_DISPLAY:

			// image first
			stored_size = (framebuffer_words_ << 1);
			copy_be_words((word_t*)pdatarun, bitmaps_[0], framebuffer_words_);
			pdatarun += stored_size;

//			compressed_size = (framebuffer_words_ << 1);
//			copy_be_words((word_t*)pdatarun, bitmaps_[0], framebuffer_words_);
//			pdatarun += compressed_size;
			break;
		default:
			break;
		}
	}
	
	// palette second
	if (success)
	{						
		switch (g_allargs.storage)
		{			
		case StorageFmt_PPRLE:
		case StorageFmt_UNCOMPRESSED:
			{
				if (g_allargs.verbose)	
					printf("compressing palette 1...\n");						
				compressed_size = compress_palette(palettes_[0]+palettefile_skipwords, pdatarun, palettefile_words, verify_);
				pdatarun += compressed_size;
				if (compressed_size == 0)
					success - false;
			}
			break;
		case StorageFmt_DIRECT_DISPLAY:
//		case StorageFmt_RAW_DISPLAY:

			stored_size = (palettefile_words << 1);
			copy_be_words((word_t*)pdatarun, palettes_[0]+palettefile_skipwords, palettefile_words);
			pdatarun += stored_size;	

//			compressed_size = (palettefile_words << 1);
//			copy_be_words((word_t*)pdatarun, palettes_[0]+palettefile_skipwords, palettefile_words);
//			pdatarun += compressed_size;
			break;
		default:
			break;
		}
	}
	
	//------------------------------------------------------------------------*
	// compress frame #2, using deltas if result is smaller

	if (flick)
	{
		for (int currentfield = 1; currentfield < exportfields; ++currentfield)
		{
		
		//------------------------------------------------------------------------*
		// image first
		
		// for shared-bitmap images, only bitmap #0 is valid so we need to refer to bitmap #0 
		// both times to retain format compatibility - the 2nd bitmap ends up being redundant
		// but will be delta-compressed out of the file. for RAW formats, a duplicate bitmap
		// will remain in the file and should be ignored by chosen MUX software.
		
		word_t* bitmap1 = bitmaps_[currentfield];
		if (!bitmap1)
			bitmap1 = bitmaps_[0];	// deltas against first bitmap will collapse away
		
		switch (g_allargs.storage)
		{			
		case StorageFmt_PPRLE:
		case StorageFmt_UNCOMPRESSED:
			{		
				if (success)
				{						
					// try direct compression
					if (g_allargs.verbose)	
						printf("compressing bitmap 2...\n");						
					compressed_size = compress_bitmap(bitmap1, pdatarun, verify_);
					if (compressed_size == 0)
						success = false;
				}
		
				// try compressing only deltas
				if (success)
				{						
					if (g_allargs.verbose)	
						printf("compressing deltas...\n");						
					make_delta(bitmaps_[currentfield-1], bitmap1, bitmap_delta_, framebuffer_words_);
					compressed_size_delta = compress_bitmap(bitmap_delta_, pdatarun + compressed_size, verify_);
					if (compressed_size_delta == 0)
					{
						success = false;
					}
					else
					{
						// if delta version is smaller, adopt that instead
						if (compressed_size_delta < compressed_size)
						{
							if (g_allargs.verbose)	
								printf("adopting deltas...\n");						
							memcpy(pdatarun, (pdatarun + compressed_size), compressed_size_delta);
							compressed_size = compressed_size_delta;
							//pheader->PCS_format_xor_bitmap = !pheader->PCS_format_xor_bitmap;
							pheader->PCS_format_flags ^= mPCS_format_xor_bitmap;
						}
						pdatarun += compressed_size;
					}
				}
			}
			break;
		case StorageFmt_DIRECT_DISPLAY:
//		case StorageFmt_RAW_DISPLAY:
			stored_size = (framebuffer_words_ << 1);
			copy_be_words((word_t*)pdatarun, bitmap1, framebuffer_words_);
			pdatarun += stored_size;
//			compressed_size = (framebuffer_words_ << 1);
//			copy_be_words((word_t*)pdatarun, bitmap1, framebuffer_words_);
//			pdatarun += compressed_size;			
			break;
		default:
			break;
		}
				
		//------------------------------------------------------------------------*
		// palette second
	
		switch (g_allargs.storage)
		{			
		case StorageFmt_PPRLE:
		case StorageFmt_UNCOMPRESSED:
			{				
				// try direct compression
				if (success)
				{	
					if (g_allargs.verbose)	
						printf("compressing palette 2...\n");						
					compressed_size = compress_palette(palettes_[currentfield]+palettefile_skipwords, pdatarun, palettefile_words);
					if (compressed_size == 0)
						success = false;
				}
		
				// try compressing only deltas
				if (success)
				{			
					if (g_allargs.verbose)	
						printf("compressing deltas...\n");						
					make_delta(palettes_[currentfield-1]+palettefile_skipwords, palettes_[currentfield]+palettefile_skipwords, palette_delta_, palettefile_words);
					compressed_size_delta = compress_palette(palette_delta_, pdatarun + compressed_size, palettefile_words);
					if (compressed_size_delta == 0)
					{
						success = false;
					}
					else
					{		
						// if delta version is smaller, adopt that instead
						if (compressed_size_delta < compressed_size)
						{
							if (g_allargs.verbose)	
								printf("adopting deltas...\n");						
							memcpy(pdatarun, (pdatarun + compressed_size), compressed_size_delta);
							compressed_size = compressed_size_delta;
							//pheader->PCS_format_xor_palette = !pheader->PCS_format_xor_palette;
							pheader->PCS_format_flags ^= mPCS_format_xor_palette;
						}
						pdatarun += compressed_size;
					}
				}
			}
			break;
		case StorageFmt_DIRECT_DISPLAY:
//		case StorageFmt_RAW_DISPLAY:

			stored_size = (palettefile_words << 1);
			copy_be_words((word_t*)pdatarun, palettes_[currentfield]+palettefile_skipwords, palettefile_words);
			pdatarun += stored_size;	

			//compressed_size = (palettefile_words << 1);
			//copy_be_words((word_t*)pdatarun, palettes_[currentfield]+palettefile_skipwords, palettefile_words);
			//pdatarun += compressed_size;
			break;
		default:
			break;
		}			
		}
	}						

	//printf("done - press any key...\n"); getchar();	

	//------------------------------------------------------------------------*
	// find total size
	
	pcs_filesize_ = pdatarun - pcs_filebuf_;

	//------------------------------------------------------------------------*
	// done - clean up claims
	
	delete[] bitmap_delta_;
	delete[] palette_delta_;

	//printf("done2 - press any key...\n"); getchar();	
		
	//------------------------------------------------------------------------*
	// return size or error=0

	return success ? pcs_filesize_ : 0;	
}




datasize_t PCSImage::create_pcr_format(int _bitmapfields, bool _sharedbitmap, bool _interleaved)
{
	bool success = true;

	int exportfields = _sharedbitmap ? 2 : _bitmapfields;
	
	bool flick = exportfields > 1;
	
	//------------------------------------------------------------------------*
	// some constants

	datasize_t bitmapfile_words = _interleaved ? (framebuffer_words_ >> 1) : framebuffer_words_;
	
//	datasize_t framebuffer_size = (_interleaved ? (framebuffer_words_ >> 1) framebuffer_words_) * sizeof(word_t);

	datasize_t palettefile_words = _interleaved ? (palettebuffer_words_ >> 1) : palettebuffer_words_;
//	datasize_t palettefile_size = palettefile_words * sizeof(word_t);

	datasize_t bitmap_linewords = ((display_width_ + 15) >> 4) * display_planes_;
	datasize_t bitmap_linebytes = bitmap_linewords * sizeof(word_t);

	datasize_t palette_linewords = palettebuffer_stride_;
	datasize_t palette_linebytes = palette_linewords * sizeof(word_t);
	
	//------------------------------------------------------------------------*
	// allocate space for output file format (assume worst case for compression)

	pcs_filebuf_ = 
		new byte_t[
			(
				(bitmapfile_words * exportfields) + 
				(palettefile_words * exportfields)
			) * sizeof(word_t)
		];

	if (!pcs_filebuf_) { printf("error - could not allocate memory for file buffer!\n"); getchar(); exit(1); }


	//------------------------------------------------------------------------*
	// create data area

	byte_t* pdatastart = pcs_filebuf_;	
	byte_t* pdatarun = pdatastart;
	
	//------------------------------------------------------------------------*
	// store image, interleaving fields as necessary

	datasize_t stored_size = 0;

	// image first

	if (_interleaved)
	{
		int y;
		for (y = 0; y < display_height_; y += 2)
		{
			copy_be_words((word_t*)pdatarun, bitmaps_[0] + (bitmap_linewords * (y+0)), bitmap_linewords);
			pdatarun += bitmap_linebytes;
			copy_be_words((word_t*)pdatarun, bitmaps_[1] + (bitmap_linewords * (y+1)), bitmap_linewords);
			pdatarun += bitmap_linebytes;
		}

		for (y = 0; y < display_height_; y += 2)
		{
			copy_be_words((word_t*)pdatarun, palettes_[0] + (palette_linewords * (y+0)), palette_linewords);
			pdatarun += palette_linebytes;
			copy_be_words((word_t*)pdatarun, palettes_[1] + (palette_linewords * (y+1)), palette_linewords);
			pdatarun += palette_linebytes;
		}
	}
	else
	{
		stored_size = (bitmapfile_words << 1);
		copy_be_words((word_t*)pdatarun, bitmaps_[0], bitmapfile_words);
		pdatarun += stored_size;

		// palette second
		stored_size = (palettefile_words << 1);
		copy_be_words((word_t*)pdatarun, palettes_[0], palettefile_words);
		pdatarun += stored_size;	
	}

	// for non-interleaved, multi-field images, emit remaining fields
	// note: in case of shared bitmap, emit the same bitmap[0] multiple times for easy testing

	if (flick && !_interleaved)
	{
		for (int currentfield = 1; currentfield < exportfields; ++currentfield)
		{
			word_t* bitmap = bitmaps_[currentfield];
			if (!bitmap)
				bitmap = bitmaps_[0];

			// image first
			stored_size = (bitmapfile_words << 1);
			copy_be_words((word_t*)pdatarun, bitmap, bitmapfile_words);
			pdatarun += stored_size;

			// palette second
			stored_size = (palettefile_words << 1);
			copy_be_words((word_t*)pdatarun, palettes_[currentfield], palettefile_words);
			pdatarun += stored_size;	
		}
	}

	//------------------------------------------------------------------------*
	// find total size
	
	pcs_filesize_ = pdatarun - pcs_filebuf_;

	//printf("done2 - press any key...\n"); getchar();	
		
	//------------------------------------------------------------------------*
	// return size or error=0

	return success ? pcs_filesize_ : 0;	
}


//------------------------------------------------------------------------*
// generate word-delta buffer from two comparable sources

void PCSImage::make_delta(word_t* _src1, word_t* _src2, word_t* _delta, datasize_t _words)
{
	for (datasize_t w = _words - 1; w >= 0; --w)
		*(_delta++) = *(_src1++) ^ *(_src2++);
}

//------------------------------------------------------------------------*
// make big-endian, plane-by-plane representation

void PCSImage::reformat_bitmap(word_t* _src, word_t* _dst)
{
	int xc = display_width_ >> 4;
	int yc = display_height_;
	int pc = display_planes_;
	int linesize = xc * pc;
	
	for (int p = 0; p < pc; ++p)
		for (int y = 0; y < yc; ++y)
			for (int x = 0; x < xc; ++x)
				write_be_word(_dst++, _src[p + (x << 2) + (y * linesize)]);	
}

//------------------------------------------------------------------------*

// compress bitmap frame in terms of bytes
datasize_t PCSImage::compress_bitmap(word_t* _src, byte_t* _dst, bool _verify)
{		
	// convert bitmap into bigendian-word, plane-by-plane representation
	
	word_t* reformatted_src = new word_t[framebuffer_words_ + 4];
	if (!reformatted_src) { printf("error - could not allocate memory for reformat buffer!\n"); getchar(); exit(1); }
	
	//printf("reformatting planes...\n");						
	reformat_bitmap(_src, reformatted_src);

	// compress plane-by-plane representation

	static const int tmpbuf_max = 32000;	
	signed long src_count = framebuffer_words_ << 1;
	signed long tmpbuf_count = 0;
	signed long tmprun_count = 1;
	signed short ctlcode_count = 0;
	
	byte_t* dst8ptr = _dst;
	byte_t* src8ptr = (byte_t*)reformatted_src;
	byte_t* tmp8baseptr = new byte_t[tmpbuf_max+1000];
	if (!tmp8baseptr) { printf("error - could not allocate memory for compression buffer!\n"); getchar(); exit(1); }
	byte_t* tmp8runptr = tmp8baseptr;
	word_t* ctlcount16ptr = (word_t*)dst8ptr;

	// arrange a run-terminator at the end of the buffer to force the RLE compressor
	// to break, re-evaluate it's current position and stop.
//	byte_t* pterm = src8ptr + src_count;
//	pterm[0] = pterm[-1] ^ 0xFF;
//	pterm[1] = pterm[-1];
	
	// skip control code count
	dst8ptr += 2;

	// prefetch
	word_t prev_data = 0;
	word_t curr_data = read_byte(src8ptr++); src_count--;
	
compress_loop:	
	
	// rotate & fill
	prev_data = curr_data;
	write_byte(tmp8runptr++, curr_data); tmpbuf_count++;
	curr_data = read_byte(src8ptr++); src_count--;
	
	// buffer contains everything except curr_data
	
	// interrupt if source or temp buffer exhausted
	if ((tmpbuf_count >= tmpbuf_max) || (src_count < 0))
 		goto store;
 		
	if (1 && (curr_data == prev_data))
	{
		// continue [run]
		tmprun_count++;
	}
	else
	{
		// run break
		// store entire block & run, but not break value
		if (tmprun_count >= 4)
			goto store;
			
		// reset run, continue fill
		tmprun_count = 1;	
	}
	goto compress_loop;
		
	//------------------------------------------------------------------------*
	// store an control pattern (>=0 literal + >=0 run)
	// at this point we may have both buffered and run data waiting, with
	// the run being the latest part of the buffer, if present

store:

 	//printf("src_count=%d\n", src_count);

	// if we interrupted a run before it matures, or if run==1 (special meaning)
	// we need to absorb the run into a block
	if (tmprun_count < 4)
		tmprun_count = 0;
			
	// adjust to find waiting [literal block] length, versus [run] length
	tmpbuf_count -= tmprun_count;

	//------------------------------------------------------------------------*
	// store [literal block] first, if one is waiting

	if (tmpbuf_count > 0)
	{
		ctlcode_count++;
		
		if (tmpbuf_count > 127)
		{
			// long [literal block] : 1 ctrl
			write_byte(dst8ptr++, 1);
			write_be_word(dst8ptr, tmpbuf_count); dst8ptr += 2; 
			//printf("<lb[%d]\n", tmpbuf_count);
		}
		else
		{
			// short [literal block] : -ve ctrl
			// (careful dealing with signed values, converting from asm)
			write_byte(dst8ptr++, (byte_t)(-tmpbuf_count));
			//printf("<sb[%d]\n", tmpbuf_count);
		}				 
 	 	
		// copy run	(words)
		tmp8runptr = tmp8baseptr;
		for (signed short c = tmpbuf_count - 1; c >= 0; --c)
		{
			write_byte(dst8ptr++, read_byte(tmp8runptr++)); 
		}		
		
	 	//printf("stored block!\n");		
	}
	// reset [literal block] size
	tmp8runptr = tmp8baseptr;
	tmpbuf_count = 0;
	
	//------------------------------------------------------------------------*
	// store [run] last
	if (tmprun_count > 1)
	{
		ctlcode_count++;
		
		if (tmprun_count > 127)
		{
			// long [run] : 0 ctrl
			write_byte(dst8ptr++, 0);
			write_be_word(dst8ptr, tmprun_count); dst8ptr += 2;
			//printf("<lr[%d]\n", tmprun_count);
		}
		else
		{
			// short [run] : +ve ctrl
			write_byte(dst8ptr++, tmprun_count);
			//printf("<sr[%d]\n", tmprun_count);
		}				 

		// run value
		write_byte(dst8ptr++, prev_data);		

	 	//printf("stored run!\n");		
	}
	// reset [run] size
	tmprun_count = 1;

	// continue compressor

	if (src_count >= 0)
		goto compress_loop;

	//------------------------------------------------------------------------*
	// compression complete

done:
	 
	// store control code count
	write_be_word(ctlcount16ptr, ctlcode_count);

	// free temporary space
	delete[] tmp8baseptr;

	datasize_t compressed_size = dst8ptr - _dst;

	if (_verify)
	{	
		if (g_allargs.verbose)
			printf("verifying...\n"); 

		datasize_t decompressed_size = decompress_bitmap(_dst, reformatted_src, /*verify=*/true);
	
		if ((int)decompressed_size < 0)
		{
			printf("error - decompression verify failed checksum: sizein=%d, sizeout=%d\n", 
			(int)framebuffer_words_<<1, (int)(~decompressed_size));
			// indicate error in verify mode
			compressed_size = 0;
		}
		else if ((int)decompressed_size != framebuffer_words_<<1)
		{
			printf("error - decompression verify failed bounds: sizein=%d, sizeout=%d\n", 
			(int)framebuffer_words_<<1, (int)decompressed_size);
			// indicate error in verify mode
			compressed_size = 0;
		}
	}

	// free reformatted source
	delete[] reformatted_src;
	
	return (compressed_size);
}

datasize_t PCSImage::decompress_bitmap(byte_t* _src, word_t* _dst, bool _verify)
{			
	signed char ctlcode_8 = 0;
	signed short ctlcode_16 = 0;
	byte_t data_8 = 0;
	signed short ctlcode_count = 0;
	signed short ctlcode_consumed = 0;
	byte_t chk = 0;

	datasize_t decompressed_size = 0;
	datasize_t predict_size = 0;
	datasize_t expected_size = framebuffer_words_ << 1;	

	byte_t* src8ptr = _src;
	byte_t* dst8ptr = (byte_t*)_dst;

	// begin
		
	ctlcode_count = read_be_word(src8ptr); src8ptr += 2;
	
decompress_loop:

	if (ctlcode_count == 0)
		goto finish;	

	ctlcode_8 = read_byte(src8ptr++); ctlcode_count--; ctlcode_consumed++;
	ctlcode_16 = ctlcode_8;

	if (ctlcode_16 < 0)
		goto b_block;
	else 
	if (ctlcode_16 == 0)
		goto w_run;
	else 
	if (ctlcode_16 == 1)
		goto w_block;
	else
		goto b_run;	
	
w_run:
	ctlcode_16 = read_be_word(src8ptr); src8ptr += 2;
	//printf(">lr[%d]\n", ctlcode_16);
b_run:
	data_8 = read_byte(src8ptr++);

	if (_verify)
	{
		predict_size = decompressed_size + ctlcode_16;
		if (predict_size > expected_size)
		{
			printf("overrun failure @ [run] code=%d, offset=%d, runsize=%d, overrun=%d\n", 
				(int)ctlcode_consumed-1, (int)decompressed_size, (int)ctlcode_16, (int)(predict_size - expected_size));
			decompressed_size = predict_size; 
			goto finish;
		}

		signed short r = 0;
		for (; r < ctlcode_16; ++r)
		{
			chk |= (read_byte(dst8ptr++) ^ data_8);
			if (chk != 0) break;
		}
			
		if (chk != 0)
		{
			printf("chksum failure @ [run] code=%d, offset=%d, runsize=%d, runpos=%d\n", 
				(int)ctlcode_consumed-1, (int)decompressed_size, (int)ctlcode_16, (int)r);
			goto finish;
		}
	}	
	else
	{
		signed short r = ctlcode_16 - 1;
		for (; r >= 0; --r)
			write_byte(dst8ptr++, data_8);
	}
	
	decompressed_size += ctlcode_16;
	goto decompress_loop;
	
b_block:
	ctlcode_16 = -ctlcode_16;
	goto block;
w_block:	
	ctlcode_16 = read_be_word(src8ptr); src8ptr += 2;
	//printf(">lb[%d]\n", ctlcode_16);
	
block:
	if (_verify)
	{
		predict_size = decompressed_size + ctlcode_16;
		if (predict_size > expected_size)
		{
			printf("overrun failure @ [block] code=%d, offset=%d, runsize=%d, overrun=%d\n", 
				(int)ctlcode_consumed-1, (int)decompressed_size, (int)ctlcode_16, (int)(predict_size - expected_size));
			decompressed_size = predict_size; 
			goto finish;
		}

		signed short b = 0;
		for (; b < ctlcode_16; ++b)
		{
			data_8 = read_byte(src8ptr++);
			chk |= (read_byte(dst8ptr++) ^ data_8);
			if (chk != 0) break;
		}
		if (chk != 0)
		{
			printf("chksum failure @ [run] code=%d, offset=%d, runsize=%d, runpos=%d\n", 
				(int)ctlcode_consumed-1, (int)decompressed_size, (int)ctlcode_16, (int)b);
			goto finish;
		}
	}	
	else
	{
		signed short b = ctlcode_16 - 1;
		for (; b >= 0; --b)
		{
			data_8 = read_byte(src8ptr++);
			write_byte(dst8ptr++, data_8);
		}
	}
	decompressed_size += ctlcode_16;
	goto decompress_loop;

finish:

	// encode error details
	if (_verify && (chk != 0))
		decompressed_size = ~decompressed_size;
				
	return decompressed_size;		
}

// compress palette frame in terms of words
datasize_t PCSImage::compress_palette(word_t* _src, byte_t* _dst, datasize_t palettefile_words, bool _verify)
{
	static const int tmpbuf_max = 16000;		
	signed long src_count = palettefile_words;
	signed long tmpbuf_count = 0;
	signed long tmprun_count = 1;
	signed short ctlcode_count = 0;
	
	byte_t* dst8ptr = _dst;
	word_t* src16ptr = _src;
	word_t* tmp16baseptr = new word_t[tmpbuf_max];
	if (!tmp16baseptr) { printf("error - could not allocate memory for compression buffer!\n"); getchar(); exit(1); }
	word_t* tmp16runptr = tmp16baseptr;
	word_t* ctlcount16ptr = (word_t*)dst8ptr;

	// arrange a run-terminator at the end of the buffer to force the RLE compressor
	// to break, re-evaluate it's current position and stop.
//	word_t* pterm = src16ptr + src_count;
//	pterm[0] = pterm[-1] ^ 0xFFFF;
//	pterm[1] = pterm[-1];
	
	// skip control code count
	dst8ptr += 2;

	// prefetch
	word_t prev_data = 0;
	word_t curr_data = read_word(src16ptr++); src_count--;
	
compress_loop:	

	// rotate & fill
	prev_data = curr_data;
	write_word(tmp16runptr++, curr_data); tmpbuf_count++;
	curr_data = read_word(src16ptr++); src_count--;	

	// buffer contains everything except curr_data

	// interrupt if source or temp buffer exhausted
	if ((tmpbuf_count >= tmpbuf_max) || (src_count < 0))
 		goto store;
 		 	
	if (curr_data == prev_data)
	{
		// continue [run]
		tmprun_count++;
	}
	else
	{
		// run break
		// store entire block & run, but not break value
		if (tmprun_count >= 4)
			goto store;
			
		// reset run, continue fill
		tmprun_count = 1;	
	}
	goto compress_loop;
	
	//------------------------------------------------------------------------*
	// store an control pattern (>=0 literal + >=0 run)
	// at this point we may have both buffered and run data waiting, with
	// the run being the latest part of the buffer, if present

store:

	// if we interrupted a run before it matures, or if run==1 (special meaning)
	// we need to absorb the run into a block
	if (tmprun_count < 4)
		tmprun_count = 0;
			
	// adjust to find waiting [literal block] length, versus [run] length
	tmpbuf_count -= tmprun_count;

	//------------------------------------------------------------------------*
	// store [literal block] first, if one is waiting

	if (tmpbuf_count > 0)
	{
		ctlcode_count++;
		
		if (tmpbuf_count > 127)
		{
			// long [literal block] : 1 ctrl
			write_byte(dst8ptr++, 1);
			write_be_word(dst8ptr, tmpbuf_count); dst8ptr += 2; 
			//printf("<lb[%d]\n", tmpbuf_count);
		}
		else
		{
			// short [literal block] : -ve ctrl
			// (careful dealing with signed values, converting from asm)
			write_byte(dst8ptr++, (byte_t)(-tmpbuf_count));
			//printf("<sb[%d]\n", tmpbuf_count);
		}				 

		// copy run	(words)
		tmp16runptr = tmp16baseptr;
		for (signed short c = tmpbuf_count - 1; c >= 0; --c)
		{
			write_be_word(dst8ptr, read_word(tmp16runptr++)); 
			dst8ptr += 2;
		}			
	}
	// reset [literal block] size
	tmp16runptr = tmp16baseptr;
	tmpbuf_count = 0;
	
	//------------------------------------------------------------------------*
	// store [run] last
	if (tmprun_count > 1)
	{
		ctlcode_count++;
		
		if (tmprun_count > 127)
		{
			// long [run] : 0 ctrl
			write_byte(dst8ptr++, 0);
			write_be_word(dst8ptr, tmprun_count); dst8ptr += 2;
			//printf("<lr[%d]\n", tmprun_count);
		}
		else
		{
			// short [run] : +ve ctrl
			write_byte(dst8ptr++, tmprun_count);
			//printf("<sr[%d]\n", tmprun_count);
		}				 

		// run value
		write_be_word(dst8ptr, prev_data); dst8ptr += 2;			
	}
	// reset [run] size
	tmprun_count = 1;

	// continue compressor

	if (src_count >= 0)
		goto compress_loop;

	//------------------------------------------------------------------------*
	// compression complete

done:
		
	// store control code count
	write_be_word(ctlcount16ptr, ctlcode_count);
	
	// free temporary space
	delete[] tmp16baseptr;
	
	datasize_t compressed_size = dst8ptr - _dst;

	if (_verify)
	{	
		datasize_t decompressed_size = decompress_palette(_dst, _src, palettefile_words, /*verify=*/true);
	
		if ((int)decompressed_size < 0)
		{
			printf("error - decompression verify failed checksum: sizein=%d, sizeout=%d\n", 
			(int)palettefile_words<<1, (int)(~decompressed_size));
			// indicate error in verify mode
			compressed_size = 0;
		}
		else if ((int)decompressed_size != palettefile_words<<1)
		{
			printf("error - decompression verify failed bounds: sizein=%d, sizeout=%d\n", 
			(int)palettefile_words<<1, (int)decompressed_size);
			// indicate error in verify mode
			compressed_size = 0;
		}
	}
	
	return (compressed_size);
}

datasize_t PCSImage::decompress_palette(byte_t* _src, word_t* _dst, datasize_t palettefile_words, bool _verify)
{			
	signed char ctlcode_8 = 0;
	signed short ctlcode_16 = 0;
	byte_t data_16 = 0;
	signed short ctlcode_count = 0;
	signed short ctlcode_consumed = 0;
	byte_t chk = 0;

	datasize_t decompressed_size = 0;
	datasize_t predict_size = 0;
	datasize_t expected_size = palettefile_words << 1;

	byte_t* src8ptr = _src;
	word_t* dst16ptr = _dst;
	

	// begin
		
	ctlcode_count = read_be_word(src8ptr); src8ptr += 2;
	
decompress_loop:

	if (ctlcode_count == 0)
		goto finish;	

	ctlcode_8 = read_byte(src8ptr++); ctlcode_count--; ctlcode_consumed++;
	ctlcode_16 = ctlcode_8;

	if (ctlcode_16 < 0)
		goto b_block;
	else 
	if (ctlcode_16 == 0)
		goto w_run;
	else 
	if (ctlcode_16 == 1)
		goto w_block;
	else
		goto b_run;	
	
w_run:
	ctlcode_16 = read_be_word(src8ptr); src8ptr += 2;
	//printf(">lr[%d]\n", ctlcode_16);
b_run:
	data_16 = read_be_word(src8ptr); src8ptr += 2;

	if (_verify)
	{
		predict_size = decompressed_size + ctlcode_16 * 2;
		if (predict_size > expected_size)
		{
			printf("overrun failure @ [run] code=%d, offset=%d, runsize=%d, overrun=%d\n", 
				(int)ctlcode_consumed-1, (int)decompressed_size, (int)ctlcode_16, (int)(predict_size - expected_size));
			decompressed_size = predict_size; 
			goto finish;
		}

		signed short r = 0;
		for (; r < ctlcode_16; ++r)
		{
			chk |= (read_word(dst16ptr++) ^ data_16);
			if (chk != 0) break;
		}
			
		if (chk != 0)
		{
			printf("chksum failure @ [run] code=%d, offset=%d, runsize=%d, runpos=%d\n", 
				(int)ctlcode_consumed-1, (int)decompressed_size, (int)ctlcode_16, (int)r);
			goto finish;
		}
	}	
	else
	{
		signed short r = ctlcode_16 - 1;
		for (; r >= 0; --r)
			write_word(dst16ptr++, data_16);
	}
	
	decompressed_size += ctlcode_16 * 2;
	goto decompress_loop;
	
b_block:
	ctlcode_16 = -ctlcode_16;
	goto block;
w_block:	
	ctlcode_16 = read_be_word(src8ptr); src8ptr += 2;
	//printf(">lb[%d]\n", ctlcode_16);
	
block:
	if (_verify)
	{
		predict_size = decompressed_size + ctlcode_16 * 2;
		if (predict_size > expected_size)
		{
			printf("overrun failure @ [block] code=%d, offset=%d, runsize=%d, overrun=%d\n", 
				(int)ctlcode_consumed-1, (int)decompressed_size, (int)ctlcode_16, (int)(predict_size - expected_size));
			decompressed_size = predict_size; 
			goto finish;
		}

		signed short b = 0;
		for (; b < ctlcode_16; ++b)
		{
			data_16 = read_be_word(src8ptr); src8ptr += 2;
			chk |= (read_word(dst16ptr++) ^ data_16);
			if (chk != 0) break;
		}
		if (chk != 0)
		{
			printf("chksum failure @ [run] code=%d, offset=%d, runsize=%d, runpos=%d\n", 
				(int)ctlcode_consumed-1, (int)decompressed_size, (int)ctlcode_16, (int)b);
			goto finish;
		}
	}	
	else
	{
		signed short b = ctlcode_16 - 1;
		for (; b >= 0; --b)
		{
			data_16 = read_be_word(src8ptr); src8ptr += 2;
			write_word(dst16ptr++, data_16);
		}
	}
	decompressed_size += ctlcode_16 * 2;
	goto decompress_loop;

finish:

	// encode error details
	if (_verify && (chk != 0))
		decompressed_size = ~decompressed_size;
				
	return decompressed_size;		
}
