// ---------------------------------------------------------------------------------------------------------------------
//	PhotoChrome v6.xx - Atari ST interlaced image reduction framework
//	1991-2016 Douglas Little (d.m.l)
// ---------------------------------------------------------------------------------------------------------------------
//	special thanks to:
//		evil@DHS, RA/PDX, Troed/SYNC for help testing and input on improvements
// ---------------------------------------------------------------------------------------------------------------------

// ====================================================================
//	reusable image processing stuff
// ====================================================================

#include "common.h"
#include "config.h"

#if defined(__ATARI__) || defined(__MINGW32__)
#include "FreeImage.3.15.3.h"
#else
#include "FreeImage.3.15.4.h"
#endif

#include "imagepkg.h"
#include "pcsimage.h"

extern Arguments g_allargs;

// --------------------------------------------------------------------

bool PCSImage::sanitise_source()
{
	if (active_source_)
	{
		int d = ImagePkg::instance().get_depth(active_source_);
		if (g_allargs.verbose)
			printf("check colour depth: %d ", d);
		if (d != 24)
		{
			if (g_allargs.verbose)
				printf("...converting to 24bit RGB... ");
			FIBITMAP *tmp = ImagePkg::instance().reformat(active_source_);
			if (!tmp)
			{
				if (g_allargs.verbose)
					printf("...failed!\n");				
				return false;
			}
			ImagePkg::instance().unload(active_source_);
			active_source_ = tmp;
		}
		if (g_allargs.verbose)
			printf("...ok.\n");

		// don't rescale in specific ccgen modes
		if ((g_allargs.ccmode != CCMode_SUPERPALETTE) &&
			(g_allargs.ccmode != CCMode_PLAYFIELD))
		{
			// read dimensions		
			int w = 0;
			int h = 0;
			ImagePkg::instance().get_size(active_source_, w, h);	
			if (g_allargs.verbose)
				printf("check dimensions: %d x %d ", w, h);
				
			if ((w != g_allargs.xsize) || 
				(h != g_allargs.ysize))
			{
				if (g_allargs.verbose)
					printf("...resizing to %d x %d... ", g_allargs.xsize, g_allargs.ysize);
				FIBITMAP *tmp = ImagePkg::instance().resize(active_source_, g_allargs.xsize, g_allargs.ysize);
				if (!tmp)
				{
					if (g_allargs.verbose)
						printf("...failed!\n");				
					return false;
				}
				ImagePkg::instance().unload(active_source_);
				active_source_ = tmp;			
			}
			if (g_allargs.verbose)
				printf("...ok.\n");				
		}
		
		return true;
	}
	
	return false;					
}

bool PCSImage::write_reference(const char* const _name)
{
	if (active_source_)
	{
		return ImagePkg::instance().write_png(active_source_, _name);
	}

	return false;
}

bool PCSImage::write_final(const char* const _name)
{
	return true;
}

FIBITMAP * PCSImage::read_image(const char* const _name, int _type)
{
	FIBITMAP *fib = 0;
	if (_type == 0)
		fib = ImagePkg::instance().read_jpg(_name);
	else if (_type == 1)
		fib = ImagePkg::instance().read_tga(_name);
	else if (_type == 2)
		fib = ImagePkg::instance().read_png(_name);
	else if (_type == 3)
		fib = ImagePkg::instance().read_bmp(_name);
//	else  if (_type < 0)
//		fib = generate_test_image((-_type)-1, 320, 200);
	else
		{ printf("error - unknown image type (%d)\n", _type); getchar(); exit(1); }
		
	return fib;
}

bool PCSImage::assign_source(FIBITMAP *fib)
{
	if (fib)
	{
		active_source_ = fib;
		
		return true;
	}
	return false;
}

bool PCSImage::read_source(int _type)
{
//	FIBITMAP *fib = read_image(name_.c_str(), _type);

	FIBITMAP *fib = 0;
	if (_type == 0)
		fib = ImagePkg::instance().read_jpg(name_.c_str());
	else if (_type == 1)
		fib = ImagePkg::instance().read_tga(name_.c_str());
	else if (_type == 2)
		fib = ImagePkg::instance().read_png(name_.c_str());
	else if (_type == 3)
		fib = ImagePkg::instance().read_bmp(name_.c_str());
	else  if (_type < 0)
		fib = generate_test_image((-_type)-1, 320, 200);
	else
		{ printf("error - unknown image type (%d)\n", _type); getchar(); exit(1); }
		
	if (fib)
	{
		active_source_ = fib;
		
		return true;
	}
	return false;
}

bool PCSImage::read_blank(int _w, int _h)
{
	FIBITMAP *fib = 0;
	fib = ImagePkg::instance().read_blank(_w, _h);
		
	if (fib)
	{
		active_source_ = fib;
		
		return true;
	}

	return false;
}

// --------------------------------------------------------------------

void PCSImage::perform_diagnostics(byte_t** _bitmaps)
{
	printf("performing diagnostic passes...\n");	

	init_diagnostic_units();

	
	for (std::vector<DiagnosticUnit*>::iterator it = diagnosticunits_.begin(); 
		 it != diagnosticunits_.end(); 
		 ++it)
	{	
		DiagnosticUnit* du = *it;		
		diagnostic_pass(_bitmaps, du);
	}

	if (fields_ > 1)
	{
		for (int f = 0; f < fields_; ++f)
		{
			char buf[1024];
			if (!g_allargs.quiet)
				printf("writing field %d diagnostic...\n", f);	
			generate_field_image(_bitmaps, f);
//			std::string err_name("field0.png");
			sprintf(buf, "field%d.png", f);
			std::string err_name(buf);
//			err_name << "field" << f << ".png");
			ImagePkg::instance().write_png(error_, err_name.c_str());
		}
	
//		{
//			if (!g_allargs.quiet)
//				printf("writing field 1 diagnostic...\n");	
//			generate_field_image(_bitmaps, 1);
//			std::string err_name("field1.png");
//			ImagePkg::instance().write_png(error_, err_name.c_str());
//		}
	}

	{
		if (!g_allargs.quiet)
			printf("writing final diagnostic...\n");	
		generate_final_image(_bitmaps);
		std::string err_name("final.png");
		ImagePkg::instance().write_png(error_, err_name.c_str());
	}

	//generate_cie_error_image();	
}

// --------------------------------------------------------------------

void PCSImage::diagnostic_pass(byte_t** _bitmaps, DiagnosticUnit* _du)
{
	if (!g_allargs.aPCT)
		return;

	ImagePkg& imgpkg = ImagePkg::instance();
		
	_du->reset();
			
	calc_real err = 0.0f;

	const calc_real fnp = 1.0f / (calc_real)pal_crange(colourdepth_);
	const calc_real fns = 1.0f / (calc_real)src_crange();
	const calc_real ffn = 1.0f / (calc_real)fields_;
	
	signed short palette_scan_size = palettebuffer_stride_;
	signed short bitmap_scan_size = display_width_;
	
	for (signed short scan = 1; scan < display_height_; ++scan)
	{
		// locate synctable for this specific scanline
		byte_t* sync_mapping_scan = ADDRESS_PCT_SCAN(g_allargs.aPCT, 0, display_width_, scan);
		
		for (int x = 0; x < display_width_; ++x)
		{					
			// extract reference pixel
			byte_t ur, ug, ub;
			imgpkg.get_rgb(active_source_, (int)x, (int)scan, display_height_, ur, ug, ub);
			
			signed short sr = ((unsigned short)ur);
			signed short sg = ((unsigned short)ug);
			signed short sb = ((unsigned short)ub);
			
			// into unit space (1.0 = saturated)
			calc_real fsx = (calc_real)sr * fns;
			calc_real fsy = (calc_real)sg * fns;
			calc_real fsz = (calc_real)sb * fns;

			// extract frame pixels

			calc_real fax = 0.0f;			
			calc_real fay = 0.0f;			
			calc_real faz = 0.0f;			
			
			for (int f = 0; f < fields_; ++f)
			{
				word_t* active_palette = palettes_[f];
				byte_t* active_bitmap = _bitmaps[f];
				word_t* scan_palette = &active_palette[scan * palette_scan_size];
				byte_t* scan_bitmap = &active_bitmap[scan * bitmap_scan_size];

				// get pixel index
				byte_t p = scan_bitmap[x];

				byte_t* sync_mapping = &sync_mapping_scan[x << MAX_PALETTE_SHIFT];
					
				// index correct palette for this x & p
				byte_t upi = sync_mapping[p];
				signed short spi = (signed short)((signed char)upi);		
				word_t* x_palette = &scan_palette[spi * g_allargs.colours];

				// find colour (0x0RGB)
				word_t pc = x_palette[p];

				// separate colour channels in palette space
				signed short pr = pc >> (r_pos_ - colourdepth_) & c_mask_;
				signed short pg = pc >> (g_pos_ - colourdepth_) & c_mask_;
				signed short pb = pc >> (b_pos_ - colourdepth_) & c_mask_;
								
				// into unit space (1.0 = saturated)
				calc_real fpr = (calc_real)pr * fnp;
				calc_real fpg = (calc_real)pg * fnp;
				calc_real fpb = (calc_real)pb * fnp;
											
				// error sum				
				fax += fpr;
				fay += fpg;
				faz += fpb;
			}
			
			// normalize for field count
			fax *= ffn;			
			fay *= ffn;	
			faz *= ffn;
			
			// quantize into diagnostic space and back, since the native virtual space
			// and source spaces doesn't map exactly in integer space. this removes 
			// artificial space translation errors which have nothing to do with image 
			// reduction itself.
			
			_du->quantize(fsx, fsy, fsz);
			_du->quantize(fax, fay, faz);
			
//			fax = col2f(f2col(fax));
//			fay = col2f(f2col(fay));
//			faz = col2f(f2col(faz));

			// sample error
			_du->sample
			(
				error_, x, scan,
				fax, fay, faz, 
				fsx, fsy, fsz
			);
		}
	}	
	
	// evaluate final error / generate result
	_du->evaluate(error_);
}

// --------------------------------------------------------------------

void PCSImage::generate_final_image(byte_t** _bitmaps)
{
	if (!g_allargs.aPCT)
		return;

	ImagePkg& imgpkg = ImagePkg::instance();

	const float fnp = 1.0f / (float)pal_crange(colourdepth_);
	const float ffn = 1.0f / (float)fields_;
				
	signed short palette_scan_size = palettebuffer_stride_;
	signed short bitmap_scan_size = display_width_;

	for (signed short scan = 1; scan < display_height_; ++scan)
	{
		//byte_t* sync_mapping_scan = &g_allargs.aPCT[(scan-1)*display_width_];	
		byte_t* sync_mapping_scan = ADDRESS_PCT_SCAN(g_allargs.aPCT, 0, display_width_, scan);
		
		for (int x = 0; x < display_width_; ++x)
		{		
			// extract frame pixels

			float fax = 0.0f;			
			float fay = 0.0f;			
			float faz = 0.0f;			
			
			for (int f = 0; f < fields_; ++f)
			{
				word_t* active_palette = palettes_[f];
				byte_t* active_bitmap = _bitmaps[f];
				word_t* scan_palette = &active_palette[scan * palette_scan_size];
				byte_t* scan_bitmap = &active_bitmap[scan * bitmap_scan_size];

				// get pixel index
				byte_t p = scan_bitmap[x];

				byte_t* sync_mapping = &sync_mapping_scan[x << MAX_PALETTE_SHIFT];
					
				// index correct palette for this x & p
				byte_t upi = sync_mapping[p];
				signed short spi = (signed short)((signed char)upi);		
				word_t* x_palette = &scan_palette[spi * g_allargs.colours];

				// find colour (0x0RGB)
				word_t pc = x_palette[p];
				
				// separate colour channels in palette space
				signed short pr = pc >> (r_pos_ - colourdepth_) & c_mask_;
				signed short pg = pc >> (g_pos_ - colourdepth_) & c_mask_;
				signed short pb = pc >> (b_pos_ - colourdepth_) & c_mask_;
								
				// into unit space (1.0 = saturated)
				float fpr = (float)pr * fnp;
				float fpg = (float)pg * fnp;
				float fpb = (float)pb * fnp;
								
				// error sum				
				fax += fpr;
				fay += fpg;
				faz += fpb;
			}
			
			// normalize for field count
			fax *= ffn;			
			fay *= ffn;	
			faz *= ffn;	
			
			// back into 8-bit space
			signed short ax = f2col(fax);
			signed short ay = f2col(fay);
			signed short az = f2col(faz);
			
			imgpkg.put_rgb(error_, (int)x, (int)scan, display_height_, (byte_t)ax, (byte_t)ay, (byte_t)az);			
		}
	}		
}
	
void PCSImage::generate_field_image(byte_t** _bitmaps, int _f)
{
	if (!g_allargs.aPCT)
		return;

	ImagePkg& imgpkg = ImagePkg::instance();
			
	const float fnp = 1.0f / (float)pal_crange(colourdepth_);

	signed short palette_scan_size = palettebuffer_stride_;
	signed short bitmap_scan_size = display_width_;

	for (signed short scan = 1; scan < display_height_; ++scan)
	{
		//byte_t* sync_mapping_scan = &g_allargs.aPCT[(scan-1)*display_width_];	
		byte_t* sync_mapping_scan = ADDRESS_PCT_SCAN(g_allargs.aPCT, 0, display_width_, scan);
		
		for (int x = 0; x < display_width_; ++x)
		{		
			// extract frame pixels
			
			word_t* active_palette = palettes_[_f];
			byte_t* active_bitmap = _bitmaps[_f];
			word_t* scan_palette = &active_palette[scan * palette_scan_size];
			byte_t* scan_bitmap = &active_bitmap[scan * bitmap_scan_size];

			// get pixel index
			byte_t p = scan_bitmap[x];

			byte_t* sync_mapping = &sync_mapping_scan[x << MAX_PALETTE_SHIFT];
				
			// index correct palette for this x & p
			byte_t upi = sync_mapping[p];
			signed short spi = (signed short)((signed char)upi);		
			word_t* x_palette = &scan_palette[spi * g_allargs.colours];

			// find colour (0x0RGB)
			word_t pc = x_palette[p];
			
			// separate colour channels in palette space
			signed short pr = pc >> (r_pos_ - colourdepth_) & c_mask_;
			signed short pg = pc >> (g_pos_ - colourdepth_) & c_mask_;
			signed short pb = pc >> (b_pos_ - colourdepth_) & c_mask_;
							
			// into unit space (1.0 = saturated)
			float fpr = (float)pr * fnp;
			float fpg = (float)pg * fnp;
			float fpb = (float)pb * fnp;
						
			// back into 8-bit space
			signed short ar = f2col(fpr);
			signed short ag = f2col(fpg);
			signed short ab = f2col(fpb);
									
			imgpkg.put_rgb(error_, (int)x, (int)scan, display_height_, (byte_t)ar, (byte_t)ag, (byte_t)ab);			
		}
	}		
}

// --------------------------------------------------------------------
	
bool PCSImage::convert()
{	
	if (!g_allargs.aPCT)
		return true;

	// create error buffer
	error_ = ImagePkg::instance().read_blank(display_width_, display_height_);

	if (g_allargs.method >= 3)
	{	
		if (g_allargs.verbose)
			printf("creating perception filter...\n");
			
		// perception mask is only used when using scoring methods, versus sampling-order methods
		generate_perception_mask();
	}
		
	if (g_allargs.verbose)
		printf("generating image...\n");
	
	init_error_unit(g_allargs.colourdepth);
	
	return transform_source_to_target();				
}

// --------------------------------------------------------------------
// generate perception-based mask for weighting pixel importance values

void PCSImage::generate_perception_mask()
{		
	// create perception mask buffer
	mask_ = new byte_t[display_width_ * display_height_];

	ImagePkg& imgpkg = ImagePkg::instance();	

	if (!g_allargs.pfm_source.empty())
	{
		if (g_allargs.verbose)
			printf("reading user defined perception filter mask... [%s]\n", g_allargs.pfm_source.c_str());
		
		// load specific PFM image from disk
		FIBITMAP* masksrc = ImagePkg::instance().read_png(g_allargs.pfm_source.c_str());
			
		// resize it to suit display, if required
		int ww, hh;
		ImagePkg::instance().get_size(masksrc, ww, hh);
		if ((ww != display_width_) || (hh != display_height_))
		{
			FIBITMAP* mask = ImagePkg::instance().resize(masksrc, display_width_, display_height_);
			ImagePkg::instance().unload(masksrc);
			masksrc = mask;
		}
			
		// use the RED channel as the mask - ignore the others
		for (signed short yo = 0; yo < display_height_; ++yo)
		{
			for (signed short xo = 0; xo < display_width_; ++xo)
			{
				// extract pixel
				byte_t uro, ugo, ubo;
				imgpkg.get_rgb(masksrc, (int)xo, (int)yo, display_height_, uro, ugo, ubo);					

				// write mask
				mask_[xo + (yo * display_width_)] = uro;					
			}
		}
		
		// release temporary image			
		ImagePkg::instance().unload(masksrc);
	}
	else
	{
		if (g_allargs.verbose)
			printf("generating internal perception filter mask...\n");
		
		// generate PFM using internal filter
		
		static const float filter_coeffs[] = 
		{ 
			0.707, 1.000, 0.707,
			1.000, 0.000, 1.000,
			0.707, 1.000, 0.707
		};

		int progress = 0;
		int last_progress_pct = -1;
					
		for (signed short yo = 0; yo < display_height_; ++yo)
		{
			for (signed short xo = 0; xo < display_width_; ++xo)
			{
				// extract pixel
				byte_t uro, ugo, ubo;
				imgpkg.get_rgb(active_source_, (int)xo, (int)yo, display_height_, uro, ugo, ubo);
				signed short sro = (unsigned short)uro;
				signed short sgo = (unsigned short)ugo;
				signed short sbo = (unsigned short)ubo;
	
				float dd = 0;
				//float ff = 0;
				//float ad = 0;
				//float adn = 0;
				for (signed short yi = 0; yi < 3; ++yi)
				{
					for (signed short xi = 0; xi < 3; ++xi)
					{
						//if (xi == 1 && yi == 1) continue;
							
						signed short x = xclamp(xo + xi - 1, 0, display_width_-1);					
						signed short y = xclamp(yo + yi - 1, 0, display_height_-1);
	
						// extract pixel
						byte_t uri, ugi, ubi;
						imgpkg.get_rgb(active_source_, (int)x, (int)y, display_height_, uri, ugi, ubi);
						signed short sri = (unsigned short)uri;
						signed short sgi = (unsigned short)ugi;
						signed short sbi = (unsigned short)ubi;
						
						// calculate relation to neighbours
						float f = filter_coeffs[xi + (yi * 3)];
						float dx = ((float)sri - (float)sro);
						float dy = ((float)sgi - (float)sgo);
						float dz = ((float)sbi - (float)sbo);
											
						// get colour distance to this neighbour
						float d = sqrt((dx * dx) + (dy * dy) + (dz * dz));
						
						// transform distance into 'significance' using mapping with smallest distance = 1.0, greatest >= 0.0 
						// note: this transform means greatest resolution assigned to smallest distances, which is what we want
						float rd = 1.0 / (2.0 + d);
						
						// attenuate sample significance with filter matrix
						rd *= f;// * 0.66666f;
	
						// combine significance using inverse product 
						// (significance can increased by small distances, but not decreased by large distances)
						dd = inverse_product<double>(dd, rd);
						//ad += rd;
						//adn += 1.0f;
					}
				}
				
				//dd = pow(dd, 1.0 / 3.0);
				dd *= 255.99999f;
				
				byte_t sample = xclamp((int)dd, 0, 255);
				
				mask_[xo + (yo * display_width_)] = sample;
			
				// now apply the gamma/influence transform directly to the diagnostic image, which will get repeated during reduction
				{					 
					// look up perception filter mask (range: 1.0 -> 0.0)
					double perception_filter_mask = (double)((unsigned int)sample) / 255.0f;
					
					// apply gamma curve
					perception_filter_mask = xclamp<double>(pow(perception_filter_mask, g_allargs.pfm_gamma), 0.0, 1.0);
					
					// attenuate influence
					perception_filter_mask = 1.0f - ((1.0f - perception_filter_mask) * g_allargs.pfm_influence);
					
					byte_t dsample = xclamp((int)(perception_filter_mask * 255.99999f), 0, 255);					
					
					imgpkg.put_rgb(error_, (int)xo, (int)yo, display_height_, dsample, dsample, dsample);			
				}
			}
			
			#ifdef __ATARI__
			// update progress
			if (!g_allargs.quiet)
			{
				int progress_pct = ((int)yo * 100) / display_height_;
				if (progress_pct != last_progress_pct)
				{
					last_progress_pct = progress_pct;		
					switch (progress & 3)
					{
					case 0:
						printf("- %2d%%\r", progress_pct); break;
					case 1:
						printf("\\ %2d%%\r", progress_pct); break;
					case 2:
						printf("| %2d%%\r", progress_pct); break;
					case 3:
						printf("/ %2d%%\r", progress_pct); break;
					};
					fflush(stdout);
					progress++;
				}
			}
			#endif
		}
		
		if (g_allargs.diagnostics)
		{
			// output our generated mask
			std::string err_name("pcfmask.png");
			ImagePkg::instance().write_png(error_, err_name.c_str());
		}
	}
}

// --------------------------------------------------------------------
// reduce all fields

bool PCSImage::transform_source_to_target()
{
	int f;
	bool success = true;
	
	datasize_t chunky_bitmap_size = display_width_ * display_height_;
	
		
	for (f = 0; f < fields_; ++f)	
	{
		chunky_bitmaps_[f] = new byte_t[chunky_bitmap_size];
		if (!chunky_bitmaps_[f]) { printf("error - could not allocate memory for chunky buffer!\n"); getchar(); exit(1); }
		clear(chunky_bitmaps_[f], chunky_bitmap_size);			
	}
		
	// each iteration 'bounces' a portion of the last error term between fields until it settles down
	// it is only meaningful in 'ErrorType_FIELD' mode
	error_iteration_ = 0;
	for (int i = 0; i < g_allargs.erroriters; ++i)
	{
		if (g_allargs.verbose && (g_allargs.erroriters > 1))
			printf("error iteration %d... \n", i);

		// sliding error feedback requires error unit to be aware of progress
		if (errorunit_)
			errorunit_->set_iteration(i, g_allargs.erroriters);
		
		for (f = 0; f < fields_; ++f)	
		{
			// error unit needs to know which field we are processing
			if (errorunit_)
				errorunit_->set_field(f);
			
			reduce_frame(chunky_bitmaps_[f], palettes_[f], f, fields_);
		}

		if (g_allargs.diagnostics)
		{	
			// all chunky bitmaps required for error calculations
			perform_diagnostics(chunky_bitmaps_);
		}
		
		++error_iteration_;
	}

	for (f = 0; f < fields_; ++f)	
	{
		if (g_allargs.verbose)
			printf("convert field to planar representation... \n");
			
		clear(bitmaps_[f], framebuffer_words_);
		c2p(chunky_bitmaps_[f], bitmaps_[f]);
	}
	
	// clean up	as early as possible
	for (f = 0; f < fields_; ++f)	
	{
		delete[] chunky_bitmaps_[f]; chunky_bitmaps_[f] = 0;
	}
	
	// swizzle palettes for HW regs
	if (g_allargs.fieldmode == FieldMode_SharedBitmap)
	{
		// todo: fixed at 2 palettes
		
		// we only generated a palette from one field, but it has additional depth
		// clone the palette in order to quantize two native-bitdepth versions from the virtual bitdepth

		if (g_allargs.verbose)
			printf("swizzling shared bitmap (dual palette) fields from single high bitdepth palette... \n");

		// 2nd palette will not have been allocated in advance when fields==1
		delete[] palettes_[1];
		palettes_[1] = new word_t[palettebuffer_words_ + 4];

		// in SharedBitmap mode, native colour depth is 1 level lower than converted depth
		int native_cdepth = colourdepth_ - 1;
		word_t native_cmask = (1 << native_cdepth) - 1;
		word_t normalized_cmask = native_cmask << (5 - native_cdepth);
		
		// error-correct palettes using half native colour depth step, not losing odd bit
		// note: this adjustment always rounds downwards - not as faithful as normal single or multi-field conversion!
		int step = (1 << (5 - native_cdepth));
		int odd_offset = (step >> 1);
		int even_offset = (step - odd_offset) - 1;
				
		// offset one field by one bit		
		word_t* src = palettes_[0];
		word_t* dst = palettes_[1];
		for (int w = 0; w < palettebuffer_words_; ++w)
		{
			word_t ps = src[w];
			
			// use binary here just for easy reading / self-documentation
			int r = (ps & binary(0111110000000000)) >> 10;
			int g = (ps & binary(0000001111100000)) >> 5;
			int b = (ps & binary(0000000000011111)) >> 0;		
//			int r = (ps & 0x7C00) >> 10;
//			int g = (ps & 0x03E0) >> 5;
//			int b = (ps & 0x001F) >> 0;		

			int re = xclamp<int>(r+even_offset,0,31);
			int ge = xclamp<int>(g+even_offset,0,31);
			int be = xclamp<int>(b+even_offset,0,31);
			
			int ro = xclamp<int>(r-odd_offset,0,31);
			int go = xclamp<int>(g-odd_offset,0,31);
			int bo = xclamp<int>(b-odd_offset,0,31);

//			word_t p1 = ps;
			word_t p1 = ((re & normalized_cmask) << 10) | ((ge & normalized_cmask) << 5) | ((be & normalized_cmask) << 0);;
			word_t p2 = ((ro & normalized_cmask) << 10) | ((go & normalized_cmask) << 5) | ((bo & normalized_cmask) << 0);;

			// alternate offset to keep intensity bias low
			int eval = (w / palettebuffer_stride_) & 1;
			if (eval)
			{
				word_t tmp = p2;
				p2 = p1;
				p1 = tmp;
			}
						
			src[w] = p1;
			dst[w] = p2;
		}
				
		// swizzle both fields
		palette_swizzle(palettes_[0], palettebuffer_words_);
		palette_swizzle(palettes_[1], palettebuffer_words_);		
	}
	else
	{
		if (g_allargs.verbose)
			printf("swizzling palette fields... \n");
		
		// just swizzle each field palette separately
		for (f = 0; f < fields_; ++f)	
			palette_swizzle(palettes_[f], palettebuffer_words_);
	}

	return success;
}

// --------------------------------------------------------------------
// process a single sub-frame of the image

void PCSImage::reduce_frame
(
	byte_t* _bitmap, 
	word_t* _palette, 
	int _index, 
	int _num_indices
)
{	
	clear(_palette, palettebuffer_words_);

	// initialise interlace, dither & reduction units for this field
	init_interlace_unit(_index);
	init_dither_unit(_index);
	init_reduce_unit(_index);

	// IMPORTANT:
	// offset by -2 of 3 palettes, since conversion starts on 2nd scanline, pixels can
	// refer backwards by 1 palette - map contains indices to previous scan!
	// palette buffer for 1st scanline has only space for a setup palette
	// in simple terms, reduce_frame_method0 assumes all scanlines are equal but
	// destination buffer is smaller for 1st scan.
	//
	// note: this doesn't happen here anymore - the pct table determines padding/offsetting of palettes

	reduceunit_->reduce(_bitmap, _palette, g_allargs.aPCT, g_allargs.aDPT);
}

// --------------------------------------------------------------------

void PCSImage::clear(word_t* _dst, datasize_t _count)
{
	for (datasize_t c = 0; c < _count; ++c)
		*(_dst++) = 0;
}

void PCSImage::clear(byte_t* _dst, datasize_t _count)
{
	for (datasize_t c = 0; c < _count; ++c)
		*(_dst++) = 0;
}

// --------------------------------------------------------------------

void PCSImage::palette_swizzle(word_t* _dst, datasize_t _count)
{
	// swizzle internal palette representation into Atari hardware palette register representation
	//
	// note: used to expect 0RGB 12bit colour format, but has been modified to expect 15bit colour following some other changes
	// it is just a storage format - the entries will only hold additional bits if the colour reduction generated them
	// it is the responsibility of earlier stages to make sure the palettes are already quantized/rounded/offset for each field
	for (datasize_t c = 0; c < _count; ++c)
	{
		word_t val = *(_dst);
//		word_t r = (val & 0x0E00) >> 1 | (val & 0x0100) << 3;
//		word_t g = (val & 0x00E0) >> 1 | (val & 0x0010) << 3;
//		word_t b = (val & 0x000E) >> 1 | (val & 0x0001) << 3;

		// use binary here just for easy reading / self-documentation
		word_t r = ((val & binary(0111000000000000)) >> 4) | ((val & binary(0000100000000000)) << 0);
		word_t g = ((val & binary(0000001110000000)) >> 3) | ((val & binary(0000000001000000)) << 1);
		word_t b = ((val & binary(0000000000011100)) >> 2) | ((val & binary(0000000000000010)) << 2);
		//word_t r = ((val & 0x7000) >> 4) | ((val & 0x0800) << 0);
		//word_t g = ((val & 0x0380) >> 3) | ((val & 0x0040) << 1);
		//word_t b = ((val & 0x001C) >> 2) | ((val & 0x0002) << 2);

		word_t vv = r | g | b;

		
		*(_dst++) = vv;
	}
}

/*
void PCSImage::palette_deswizzle(word_t* _dst, datasize_t _count)
{
	for (datasize_t c = 0; c < _count; ++c)
	{
		word_t val = *(_dst);
//		word_t r = (val & 0x0700) << 1 | (val & 0x0800) >> 3;
//		word_t g = (val & 0x0070) << 1 | (val & 0x0080) >> 3;
//		word_t b = (val & 0x0007) << 1 | (val & 0x0008) >> 3;
		*(_dst++) = r | g | b;
	}
}
*/
// --------------------------------------------------------------------
// simple c2p transpose (not fast, but does handle varying plane count)

void PCSImage::c2p(byte_t* _cin, word_t* _pout)
{	
	byte_t* src = _cin;
	word_t* dst = _pout;

	// v line loop
	for (signed short yc = display_height_ - 1; yc >= 0; --yc)
	{
		// h word loop
		for (signed short xc = (display_width_ >> 4) - 1; xc >= 0; --xc)
		{
			// h pixel loop				
			for (signed short xbc = 16-1; xbc >= 0; --xbc)
			{
				byte_t val = *(src++);
				
				// plane loop
				word_t* o = dst;
				for (signed short p = display_planes_ - 1; p >= 0; --p)
				{
					*(o) = (*(o) << 1) | (val & 1); val >>= 1; o++;
				}
			}
			
			// advance per word
			dst += display_planes_;
		}
	}
}	

// --------------------------------------------------------------------
// create test images

FIBITMAP* PCSImage::generate_test_image(int _index, signed short _w, signed short _h)
{
	ImagePkg& imgpkg = ImagePkg::instance();

	FIBITMAP* img = imgpkg.read_blank(_w, _h);
	
	switch (_index)
	{
	default:
		printf("unknown testcard index (%d)\n", _index);
	case 0:
		// perfect 8-bit greyscale gradient
		{		
			// 16 divisions (20 pixels for 320-wide)
			int dw = _w / 256;
			// 16 divisions (12 pixels for 200-high)
			int dh = _h / 256;
						
			for (int y = 1; y < _h; ++y)
			{
				for (int x = 0; x < _w; ++x)
				{
					int u = x / dw;
					int v = y / dh;
					
					int c = u + (v * 0);
					
					imgpkg.put_rgb(img, x, y, _h, (byte_t)c, (byte_t)c, (byte_t)c);					
				}
			}
		}
		break;
	case 1:
		// perfect 'native virtual' greyscale gradient (i.e. best depth achievable in current mode)
		{		
			for (int y = 1; y < _h; ++y)
			{
				for (int x = 0; x < _w; ++x)
				{
					// unit space
					float u = (float)x / (float)_w;

					// quantize into native virtual space
					int p = f2range(u, (float)nvi_crange(colourdepth_, fields_));
					// back to unit space
					float u2 = range2f(p, (float)nvi_crange(colourdepth_, fields_));

					// into 8-bit space / source format					
					int c = f2col(u2);
					
					imgpkg.put_rgb(img, x, y, _h, (byte_t)c, (byte_t)c, (byte_t)c);					
				}
			}
		}
		break;
	case 2:
		// pure colour noise - worst case scenario
		{		
			LCG r(12951);
									
			for (int y = 1; y < _h; ++y)
			{
				for (int x = 0; x < _w; ++x)
				{
					unsigned short cr = r.lcg_sample();
					unsigned short cg = r.lcg_sample();
					unsigned short cb = r.lcg_sample();
					
					imgpkg.put_rgb(img, x, y, _h, (byte_t)cr, (byte_t)cg, (byte_t)cb);					
				}
			}
		}
		break;
	}	
	
	return img;
}

// --------------------------------------------------------------------
