// ---------------------------------------------------------------------------------------------------------------------
//	PhotoChrome v5.xx - Atari ST interlaced image reduction framework
//	1991-2012 Douglas Little (d.m.l)
// ---------------------------------------------------------------------------------------------------------------------
//	special thanks to:
//
//		evil@DHS for noticing faults, making additional suggestions & testing ideas
//		Frank B for encouraging me back into Atari land
//		Vincent Riviere's port of GCC 4.x.x for MiNT/TOS
//		various members of Atari-Forum for adding their 2p (you know who you are!)
//		*all* users of PhotoChrome, Apex etc. - past and future
// ---------------------------------------------------------------------------------------------------------------------

#ifndef pcsimage_h_
#define pcsimage_h_

// ====================================================================
//	reusable image processing stuff
// ====================================================================

#include "ditherunit.h"
#include "laceunit.h"
#include "errorunit.h"
#include "reduceunit.h"
#include "diagnosticunit.h"

#include <assert.h>

extern Arguments g_allargs;

// --------------------------------------------------------------------

#define USE_MULTILINE_PCT (1)

#if (USE_MULTILINE_PCT)
#define ADDRESS_PCT_SCAN(_PCT_, _x_, _display_width_, _scan_) (&_PCT_[(_x_ + (_display_width_ * (_scan_))) << MAX_PALETTE_SHIFT])
#else
#define ADDRESS_PCT_SCAN(_PCT_, _x_, _display_width_, _scan_) (&_PCT_[_x_ << MAX_PALETTE_SHIFT])
#endif

#define ADDRESS_PCT(_PCT_, _x_) (&_PCT_[_x_ << MAX_PALETTE_SHIFT])

// --------------------------------------------------------------------
// object representing a raw true colour source image pulled from a file

class PCSImage
{
	
	friend class ReduceUnit;
	
protected:

	// --------------------------------------------------------------------
	// constants

	static const int max_fields_ = 4;							// unlikely to ever want more than 4 fields!
	static const int displayword_bits_ = sizeof(word_t) << 3;	// unlikely to need specialised

//	static const int display_planes_ = 4;						// specialise only if attempting medium-res or TT/falcon mode
//	static const int display_colours_ = 1 << display_planes_;

	//static const signed short palettebuffer_stride_ = (3 * 16);
	//static const signed short palettebuffer_preload_ = 16;
	
	// --------------------------------------------------------------------
	// storage/state
	
	FIBITMAP *active_source_;
	FIBITMAP *error_;
	
	// cryptochrome specific
	FIBITMAP *cc_field0_;
	FIBITMAP *cc_field1_;
	
	byte_t* mask_;
	std::string name_;
	unsigned char* pcs_filebuf_;
	datasize_t pcs_filesize_;
	bool verify_;

	//unsigned long* reference_;	
	word_t* bitmaps_[max_fields_];	
	word_t* palettes_[max_fields_];
	byte_t* chunky_bitmaps_[max_fields_];
	
	int error_iteration_;
		
	datasize_t framebuffer_words_;
	datasize_t palettebuffer_words_;
	int palette_skip_;
	int palette_count_;
	signed short palettebuffer_stride_;
	int display_planes_;						// specialise only if attempting medium-res or TT/falcon mode
	int display_colours_;


private:
	
	// --------------------------------------------------------------------
	// off-limits
	
	// prevent copying of this type
	PCSImage(PCSImage& _other)
	{
	}

	PCSImage(const PCSImage& _other)
	{
	}
			
public:

	// --------------------------------------------------------------------

	int display_width_;
	int display_height_;

	int fields_;
	int colourdepth_;
	signed short r_pos_;
	signed short g_pos_;
	signed short b_pos_;
	unsigned short c_mask_;

	// --------------------------------------------------------------------
	// primary constructor (from source filename)
	
	PCSImage
	(
		const char* const _name, 
		int _w, int _h, int _d,
		int _colours, int _planes,
		int _palette_count, int _palette_skip
	)
		: name_(_name), verify_(true),
		  active_source_(0), error_(0), mask_(0), 
		  cc_field0_(0), cc_field1_(0),
		  pcs_filebuf_(0), pcs_filesize_(0),
		  framebuffer_words_(0), palettebuffer_words_(0), 
		  display_colours_(_colours), display_planes_(_planes),
		  palette_count_(_palette_count), palette_skip_(_palette_skip),
		  fields_(0),
		  colourdepth_(_d), r_pos_(0), g_pos_(0), b_pos_(0), c_mask_(0),
		  ditherunit_(0), laceunit_(0), errorunit_(0), reduceunit_(0),
		  display_width_(_w),
		  display_height_(_h),
		  error_iteration_(0)
	{		
		for (int f = 0; f < max_fields_; ++f)
		{
			bitmaps_[f] = 0;
			palettes_[f] = 0;
			chunky_bitmaps_[f] = 0;
		}
		
		palettebuffer_stride_ = palette_count_ * g_allargs.colours;
		
		// generic setup for RGB positions and masks for this specified depth
		
		// R,G,B bit position markers
//		if (colourdepth_ <= 4)
//		{
//			// normal bit positions for palettes up to 12bit, pre-swizzling
//			r_pos_ = 12;
//			g_pos_ = 8;
//			b_pos_ = 4;
//		}
//		else if (colourdepth_ == 5)
		if (colourdepth_ <= 5)
		{
			// for palettes (temporarily) exceeding 12bit before transformation, special arrangement required
			// note: this is really only required for STE images generating a single 'shared' bitmap and dual palette,
			// to minimise impact on the colour reduction system - the palette will get split and exported as two
			// 12-bit palettes in Atari format.
			r_pos_ = 15;
			g_pos_ = 10;
			b_pos_ = 5;
		}
		else
		{
			assert(false); // oops...
		}
		
		c_mask_ = (1 << colourdepth_) - 1;	// channel mask, when denormalized

		printf("output: %dx%d @ %d bpp\n", display_width_, display_height_, colourdepth_);
	}

	// --------------------------------------------------------------------
	// destructor
		
	virtual ~PCSImage()
	{
		// buffers/arrays
		for (int f = 0; f < max_fields_; ++f)
		{
			delete[] bitmaps_[f]; bitmaps_[f] = 0;
			delete[] palettes_[f]; palettes_[f] = 0;
			delete[] chunky_bitmaps_[f]; chunky_bitmaps_[f] = 0;
		}
		
		delete[] mask_; mask_ = 0;

		// objects		
		delete ditherunit_; ditherunit_ = 0;
		delete laceunit_; laceunit_ = 0;
		delete errorunit_; errorunit_ = 0;
		delete reduceunit_; reduceunit_ = 0;

		while (!diagnosticunits_.empty())
		{
			delete diagnosticunits_.back(); 
			diagnosticunits_.pop_back();
		}

		// images		
		// note: no refcounting - can only delete once. i.e. prevent copying for safety
		// note to self - get smartptr impl. in here to clean this nonsense up? or better leave it almost-ANSI?
		if (active_source_)
			ImagePkg::instance().unload(active_source_);
		if (error_)
			ImagePkg::instance().unload(error_);
				
		if (cc_field0_)
			ImagePkg::instance().unload(cc_field0_);
		if (cc_field1_)
			ImagePkg::instance().unload(cc_field1_);
	}

	// --------------------------------------------------------------------
	// public common methods

	FIBITMAP* get_active_source() { return active_source_; }

	// image file i/o
	bool read_source(int _type);
	bool assign_source(FIBITMAP *fib);
	bool read_blank(int _w, int _h);	
	bool write_reference(const char* const _name);
	bool write_final(const char* const _name);

	// image processing
		
	// processes the image into a common display depth (24bpp) and output dimensions for chosen display mode
	bool sanitise_source();	

	// converts the source image from 24bpp into display format
	bool convert();	
	bool convert_cc();	
	void perform_diagnostics(byte_t** _bitmaps);
	void diagnostic_pass(byte_t** _bitmaps, DiagnosticUnit* _du);
	void generate_field_image(byte_t** _bitmaps, int _f);
	void generate_final_image(byte_t** _bitmaps);

	void init_dither_unit(int _field);
	void init_interlace_unit(int _field);
	void init_error_unit(int _colourdepth);
	void init_reduce_unit(int _field);
	void init_diagnostic_units();
		
	// final step, formats & compresses display format to file
	bool generate_pcs_file(const char* const _name);


	// --------------------------------------------------------------------
	// public virtual methods (for specialization)
	
	// reduce source image to custom display format
	virtual bool transform_source_to_target();

	static FIBITMAP* read_image(const char* const _name, int _type);

protected:
	
	DitherUnit* ditherunit_;
	InterlaceUnit* laceunit_;
	ErrorUnit* errorunit_;
	ReduceUnit* reduceunit_;
	std::vector<DiagnosticUnit*> diagnosticunits_;

	// --------------------------------------------------------------------
	// private virtual methods (for specialization)
	
	virtual void reduce_frame
	(
		byte_t* _bitmap, 
		word_t* _palette, 
		int _index, 
		int _num_indices
	);

	// --------------------------------------------------------------------
	// private common methods

	//datasize_t create_pcs_format(bool _flick, bool _ste_palette, bool _60hz);	
	datasize_t create_pcs_format(int _bitmapfields, bool _sharedbitmap, bool _ste_palette, bool _60hz);
	datasize_t create_pcr_format(int _bitmapfields, bool _sharedbitmap, bool _interleaved);

	// bitmap only
	void c2p(byte_t* _cin, word_t* _pout);
	void reformat_bitmap(word_t* _src, word_t* _dst);
	datasize_t compress_bitmap(word_t* _src, unsigned char* _dst, bool _verify = false);
	datasize_t decompress_bitmap(byte_t* _src, word_t* _dst, bool _verify = false);

	// palette only
	void palette_swizzle(word_t* _dst, datasize_t _count);
	void palette_deswizzle(word_t* _dst, datasize_t _count);
	datasize_t compress_palette(word_t* _src, unsigned char* _dst, datasize_t palettefile_words, bool _verify = false);
	datasize_t decompress_palette(byte_t* _src, word_t* _dst, datasize_t palettefile_words, bool _verify = false);

	// bitmap & palette
	void clear(byte_t* _dst, datasize_t _count);
	void clear(word_t* _dst, datasize_t _count);
	void make_delta(word_t* _src1, word_t* _src2, word_t* _delta, datasize_t _words);

	// other
	void generate_perception_mask();

	FIBITMAP* generate_test_image(int _index, signed short _w, signed short _h);

	// --------------------------------------------------------------------
	// simple LCG dither
	
//	inline unsigned short dither(unsigned long& dstate)
//	{
//		dstate = (dstate * 19387) + 27791;
//		return (dstate >> 15) & 255;
//	}

	// --------------------------------------------------------------------
	// private virtual methods (for specialization)
		
};

#endif // pcsimage_h_
