// ---------------------------------------------------------------------------------------------------------------------
//	PhotoChrome v6.xx - Atari ST interlaced image reduction framework
//	1991-2016 Douglas Little (d.m.l)
// ---------------------------------------------------------------------------------------------------------------------
//	special thanks to:
//		evil@DHS, RA/PDX, Troed/SYNC for help testing and input on improvements
// ---------------------------------------------------------------------------------------------------------------------

// ====================================================================
// 	image colour reduction unit
// ====================================================================

#include "common.h"
#include "config.h"

#if defined(__ATARI__) || defined(__MINGW32__)
#include "FreeImage.3.15.3.h"
#else
#include "FreeImage.3.15.4.h"
#endif

#include "imagepkg.h"
#include "pcsimage.h"

#include <math.h>

#include <iostream>
#include <map>
#include <assert.h>

#include "quantize_som.h"

extern Arguments g_allargs;

// --------------------------------------------------------------------

inline int Quantize_SOM::match_colour(signed short sr, signed short sg, signed short sb)
{				
	// look for match in visible palette(s)
	int best_index = 0;
	signed long best_error = 0x7FFFFFFF;
	for (int p = (target_colours_ - 1); p >= 0; --p)
	{
		{
			RGB8 &col = palette_rgb_[p];
						
			// check allocated status - top bits signify palette entry has been claimed
//			if (col.a)
			{
				// separate colour channels	
				signed short pr = (unsigned short)col.r;
				signed short pg = (unsigned short)col.g;
				signed short pb = (unsigned short)col.b;
							
				// calculate error 
				signed short rerr = pr - sr;
				signed short gerr = pg - sg;
				signed short berr = pb - sb;

				signed long errorsum = (rerr * rerr) + (gerr * gerr) + (berr * berr);
				
				// track smallest error
				if (errorsum < best_error)
				{
					best_error = errorsum;
					best_index = p;

					// can stop if we have exact match
					if (errorsum == 0)
						break;
				}
			}
		}
	}
	
	return best_index;
}

// --------------------------------------------------------------------
// colour reduction/allocation method 5 (self organising network / map)
// --------------------------------------------------------------------
// rules:
// 	1) create link from every pixel to every neuron representing a potential palette entry for that pixel
//	2) train neurons against pixels they can see, over N iterations
//	3) convert each neuron to a single colour
//	4) remap image to nearest matching colour
// notes:
//	* typically 3 palettes per scan, so 320/3 = ~100 pixels visible to each palette (maximum of 172, by measurement)
//	* ~48 palette bins per scan, so 48*100 = ~5000 colour records per scan
// --------------------------------------------------------------------

Quantize_SOM::Quantize_SOM(FIBITMAP *_image, int _target_colours)
	: image_(_image),
	  target_colours_(_target_colours),
	  pixel_records_(0), num_pixel_records_(0),
	  palette_rgb_(0),
	  neurons_(0), num_neurons_(0),
	  histogram_(0)
{			
	// create pixel and colour association buffers
	palette_rgb_ = new RGB8[_target_colours];
	
	num_neurons_ = _target_colours;
	neurons_ = new neuron[num_neurons_];
	if (!neurons_) { printf("error - could not allocate memory for colour neurons!\n"); getchar(); exit(1); }

	ImagePkg::instance().get_size(_image, w_, h_);	

	//num_pixel_records_ = w_ * h_;
	//printf("pixels = %d\n", num_pixel_records_);
	//pixel_records_ = new pixel_record[num_pixel_records_];
	//if (!pixel_records_) { printf("error - could not allocate memory for pixel records!\n"); getchar(); exit(1); }	

	histogram_ = new int[64*64*64];
}	

Quantize_SOM::~Quantize_SOM()
{
	delete[] palette_rgb_;
	delete[] pixel_records_;
	delete[] neurons_;
	delete[] histogram_;
}

// --------------------------------------------------------------------

void Quantize_SOM::generate_colours()
{
	if (g_allargs.verbose)
		printf("quantizing image to %d colours...\n", target_colours_);

	generate_colours_pass1_associate();
	generate_colours_pass2_train();
	generate_colours_pass3_merge();
}

// --------------------------------------------------------------------

void Quantize_SOM::init()
{
	memset(palette_rgb_, 0, sizeof(RGB8) * target_colours_);
	memset(neurons_, 0, sizeof(neuron) * num_neurons_);
//	memset(pixel_records_, 0, sizeof(pixel_record) * num_pixel_records_);
	memset(histogram_, 0, sizeof(int) * 64*64*64);
}

inline void Quantize_SOM::init_exp_solutions(arg_real _radius)
{
	for (int i = 0; i < max_neuron_references; ++i)
	{
		calc_real rho = 1.0f;
		if (i > 0)
			rho = exp(-1.0 * 4.0 * (calc_real(i) / _radius));
		exp_solutions_[i] = rho;
	}
}
	
// --------------------------------------------------------------------

bool Quantize_SOM::analyse()
{		
	int progress = 0;
	int last_progress_pct = -1;

	init();

	ImagePkg& imgpkg = ImagePkg::instance();	
		
	// pass #1 - create palette bins with pixel references, and pixel records with bin references
	
	num_unique_source_cols_ = 0;

	for (int y = 0; y < h_; ++y)
	{	
		for (int x = 0; x < w_; ++x)
		{
			// extract pixel
			byte_t ur, ug, ub;
			imgpkg.get_rgb(image_, x, y, h_, ur, ug, ub);
			signed short srn = ((signed short)(unsigned short)ur);
			signed short sgn = ((signed short)(unsigned short)ug);
			signed short sbn = ((signed short)(unsigned short)ub);

			// pack key
			int key = 
				((int)(ur & 0xFC) << (12-2)) |
				((int)(ug & 0xFC) << (6-2)) |
				((int)(ub & 0xFC) >> 2);

			// count refs
			if (histogram_[key] == 0)
				num_unique_source_cols_++;

			histogram_[key]++;
		}
	}

	if (g_allargs.verbose)
		printf("counted %d source colour references...\n", num_unique_source_cols_);

	return (can_reduce());
}

void Quantize_SOM::generate_colours_pass1_associate()
{		
	int progress = 0;
	int last_progress_pct = -1;
		
	ImagePkg& imgpkg = ImagePkg::instance();	

	if (can_reduce())
	{
		delete[] pixel_records_;
		num_pixel_records_ = num_unique_source_cols_;
		pixel_records_ = new pixel_record[num_pixel_records_];
		memset(pixel_records_, 0, sizeof(pixel_record) * num_pixel_records_);

		// process all used keys
		int refs = 0;
		for (int key = 0; key < (64*64*64); key++)
		{
			if (histogram_[key])
			{
				pixel_record& record = pixel_records_[refs++];	

				// decode key
				signed short srn = (key >> (12-2)) & 0xFC;
				signed short sgn = (key >> (6-2)) & 0xFC;
				signed short sbn = (key << 2) & 0xFC;

				// into unit space, combining lace/dither adjustments - this is the intended colour
				calc_real fsx = srn / 255.0f;//col2f(srn);
				calc_real fsy = sgn / 255.0f;//col2f(sgn);
				calc_real fsz = sbn / 255.0f;//col2f(sbn);
					
				// final intended colour
				record.ix = fsx;
				record.iy = fsy;
				record.iz = fsz;
				record.weight = (float)histogram_[key];
										
				// insert reference to this pixel into every visible palette index
				for (int p = num_neurons_-1; p >= 0; --p)
				{
					int ineuron = p;

					calc_real fp = (calc_real)p / (calc_real)num_neurons_;

					neuron& rneuron = neurons_[ineuron];
					// assign bin index/name
					rneuron.index = p;
					
					// init neuron
					rneuron.wx = fp;
					rneuron.wy = fp;
					rneuron.wz = fp;					
					//rneuron.wx = exposure_twister_.genFloat();
					//rneuron.wy = exposure_twister_.genFloat();
					//rneuron.wz = exposure_twister_.genFloat();

					rneuron.vx = 0.0f;
					rneuron.vy = 0.0f;
					rneuron.vz = 0.0f;
					
					// add bin reference to pixel record, for reduction later
					neuron_reference& rnref = record.neuron_references[record.num_neuron_references++];
					rnref.ineuron = ineuron;
				}	
			}
		}
	}

if (0)
{
	for (int y = 0; y < h_; ++y)
	{	
		// linear left-to-right ordering
		for (int x = 0; x < w_; ++x)
		{							
			unsigned long pixel_index = (y * w_) + x;

			pixel_record& record = pixel_records_[pixel_index];	


			// extract pixel
			byte_t ur, ug, ub;
			imgpkg.get_rgb(image_, x, y, h_, ur, ug, ub);
			signed short srn = ((signed short)(unsigned short)ur);
			signed short sgn = ((signed short)(unsigned short)ug);
			signed short sbn = ((signed short)(unsigned short)ub);
		
			// into unit space, combining lace/dither adjustments - this is the intended colour
			calc_real fsx = srn / 255.0f;//col2f(srn);
			calc_real fsy = sgn / 255.0f;//col2f(sgn);
			calc_real fsz = sbn / 255.0f;//col2f(sbn);
				
			// final intended colour
			record.ix = fsx;
			record.iy = fsy;
			record.iz = fsz;
			record.weight = 1.0f;
									
			// insert reference to this pixel into every visible palette index
			for (int p = num_neurons_-1; p >= 0; --p)
			{
				{
					int ineuron = p;

					calc_real fp = (calc_real)p / (calc_real)num_neurons_;

					neuron& rneuron = neurons_[ineuron];
					// assign bin index/name
					rneuron.index = p;
					
					// init neuron
					rneuron.wx = fp;
					rneuron.wy = fp;
					rneuron.wz = fp;					
					//rneuron.wx = exposure_twister_.genFloat();
					//rneuron.wy = exposure_twister_.genFloat();
					//rneuron.wz = exposure_twister_.genFloat();

					rneuron.vx = 0.0f;
					rneuron.vy = 0.0f;
					rneuron.vz = 0.0f;
					
					// add bin reference to pixel record, for reduction later
					neuron_reference& rnref = record.neuron_references[record.num_neuron_references++];
					rnref.ineuron = ineuron;
				}
			}			
		}
		
		#ifdef __ATARI__
		// update progress
		if (!g_allargs.quiet)
		{
			int progress_pct = ((int)y * 100) / h_;
			if (progress_pct != last_progress_pct)
			{
				last_progress_pct = progress_pct;		
				switch (progress & 3)
				{
				case 0:
					printf("- %2d%%\r", progress_pct); break;
				case 1:
					printf("\\ %2d%%\r", progress_pct); break;
				case 2:
					printf("| %2d%%\r", progress_pct); break;
				case 3:
					printf("/ %2d%%\r", progress_pct); break;
				};
				fflush(stdout);
				progress++;
			}
		}
		#endif
	}
}

}

inline void Quantize_SOM::present(pixel_record& rrec /*, arg_real alpha, arg_real radius*/)
{
	// find winning *visible* neuron via references
	
	calc_real win_d = 1e6f;
	int win_ni = 0;
	
	for (int ni = 0; ni < rrec.num_neuron_references; ++ni)
	{
		// find neuron for this reference
		const neuron_reference& rnref = rrec.neuron_references[ni];

		int ineuron = rnref.ineuron;

		neuron& rneuron = neurons_[ineuron];
		
		// find biased distance
		calc_real dx = (rrec.ix - rneuron.wx);
		calc_real dy = (rrec.iy - rneuron.wy);
		calc_real dz = (rrec.iz - rneuron.wz);
		calc_real d = (dx * dx) + (dy * dy) + (dz * dz);
		
		if (d < win_d)
		{
			win_d = d;
			win_ni = ni;
		}
	}

	// calculate neighbour influence 
	for (int ni = 0; ni < rrec.num_neuron_references; ++ni)
	{
		// find neuron for this reference
		neuron_reference& rnref = rrec.neuron_references[ni];

		int ineuron = rnref.ineuron;

		neuron& rneuron = neurons_[ineuron];

		int i = xabs(ni - win_ni);
		assert(i < max_neuron_references);
		rnref.rho = exp_solutions_[i];
	}	
}

#include <list>


namespace
{
	mersenne_twister g_twister;
	inline unsigned int grand() { return g_twister.genU32(); }
	ptrdiff_t myrandom (ptrdiff_t i) { return grand() % i;}
	ptrdiff_t (*p_myrandom)(ptrdiff_t) = myrandom;
}

void Quantize_SOM::generate_colours_pass2_train()
{
	int xxx = 0;
		
	int sampling_breadth = xftoi_floor(float(num_pixel_records_) * g_allargs.nn_breadth);

	std::vector<signed long> sequence;		
	for (int i = 0; i < num_pixel_records_; ++i)
		sequence.push_back(i);
	g_twister.reset(exposure_twister_.genU32());
				
	float learning_rate = g_allargs.nn_rate;
	float learning_radius = g_allargs.nn_radius;
	
	int total_cycles = 1 << g_allargs.nn_depth;
	for (int cycle = 0; cycle < total_cycles; ++cycle)
	{		
		std::random_shuffle(sequence.begin(), sequence.end(), p_myrandom);
			
		// calculate alpha
		calc_real alpha = learning_rate * (0.5 * exp(-1.0 * 1.5 * (calc_real(cycle) / calc_real(total_cycles))));
		calc_real radius = learning_radius * exp(-1.0 * 5.0 * (calc_real(cycle) / calc_real(total_cycles)));
		init_exp_solutions(radius);
		
		if (--xxx <= 0)
		{
			xxx = 10;
			printf("\riteration:%5d/%5d, radius(%.3f), alpha(%.6f), ", 
				cycle, total_cycles, 
				(double)radius, 
				alpha);
				
			generate_colours_pass3_merge();
			calc_real ee = score();
			printf("sqerr:%.2f     ", (double)ee);
			fflush(stdout);			
		}
	
		// present each pixel to each relevant neuron
		signed long* psequence = &sequence[0];
		for (int bi = 0; bi < sampling_breadth; ++bi)
		{
			//signed long irec = *(sit++);
			signed long irec = psequence[bi];
			pixel_record& rrec = pixel_records_[irec];
			present(rrec /*, alpha, radius*/);	
		}
		
		// update neurons via references
		for (int irec = 0; irec < num_pixel_records_; ++irec)
		{
			pixel_record& rrec = pixel_records_[irec];			
					
			for (int ni = 0; ni < rrec.num_neuron_references; ++ni)
			{
				// find neuron for this reference
				neuron_reference& rnref = rrec.neuron_references[ni];

				int ineuron = rnref.ineuron;

				neuron& rneuron = neurons_[ineuron];
								
				// train neuron
				calc_real t = rnref.rho * alpha * rrec.weight;

				rneuron.vx += (rrec.ix - rneuron.wx) * t;
				rneuron.vy += (rrec.iy - rneuron.wy) * t;
				rneuron.vz += (rrec.iz - rneuron.wz) * t;

				rneuron.wx += rneuron.vx;
				rneuron.wy += rneuron.vy;
				rneuron.wz += rneuron.vz;
				
				rneuron.vx *= 0.95f;
				rneuron.vy *= 0.95f;
				rneuron.vz *= 0.95f;
			}	
		}
	}
	printf("\n");	
}

void Quantize_SOM::generate_colours_pass3_merge()
{
	for (signed long ineuron = 0; ineuron < num_neurons_; ++ineuron)
	{
		neuron& rneuron = neurons_[ineuron];
		
		calc_real dcxm = rneuron.wx;
		calc_real dcym = rneuron.wy;
		calc_real dczm = rneuron.wz;

		// into palette space for this field
		signed short sr = (signed short)xclamp<calc_real>((dcxm * 255.0f), 0.0f, 255.0f);
		signed short sg = (signed short)xclamp<calc_real>((dcym * 255.0f), 0.0f, 255.0f);
		signed short sb = (signed short)xclamp<calc_real>((dczm * 255.0f), 0.0f, 255.0f);

		RGB8 &col = palette_rgb_[ineuron];
		col.r = (byte_t)sr;					
		col.g = (byte_t)sg;					
		col.b = (byte_t)sb;
		col.a = 1;
	}
}

// --------------------------------------------------------------------

void Quantize_SOM::remap_pixels()
{
	allocate_pixels_remap_();
}

// --------------------------------------------------------------------


void Quantize_SOM::allocate_pixels_remap_()
{	
	int progress = 0;
	int last_progress_pct = -1;

	ImagePkg& imgpkg = ImagePkg::instance();	

	for (int y = 0; y < h_; ++y)
	{
		for (int x = 0; x < w_; ++x)
		{	
			byte_t ur, ug, ub;
			imgpkg.get_rgb(image_, x, y, h_, ur, ug, ub);
		
			// into unit space, combining spatial error adjustment
			signed short sr = (signed short)(unsigned short)ur;
			signed short sg = (signed short)(unsigned short)ug;
			signed short sb = (signed short)(unsigned short)ub;

			int chosen_index = match_colour(sr, sg, sb);

			RGB8 &col = palette_rgb_[chosen_index];

			imgpkg.put_rgb(image_, x, y, h_, col.r, col.g, col.b);
		}
		
		#ifdef __ATARI__
		// update progress
		if (!g_allargs.quiet)
		{
			int progress_pct = ((int)y * 100) / h_;
			if (progress_pct != last_progress_pct)
			{
				last_progress_pct = progress_pct;		
				switch (progress & 3)
				{
				case 0:
					printf("- %2d%%\r", progress_pct); break;
				case 1:
					printf("\\ %2d%%\r", progress_pct); break;
				case 2:
					printf("| %2d%%\r", progress_pct); break;
				case 3:
					printf("/ %2d%%\r", progress_pct); break;
				};
				fflush(stdout);
				progress++;
			}
		}
		#endif
		
	}
}


calc_real Quantize_SOM::score()
{	
	int progress = 0;
	int last_progress_pct = -1;

	calc_real ee = 0.0f;

	for (int sample = 0; sample < num_pixel_records_; sample++)
	{
		pixel_record& rrec = pixel_records_[sample];			
	
		if (rrec.num_neuron_references <= 0)
			continue;

		// into unit space, combining spatial error adjustment
		calc_real fsx = rrec.ix;
		calc_real fsy = rrec.iy;
		calc_real fsz = rrec.iz;
								
		// into palette space
		signed short sr = (signed short)xclamp<calc_real>((fsx * 255.0f), 0.0f, 255.0f);
		signed short sg = (signed short)xclamp<calc_real>((fsy * 255.0f), 0.0f, 255.0f);
		signed short sb = (signed short)xclamp<calc_real>((fsz * 255.0f), 0.0f, 255.0f);					

		int chosen_index = match_colour(sr, sg, sb);

		{
			RGB8 &col = palette_rgb_[chosen_index];
		
			// nearest palette colour
			signed short pr = col.r;
			signed short pg = col.g;
			signed short pb = col.b;
		
			// into unit space
			calc_real fpr = (calc_real)pr / 255.0f;
			calc_real fpg = (calc_real)pg / 255.0f;
			calc_real fpb = (calc_real)pb / 255.0f;
		
			// calculate error		
			calc_real fedr = (fsx - fpr);
			calc_real fedg = (fsy - fpg);
			calc_real fedb = (fsz - fpb);
													
			ee += (fedr * fedr) + (fedg * fedg) + (fedb * fedb);				
		}		
	}

	return ee;
}

// --------------------------------------------------------------------
