// ---------------------------------------------------------------------------------------------------------------------
//	PhotoChrome v6.xx - Atari ST interlaced image reduction framework
//	1991-2016 Douglas Little (d.m.l)
// ---------------------------------------------------------------------------------------------------------------------
//	special thanks to:
//		evil@DHS, RA/PDX, Troed/SYNC for help testing and input on improvements
// ---------------------------------------------------------------------------------------------------------------------

// ====================================================================
//	error diffusion unit
// ====================================================================

#include "common.h"
#include "config.h"

#if defined(__ATARI__) || defined(__MINGW32__)
#include "FreeImage.3.15.3.h"
#else
#include "FreeImage.3.15.4.h"
#endif

#include "imagepkg.h"
#include "pcsimage.h"

extern Arguments g_allargs;

// --------------------------------------------------------------------

// --------------------------------------------------------------------

ReduceUnit::ReduceUnit(PCSImage* _host, int _index)
	: LCG(0),
	  //
	  // state for all jobs
	  //
	  imgpkg_(ImagePkg::instance()),	
	  host_(_host),
	  //
	  active_source_(_host->active_source_),
	  cc_field0_(_host->cc_field0_),
	  cc_field1_(_host->cc_field1_),
	  mask_(_host->mask_),
	  display_width_(_host->display_width_),
	  display_height_(_host->display_height_),
	  fields_(_host->fields_),
	  palettebuffer_stride_(_host->palettebuffer_stride_),
	  //
	  colourdepth_(_host->colourdepth_),
	  r_pos_(_host->r_pos_),
	  g_pos_(_host->g_pos_),
	  b_pos_(_host->b_pos_),
	  c_mask_(_host->c_mask_),
	  //
	  laceunit_(_host->laceunit_),
	  ditherunit_(_host->ditherunit_),
	  errorunit_(_host->errorunit_),
	  error_iteration_(_host->error_iteration_),
	  //
	  // state for this job
	  //
	  index_(_index),
	  bitmap_(0),
	  palette_(0),
	  PCT_(0),
	  DPT_(0)
{
}

ReduceUnit::~ReduceUnit()
{
}

void ReduceUnit::reduce
(
	byte_t* _bitmap, 
	word_t* _palette,
	byte_t* _PCT, word_t* _DPT
)
{
	// set state
	bitmap_ = _bitmap;
	palette_ = _palette;
	PCT_ = _PCT;
	DPT_ = _DPT;
	
	// act
	
	if (laceunit_)
		laceunit_->reset();
	if (ditherunit_)
		ditherunit_->reset();
	
	generate_colours();
	
	if (laceunit_)
		laceunit_->reset();
	if (ditherunit_)
		ditherunit_->reset();
	
	allocate_pixels();
}

// --------------------------------------------------------------------
// common field error feedback step
// --------------------------------------------------------------------

arg_real ReduceUnit::feedback_field_error(int x, int scan, byte_t chosen_index, arg_real fsx, arg_real fsy, arg_real fsz)
{
	// get assigned colour
	
	// index correct palette for this x & p
	byte_t* sync_mapping = ADDRESS_PCT_SCAN(PCT_, x, display_width_, scan);
	
	byte_t upi = sync_mapping[chosen_index];
	signed short spi = (signed short)((signed char)upi);		
	word_t* scan_palette = &palette_[scan * palettebuffer_stride_];
	word_t* x_palette = &scan_palette[spi * g_allargs.colours];

	// find colour (0x0RGB)
	word_t pc = x_palette[chosen_index];

	// palette colour
	signed short pr = ((pc >> (r_pos_ - colourdepth_)) & c_mask_);
	signed short pg = ((pc >> (g_pos_ - colourdepth_)) & c_mask_);
	signed short pb = ((pc >> (b_pos_ - colourdepth_)) & c_mask_);

	// into unit space
	const float fpn = (calc_real)pal_crange(colourdepth_);
	calc_real fpr = range2f(pr, fpn);
	calc_real fpg = range2f(pg, fpn);
	calc_real fpb = range2f(pb, fpn);	

	// quantize in native virtual space
	//fsx = range2f(f2range(fsx, (float)nvi_crange(colourdepth_, fields_)), (float)nvi_crange(colourdepth_, fields_));
	//fsy = range2f(f2range(fsy, (float)nvi_crange(colourdepth_, fields_)), (float)nvi_crange(colourdepth_, fields_));
	//fsz = range2f(f2range(fsz, (float)nvi_crange(colourdepth_, fields_)), (float)nvi_crange(colourdepth_, fields_));

	// calculate error
	calc_real fedr = fsx - fpr;
	calc_real fedg = fsy - fpg;
	calc_real fedb = fsz - fpb;
				
	// feedback field error				
	errorunit_->feedback_field(x, scan, fedr, fedg, fedb);

	return (fedr * fedr) + (fedg * fedg) + (fedb * fedb);
}

arg_real ReduceUnit::feedback_error(int x, int scan, byte_t chosen_index, arg_real fsx, arg_real fsy, arg_real fsz)
{
	// get assigned colour
	
	// index correct palette for this x & p
	byte_t* sync_mapping = ADDRESS_PCT_SCAN(PCT_, x, display_width_, scan);

	byte_t upi = sync_mapping[chosen_index];
	signed short spi = (signed short)((signed char)upi);		
	word_t* scan_palette = &palette_[scan * palettebuffer_stride_];
	word_t* x_palette = &scan_palette[spi * g_allargs.colours];

	// find colour (0x0RGB)
	word_t pc = x_palette[chosen_index];

	// palette colour
	signed short pr = ((pc >> (r_pos_ - colourdepth_)) & c_mask_);
	signed short pg = ((pc >> (g_pos_ - colourdepth_)) & c_mask_);
	signed short pb = ((pc >> (b_pos_ - colourdepth_)) & c_mask_);

	// into unit space
	const calc_real fpn = (calc_real)pal_crange(colourdepth_);
	calc_real fpr = range2f(pr, fpn);
	calc_real fpg = range2f(pg, fpn);
	calc_real fpb = range2f(pb, fpn);			

	// quantize in native virtual space
	//fsx = range2f(f2range(fsx, (float)nvi_crange(colourdepth_, fields_)), (float)nvi_crange(colourdepth_, fields_));
	//fsy = range2f(f2range(fsy, (float)nvi_crange(colourdepth_, fields_)), (float)nvi_crange(colourdepth_, fields_));
	//fsz = range2f(f2range(fsz, (float)nvi_crange(colourdepth_, fields_)), (float)nvi_crange(colourdepth_, fields_));
	
	// calculate error
	calc_real fedr = fsx - fpr;
	calc_real fedg = fsy - fpg;
	calc_real fedb = fsz - fpb;
				
	// feedback field error				
	errorunit_->feedback(x, scan, fedr, fedg, fedb);

	return (fedr * fedr) + (fedg * fedg) + (fedb * fedb);
}

// --------------------------------------------------------------------

byte_t ReduceUnit::match_colour(int x, int scan, signed short sr, signed short sg, signed short sb)
{

	byte_t* sync_mapping = ADDRESS_PCT_SCAN(PCT_, x, display_width_, scan);

	word_t* scan_palette = &palette_[scan * palettebuffer_stride_];
				
	// look for match in visible palette(s)
	signed short best_index = 0;
	signed long best_error = 0x7FFFFFFF;
	for (signed short p = g_allargs.colours-1; p >= 0; --p)
	{
		word_t mask = 0;
		if (DPT_) 
			mask = DPT_[x + (scan*display_width_)];
					
		// avoid masked colours for this pixel
		if ((mask & (1 << p)) == 0)
		{
			// index correct palette for this x & p
			byte_t upi = sync_mapping[p];
			signed short spi = (signed short)((signed char)upi);		
			word_t* x_palette = &scan_palette[spi * g_allargs.colours];

			// find colour (0x0RGB)
			word_t pc = x_palette[p];
			
			// check allocated status - top bits signify palette entry has been claimed
			if ((pc & 0x8000) != 0)
			{
				// separate colour channels	
				signed short pr = (pc >> (r_pos_ - colourdepth_) & c_mask_);
				signed short pg = (pc >> (g_pos_ - colourdepth_) & c_mask_);
				signed short pb = (pc >> (b_pos_ - colourdepth_) & c_mask_);
														
				// calculate error 
				signed short rerr = pr - sr;
				signed short gerr = pg - sg;
				signed short berr = pb - sb;
				signed long errorsum = (rerr * rerr) + (gerr * gerr) + (berr * berr);
				
				// track smallest error
				if (errorsum < best_error)
				{
					best_error = errorsum;
					best_index = p;

					// can stop if we have exact match
					if (errorsum == 0)
						break;
				}
			}
		}
	}
	
	return (byte_t)best_index;
}

// --------------------------------------------------------------------

void PCSImage::init_reduce_unit(int _index)
{
	delete reduceunit_; reduceunit_ = 0;
	
	ReduceUnit* newunit = 0;	
	switch (g_allargs.method)
	{
	default:
	case ReduceMethod_SIMPLE:
		newunit = new ReduceUnit_0_SIMPLE(this, _index);
		break;
	case ReduceMethod_SCAN_SUBDIVISION:
		newunit = new ReduceUnit_1_SCAN_SUBDIVISION(this, _index);
		break;
	case ReduceMethod_LCG:
		newunit = new ReduceUnit_2_LCG(this, _index);
		break;
	case ReduceMethod_BIN_BALANCING_L:
		newunit = new ReduceUnit_3_BIN_BALANCER(this, _index, ReduceUnit_3_BIN_BALANCER::Quality_LOW);
		break;
	case ReduceMethod_SCAN_SUBDIVISION2:
		newunit = new ReduceUnit_4_SCAN_SUBDIVISION2(this, _index);
		break;
	case ReduceMethod_SOM:
		newunit = new ReduceUnit_5_SOM(this, _index);
		break;
	case ReduceMethod_BIN_BALANCING_H:
		newunit = new ReduceUnit_3_BIN_BALANCER(this, _index, ReduceUnit_3_BIN_BALANCER::Quality_HIGH);
		break;
	}
	
	reduceunit_ = newunit;
}

// --------------------------------------------------------------------
