// ---------------------------------------------------------------------------------------------------------------------
//	PhotoChrome v6.xx - Atari ST interlaced image reduction framework
//	1991-2016 Douglas Little (d.m.l)
// ---------------------------------------------------------------------------------------------------------------------
//	special thanks to:
//		evil@DHS, RA/PDX, Troed/SYNC for help testing and input on improvements
// ---------------------------------------------------------------------------------------------------------------------

#ifndef reduceunit_h_
#define reduceunit_h_

// ====================================================================
//  colour reduction unit
// ====================================================================

#include "lcg.h"

// --------------------------------------------------------------------
// object representing an error-diffusion implementation
// --------------------------------------------------------------------

class PCSImage;

class ReduceUnit : public LCG
{
protected:
	
	PCSImage* host_;
	FIBITMAP* active_source_;
	
	// cryptochrome specific
	FIBITMAP* cc_field0_;
	FIBITMAP* cc_field1_;
	
	byte_t* mask_;
	signed short display_width_;
	signed short display_height_;
	signed short fields_;
	signed short palettebuffer_stride_;
	int colourdepth_;
	signed short r_pos_;
	signed short g_pos_;
	signed short b_pos_;
	unsigned short c_mask_;
	InterlaceUnit* laceunit_;
	DitherUnit* ditherunit_;
	ErrorUnit* errorunit_;
	ImagePkg& imgpkg_;


	byte_t* bitmap_;
	word_t* palette_;
	byte_t* PCT_;
	word_t* DPT_;
	int index_;
	int error_iteration_;
			
	virtual void generate_colours() = 0;
	virtual void allocate_pixels() = 0;

	virtual arg_real feedback_error(int x, int scan, byte_t chosen_index, arg_real fsx, arg_real fsy, arg_real fsz);
	virtual arg_real feedback_field_error(int x, int scan, byte_t chosen_index, arg_real fsx, arg_real fsy, arg_real fsz);

	byte_t match_colour(int x, int scan, signed short sr, signed short sg, signed short sb);
				
public:
	
	virtual ~ReduceUnit();

	ReduceUnit(PCSImage* _host, int _index);
	
	void reduce
	(
		byte_t* _bitmap, 
		word_t* _palette,
		byte_t* _PCT, word_t* _DPT
	);
};

// --------------------------------------------------------------------
// direct pixel sampling allocators share a lot of common behaviour

class ReduceUnit_DIRECT_SAMPLING : public ReduceUnit
{
public:
	
	ReduceUnit_DIRECT_SAMPLING(PCSImage* _host, int _index)
		: ReduceUnit(_host, _index)
	{
	}	
	
	~ReduceUnit_DIRECT_SAMPLING()
	{
	}

	virtual void allocate_pixels();	
	
	byte_t allocate_colour(int x, int scan, signed short sr, signed short sg, signed short sb);
	void sample(int x, int scan);

};

// --------------------------------------------------------------------

class ReduceUnit_0_SIMPLE : public ReduceUnit_DIRECT_SAMPLING
{
public:
	
	ReduceUnit_0_SIMPLE(PCSImage* _host, int _index)
		: ReduceUnit_DIRECT_SAMPLING(_host, _index)
	{
	}

	virtual void generate_colours();	
};

// --------------------------------------------------------------------

class ReduceUnit_1_SCAN_SUBDIVISION : public ReduceUnit_DIRECT_SAMPLING
{
public:
	
	ReduceUnit_1_SCAN_SUBDIVISION(PCSImage* _host, int _index)
		: ReduceUnit_DIRECT_SAMPLING(_host, _index)
	{
	}

	virtual void generate_colours();	
};

// --------------------------------------------------------------------

class ReduceUnit_4_SCAN_SUBDIVISION2 : public ReduceUnit_DIRECT_SAMPLING
{
public:
	
	ReduceUnit_4_SCAN_SUBDIVISION2(PCSImage* _host, int _index)
		: ReduceUnit_DIRECT_SAMPLING(_host, _index)
	{
	}

	virtual void generate_colours();	
};

// --------------------------------------------------------------------

class ReduceUnit_2_LCG : public ReduceUnit_DIRECT_SAMPLING
{
public:
	
	ReduceUnit_2_LCG(PCSImage* _host, int _index)
		: ReduceUnit_DIRECT_SAMPLING(_host, _index)
	{
	}

	virtual void generate_colours();	
};

// --------------------------------------------------------------------
// direct pixel sampling allocators share a lot of common behaviour

class ReduceUnit_INDIRECT_ANALYTICAL : public ReduceUnit
{

public:
	
	ReduceUnit_INDIRECT_ANALYTICAL(PCSImage* _host, int _index)
		: ReduceUnit(_host, _index)
	{
		// create perception filter mask buffer
	
		int num_pixels = display_width_ * display_height_;
		fmask_ = new float[num_pixels];
		if (!fmask_) { printf("error - could not allocate memory for float mask!\n"); getchar(); exit(1); }
	}	
	
	~ReduceUnit_INDIRECT_ANALYTICAL()
	{
		delete[] fmask_;
	}

	virtual arg_real feedback_error(int x, int scan, byte_t chosen_index, arg_real fsx, arg_real fsy, arg_real fsz);
	virtual arg_real feedback_field_error(int x, int scan, byte_t chosen_index, arg_real fsx, arg_real fsy, arg_real fsz);	

protected:
	
	float* fmask_;
};

// --------------------------------------------------------------------

class ReduceUnit_3_BIN_BALANCER : public ReduceUnit_INDIRECT_ANALYTICAL
{
public:
	
	enum Quality
	{
		Quality_LOW,
		Quality_HIGH
	};

private:

	// --------------------------------------------------------------------
	// some specific structures for reduce method #3
	
	struct pixel_reference
	{
		unsigned short pixel_x;
		unsigned short pixel_y;	
		float error; 
	} __attribute__((packed));
	
	// for PCS, each bin can see up to 171 pixels. will vary for other display rules.
	// if set lower than the 'magic figure', there will be some impact on image fidelity, but each reference consumes memory...
#ifdef __ATARI__
	static const int max_pixel_references = 172;	// keeps memory consumption sensible - but may be too low for some custom display formats
#else
	static const int max_pixel_references = 1024;	//416;	//200;
#endif	
	static const int max_bin_references = MAX_PALETTE_SIZE;	
	
	struct palette_bin
	{
		signed short num_pixel_references;
		pixel_reference pix_references[max_pixel_references];
		byte_t index;
	} __attribute__((packed));
	
	struct bin_reference
	{
#if defined(__ATARI__)
		signed short bin;
#else
		signed long bin;	// needed for images > 512 lines high
#endif
		//float error; 
	} __attribute__((packed));
	
	struct pixel_record
	{
		s_real cx;	// cached colour vector - defined as floats for CIE experiments - 8bit RGB would do otherwise. 
		s_real cy;	// will probably change this again after more CIE experiments, or split the impl.
		s_real cz;	// note: performing indirect re-lookup/transform each time during solver eliminates these - but is a way too slow for CIE & other spaces
		s_real ix;
		s_real iy;
		s_real iz;
		signed short num_bin_references;
		bin_reference bin_references[max_bin_references];
	} __attribute__((packed));


	unsigned long num_bins_;
	unsigned long num_records_;
	palette_bin* palette_bins_;
	pixel_record* pixel_records_;

	virtual void generate_colours();
	void generate_colours_init();
	void generate_colours_pass1_associate();
	void generate_colours_pass2_reduce();
	void generate_colours_pass2a_score();
	bool generate_colours_pass2b_simplify();
	void generate_colours_pass3_merge();

	virtual void allocate_pixels();
	void allocate_pixels_stable_();
	void allocate_pixels_remap_();
	
	void update(signed long ibin);

	Quality quality_;

public:
		
	ReduceUnit_3_BIN_BALANCER(PCSImage* _host, int _index, ReduceUnit_3_BIN_BALANCER::Quality _quality);
	
	~ReduceUnit_3_BIN_BALANCER();
	
};

class ReduceUnit_5_SOM : public ReduceUnit_INDIRECT_ANALYTICAL
{
private:
	
	// --------------------------------------------------------------------
	// some specific structures for reduce method #5
	struct pixel_reference
	{
		unsigned short pixel_x;
		unsigned short pixel_y;
#if defined(__ATARI__)
		signed short inref;
#else
		signed long inref;
#endif
	} __attribute__((packed));
	
	// for PCS, each bin can see up to 171 pixels. will vary for other display rules.
	// if set lower than the 'magic figure', there will be some impact on image fidelity, but each reference consumes memory...
#ifdef __ATARI__
	static const int max_pixel_references = 172;	// keeps memory consumption sensible - but may be too low for some custom display formats
#else
	static const int max_pixel_references = 1024;
#endif	
	
	static const int max_neuron_references = MAX_PALETTE_SIZE;
	
	struct neuron
	{
		signed short num_pixel_references;
		pixel_reference pix_references[max_pixel_references];
		byte_t index;
		storage_real wx;
		storage_real wy;
		storage_real wz;
		storage_real vx;
		storage_real vy;
		storage_real vz;
	} __attribute__((packed));
	
	struct neuron_reference
	{
#if defined(__ATARI__)
		unsigned short ineuron;
#else
		unsigned long ineuron;
#endif
		storage_real rho;
	} __attribute__((packed));
	
	struct pixel_record
	{
		s_real cx;	// cached colour vector - defined as floats for CIE experiments - 8bit RGB would do otherwise. 
		s_real cy;	// will probably change this again after more CIE experiments, or split the impl.
		s_real cz;	// note: performing indirect re-lookup/transform each time during solver eliminates these - but is a way too slow for CIE & other spaces
		s_real ix;
		s_real iy;
		s_real iz;
		unsigned int cycle;
		signed short num_neuron_references;
		neuron_reference neuron_references[max_neuron_references];
	} __attribute__((packed));

	unsigned long num_neurons_;
	neuron* neurons_;
	unsigned long num_pixel_records_;
	pixel_record* pixel_records_;
	mersenne_twister exposure_twister_;

	float exp_solutions_[max_neuron_references];
	void init_exp_solutions(arg_real _radius);

	virtual void generate_colours();
	void generate_colours_init();
	void generate_colours_pass1_associate();
	void generate_colours_pass2_train();
	void generate_colours_pass3_merge();

	virtual void allocate_pixels();
	void allocate_pixels_remap_();
	calc_real measure_pixels();

	void present(pixel_record& rrec, arg_real alpha, arg_real radius);
	
public:
	
	ReduceUnit_5_SOM(PCSImage* _host, int _index);
	
	~ReduceUnit_5_SOM();
	
};
// --------------------------------------------------------------------

#endif // errorunit_h_
