// ---------------------------------------------------------------------------------------------------------------------
//	PhotoChrome v6.xx - Atari ST interlaced image reduction framework
//	1991-2016 Douglas Little (d.m.l)
// ---------------------------------------------------------------------------------------------------------------------
//	special thanks to:
//		evil@DHS, RA/PDX, Troed/SYNC for help testing and input on improvements
// ---------------------------------------------------------------------------------------------------------------------

// ====================================================================
// 	image colour reduction unit
// ====================================================================

#include "common.h"
#include "config.h"

#if defined(__ATARI__) || defined(__MINGW32__)
#include "FreeImage.3.15.3.h"
#else
#include "FreeImage.3.15.4.h"
#endif

#include "imagepkg.h"
#include "pcsimage.h"

#include <math.h>

#include <iostream>
#include <map>
#include <assert.h>

extern Arguments g_allargs;

// --------------------------------------------------------------------
// colour reduction/allocation method 3 (global bin rebalancing)
// --------------------------------------------------------------------
// rules:
//	0) [convert pixels to CIELAB]
// 	1) insert reference to every pixel into every palette entry visible to that pixel
//	2) score every pixel reference relative to distance (error) from others in bin
//	3) remove (N > 0) references belonging to each pixel, with the highest error
//	4) repeat from step 2) until 1 reference remains per pixel
//	5) merge each bin down to a single colour
//	6) [convert back to RGB] & allocate palette entry
// notes:
//	* typically 3 palettes per scan, so 320/3 = ~100 pixels visible to each palette (maximum of 171, by measurement)
//	* ~48 palette bins per scan, so 48*100 = ~5000 colour records per scan
//	* ~1.5M colour records per image, with each record requiring: (C + I + E + errorscore) minimum
//	* solution must be global/simultaneous, so can't solve scan-by-scan to save memory
//	* perhaps convert to CIE on demand, storing only (pixel index + error = 4 or 8 bytes depending) per record
//	* abandoned CIE for now - results still not as good as RGB for the extra cost, will try again later
// --------------------------------------------------------------------

ReduceUnit_3_BIN_BALANCER::ReduceUnit_3_BIN_BALANCER(PCSImage* _host, int _index, ReduceUnit_3_BIN_BALANCER::Quality _quality)
	: ReduceUnit_INDIRECT_ANALYTICAL(_host, _index),
	  quality_(_quality)
{			
	// create pixel and colour association buffers
	
	num_bins_ = (palettebuffer_stride_ * display_height_);
	
	printf("ps: %d, nb: %d\n", palettebuffer_stride_, num_bins_);
	palette_bins_ = new palette_bin[num_bins_];
	if (!palette_bins_) { printf("error - could not allocate memory for colour reduction bins!\n"); getchar(); exit(1); }

	num_records_ = display_width_ * display_height_;
	pixel_records_ = new pixel_record[num_records_];
	if (!pixel_records_) { printf("error - could not allocate memory for pixel records!\n"); getchar(); exit(1); }	
}	

ReduceUnit_3_BIN_BALANCER::~ReduceUnit_3_BIN_BALANCER()
{
	delete[] pixel_records_;
	delete[] palette_bins_;
}

// --------------------------------------------------------------------

void ReduceUnit_3_BIN_BALANCER::generate_colours()
{
	generate_colours_init();
	generate_colours_pass1_associate();

#ifdef _DEBUG
	printf("\n");	

	// dump the pixel reference count per bin
	int last_pal = 0;
	int last_scan = 0;
	for (signed long ibin = 0; ibin < num_bins_; ++ibin)
	{
		palette_bin& rbin = palette_bins_[ibin];
		
		int scan = (ibin / palettebuffer_stride_);
		int scani = ibin - (scan * palettebuffer_stride_);
		int pal = scani / g_allargs.colours;
		int col = scani - (pal * g_allargs.colours);

		if (scan != last_scan)
		{
			printf("\n");			
		}
		else if (pal != last_pal)
		{
			printf("  ");			
		}
		
		printf("%2x ", rbin.num_pixel_references);
		
		last_scan = scan;
		last_pal = pal;
	}	

	printf("\n");	
#endif
	
	generate_colours_pass2_reduce();
	generate_colours_pass3_merge();
}

// --------------------------------------------------------------------

void ReduceUnit_3_BIN_BALANCER::generate_colours_init()
{
	if (g_allargs.verbose)
		printf("reducing field %d of %d / method %d (balancing solver / Q:%s)...\n", 
			index_, fields_,
			(quality_ == Quality_HIGH) ? 6 : 3, 
			(quality_ == Quality_HIGH) ? "HIGH":"LOW"
		);

	memset(palette_bins_, 0, sizeof(palette_bin) * num_bins_);
	memset(pixel_records_, 0, sizeof(pixel_record) * num_records_);

	// prepare perception filter mask

	calc_real pfm_power_curve = g_allargs.pfm_gamma;
	calc_real pfm_influence = g_allargs.pfm_influence;

	memset(fmask_, 0, sizeof(float) * num_records_);

	for (signed long mi = 0; mi < num_records_; ++mi)
	{
		// look up perception filter mask (range: 1.0 -> 0.0)
		calc_real perception_filter_mask = (calc_real)((unsigned int)(mask_[mi])) / 255.0f;
		
		// apply gamma curve
		perception_filter_mask = xclamp<calc_real>(pow(perception_filter_mask, pfm_power_curve), 0.0, 1.0);
		
		// attenuate influence
		perception_filter_mask = inverse_attenuation<calc_real>(perception_filter_mask, pfm_influence);
		
		fmask_[mi] = (float)perception_filter_mask;
	}	
}
	
// --------------------------------------------------------------------

void ReduceUnit_3_BIN_BALANCER::generate_colours_pass1_associate()
{		
	int progress = 0;
	int last_progress_pct = -1;

	signed short track_max_pixrefs = 0;
		
	ImagePkg& imgpkg = ImagePkg::instance();	

		
	// pass #1 - create palette bins with pixel references, and pixel records with bin references
	
	if (g_allargs.verbose)
		printf("generating colour bins...\n");
	
	//int prnt = 256;

	for (signed short scan = 1; scan < display_height_; ++scan)
	{
		//byte_t* sync_mapping_scan = &PCT_[(scan-1)*display_width_];
		byte_t* sync_mapping_scan = ADDRESS_PCT_SCAN(PCT_, 0, display_width_, scan);

//		printf("scan: %d\n", scan);
		
		// linear left-to-right ordering
		for (int x = 0; x < display_width_; ++x)
		{							
			unsigned long pixel_index = (scan * display_width_) + x;

			pixel_record& record = pixel_records_[pixel_index];	


			// extract pixel
			byte_t ur, ug, ub;
			imgpkg_.get_rgb(active_source_, x, (int)scan, display_height_, ur, ug, ub);
			signed short srn = ((signed short)(unsigned short)ur);
			signed short sgn = ((signed short)(unsigned short)ug);
			signed short sbn = ((signed short)(unsigned short)ub);
		
			// fetch input adjustments (lace, dither)
			
			float fladj = 0.0f;
			if (laceunit_)
				fladj = laceunit_->fsample(x, scan);
		
			float fdxadj = 0.0f;
			float fdyadj = 0.0f;
			float fdzadj = 0.0f;
			if (ditherunit_)
				ditherunit_->fsample(x, scan, fdxadj, fdyadj, fdzadj);
		
			// into unit space, combining lace/dither adjustments - this is the intended colour
			float fsx = col2f(srn) + fladj + fdxadj;
			float fsy = col2f(sgn) + fladj + fdyadj;
			float fsz = col2f(sbn) + fladj + fdzadj;
				
			// intended field colour
			record.ix = fsx;
			record.iy = fsy;
			record.iz = fsz;

			// field error adjustment should offset colour being generated, but only a portion (if any)
			// of this error should be fed back into next field error otherwise it gets amplified
			float fexadj = 0.0f;
			float feyadj = 0.0f;
			float fezadj = 0.0f;
			if (errorunit_)
				errorunit_->sample_field(x, scan, fexadj, feyadj, fezadj);

			//if (--prnt > 0)
			//	printf("fedr: %.4f, fedg: %.4f, fedb: %.4f\n", fexadj, feyadj, fezadj);

			// aim colour (including feedback error offset)
			record.cx = fsx + fexadj;
			record.cy = fsy + feyadj;
			record.cz = fsz + fezadj;

/*				
			// extract pixel
			byte_t ur, ug, ub;
			imgpkg.get_rgb(active_source_, (int)x, (int)scan, display_height_, ur, ug, ub);

			// sample lace & dither units (1d offset)
			int fdither = 0;//laceunit_->sample(x, scan) + ditherunit_->sample(x, scan);

			// sample error unit (only sample error from opposite field, when generating palette)
			signed short edr = 0, edg = 0, edb = 0;
			//if (errorunit_)
				//errorunit_->sample_field(x, scan, edr, edg, edb);
							
			// offset & quantize colour before matching
			signed short srn = (((unsigned short)ur) + fdither + edr);
			signed short sgn = (((unsigned short)ug) + fdither + edg);
			signed short sbn = (((unsigned short)ub) + fdither + edb);

			// clamp
			//srn = xclamp<signed short>(srn, 0, 255);
			//sgn = xclamp<signed short>(sgn, 0, 255);
			//sbn = xclamp<signed short>(sbn, 0, 255);
				
			// convert to CIELAB		
			double dsr = srn, dsg = sgn, dsb = sbn;							
			double dcx,dcy,dcz;
			
			//rgb2cielab(dsr,dsg,dsb,dcx,dcy,dcz);
			dcx = dsr;
			dcy = dsg;
			dcz = dsb;
			
			//dcy += 128.0;
			//dcz += 128.0;
			
			//rgb2xyz(dsr,dsg,dsb,dcx,dcy,dcz);
			//dcx = dsr;
			//dcy = dsg;
			//dcz = dsb;
	
			// update record
//			record.cx = (unsigned short)xmin<signed long>((signed long)(dcx * 255.0f), 65535);
//			record.cy = (unsigned short)xmin<signed long>((signed long)(dcy * 255.0f), 65535);
//			record.cz = (unsigned short)xmin<signed long>((signed long)(dcz * 255.0f), 65535);
			record.cx = dcx;
			record.cy = dcy;
			record.cz = dcz;
*/
						
			// find palette mapping for this 'x'			
			byte_t* sync_mapping = &sync_mapping_scan[x << MAX_PALETTE_SHIFT];
			
			// insert reference to this pixel into every visible palette index
			
			for (signed short p = g_allargs.colours-1; p >= 0; --p)
			{
				word_t mask = 0;

				if (DPT_) 
					mask = DPT_[x + (scan*display_width_)];
							
				// avoid masked colours for this pixel
				if ((mask & (1 << p)) == 0)
				{
					// index correct palette for this x & p
					byte_t upi = sync_mapping[p];						
					signed short spi = (signed short)((signed char)upi);
	
					// calculate absolute palette bin to fill, from:				
					//  scan = current line
					// 	spi = relative palette offset from this scan
					// 	p = colour index
#if defined(__ATARI__)
					signed short ibin = (scan * palettebuffer_stride_) + (spi * g_allargs.colours) + p;
#else
					int ibin = (scan * palettebuffer_stride_) + (spi * g_allargs.colours) + p;
#endif
						
					palette_bin& rbin = palette_bins_[ibin];
					// assign bin index/name
					rbin.index = p;
					
					if (rbin.num_pixel_references < max_pixel_references)
					{
						// add pixel reference										
						pixel_reference& rref = rbin.pix_references[rbin.num_pixel_references];
						++rbin.num_pixel_references;
								
						track_max_pixrefs = xmax(track_max_pixrefs, rbin.num_pixel_references);
						
						rref.pixel_x = x;
						rref.pixel_y = scan;												
						rref.error = 0.0f;
						
						// add bin reference to pixel record, for reduction later
						bin_reference& rbref = record.bin_references[record.num_bin_references++];
						rbref.bin = ibin;
						//rbref.error = 0.0f;
						
						if (record.num_bin_references > g_allargs.colours)
							printf("******************************* error on refcount *****************\n");
					}
					else
					{
						static bool s_bin_warning = false;
						if (!s_bin_warning)
						{
							s_bin_warning = true;
							std::cerr << "error - number of possible pixel references per colour has exceeded fixed limit of (" << max_pixel_references << ")" << std::endl;
							std::cerr << "\tthis will cause some pixel positions to remain unprocessed and will display incorrect colours" << std::endl;
							std::cerr << "\teither uses more colour changes per scan to reduce the pixel->colour refcount" << std::endl;
							std::cerr << "\tor rebuild from source with larger 'max_pixel_references'," << std::endl;
							std::cerr << "\tor use a reduction method which does not use pixel<->colour association (-m 0,1,2,4)" << std::endl;
						}
					}
				}
			}			
		}
		
		#ifdef __ATARI__
		// update progress
		if (!g_allargs.quiet)
		{
			int progress_pct = ((int)scan * 100) / display_height_;
			if (progress_pct != last_progress_pct)
			{
				last_progress_pct = progress_pct;		
				switch (progress & 3)
				{
				case 0:
					printf("- %2d%%\r", progress_pct); break;
				case 1:
					printf("\\ %2d%%\r", progress_pct); break;
				case 2:
					printf("| %2d%%\r", progress_pct); break;
				case 3:
					printf("/ %2d%%\r", progress_pct); break;
				};
				fflush(stdout);
				progress++;
			}
		}
		#endif
	}
	
	if (g_allargs.verbose)
		printf("max pixel refs per bin = %d\n", track_max_pixrefs);
}

void ReduceUnit_3_BIN_BALANCER::update(signed long ibin)
{
		palette_bin& rbin = palette_bins_[ibin];
		
		// perform error scoring if pixel is sharing a bin with others, 
		// otherwise allocate perfect score
		if (rbin.num_pixel_references <= 1)
		{
			if (rbin.num_pixel_references == 1)
			{
				pixel_reference& rref = rbin.pix_references[0];
				// for bins with a population==1, allocate perfect score
				rref.error = 0.0f;
			}
		}
		else
		{				
			// iterate all pixels references
			for (signed short ipref1 = 0; ipref1 < rbin.num_pixel_references; ++ipref1)
			{
				pixel_reference& rpref1 = rbin.pix_references[ipref1];
				signed long iprec1 = rpref1.pixel_x + (rpref1.pixel_y * display_width_);
				pixel_record& rprec1 = pixel_records_[iprec1];
								
				// calculate error relative to all contemproraries
				
				//double dcl1 = (double)(signed long)(unsigned long)rrec1.cx;
				//double dca1 = (double)(signed long)(unsigned long)rrec1.cy;
				//double dcb1 = (double)(signed long)(unsigned long)rrec1.cz;
				calc_real dcl1 = rprec1.cx;
				calc_real dca1 = rprec1.cy;
				calc_real dcb1 = rprec1.cz;

				calc_real e = 0.0f;
				calc_real ne = 0.0f;
				
				//double matchweight = 0.0f;
				
				for (signed short ipref2 = 0; ipref2 < rbin.num_pixel_references; ++ipref2)
				{
					if (ipref1 != ipref2)
					{
						pixel_reference& rpref2 = rbin.pix_references[ipref2];
						signed long iprec2 = rpref2.pixel_x + (rpref2.pixel_y * display_width_);
						pixel_record& rprec2 = pixel_records_[iprec2];							
						
						//double dcl2 = (double)(signed long)(unsigned long)rrec2.cx;
						//double dca2 = (double)(signed long)(unsigned long)rrec2.cy;
						//double dcb2 = (double)(signed long)(unsigned long)rrec2.cz;
						calc_real dcl2 = rprec2.cx;
						calc_real dca2 = rprec2.cy;
						calc_real dcb2 = rprec2.cz;
						
						// combine influence of competing pixels for error adjustment
						calc_real perception_filter_mask = inverse_product<calc_real>(fmask_[iprec1], fmask_[iprec2]);
						
						// accumulate squared error, biased by importance of competing pixels
						calc_real ecl = (dcl2 - dcl1) * perception_filter_mask;
						calc_real eca = (dca2 - dca1) * perception_filter_mask;
						calc_real ecb = (dcb2 - dcb1) * perception_filter_mask;
						calc_real ee = ((ecl * ecl) + (eca * eca) + (ecb * ecb));
						e += ee;
						
						//double c = 1.0 - (ee / 3.0);
						//assert((c >= 0) && (c <= 1.0));
																
						// accumulate match weight
						//matchweight += xclamp(c, 0.0, 1.0);

						//
						ne += 1.0f;
					}
				}
				//matchweight /= ne;
				
				//double ae = (1.0 / matchweight);
				
				calc_real ae = e / ne;
				
				//w /= ne;
				
				// renormalize bin error
				//double ae = e / ne;
				//ae = e;
				
				//ae /= w;
				//assert(!isnan(ae));
				//assert(!isinfinite(ae));
		
				//if (ae > max_ae) 
				//	max_ae = ae;			
				//if (ae > (1 << 30))
				//	ae = 1 << 30;
					
				//printf("error: %.8f\n", (float)ae);
				rpref1.error = ae;// / 255.0;
/*				
				// load error into bin reference from this pixel
				int bi = 0;
				for (; bi < rprec1.num_bin_references; ++bi)
				{
					bin_reference& rbref = rprec1.bin_references[bi];
					if (rbref.bin == ibin)
					{
						rbref.error = rpref1.error;
						break;
					}
				}
				if (bi == rprec1.num_bin_references)
					printf("******************************* error on bref search! *****************\n");
*/						
			}
		}
}

void ReduceUnit_3_BIN_BALANCER::generate_colours_pass2a_score()
{
	// pass #2, step #1 - recalculate sharing error for every bin reference
	
	int progress = 0;
	int last_progress_pct = -1;
	
	// iterate all bins
	for (signed long ibin = 0; ibin < num_bins_; ++ibin)
	{
		update(ibin);
				
		#ifdef __ATARI__
		// update progress
		if (!g_allargs.quiet)
		{
			int progress_pct = (ibin * 100) / num_bins_;
			if (progress_pct != last_progress_pct)
			{
				last_progress_pct = progress_pct;		
				switch (progress & 3)
				{
				case 0:
					printf("- %2d%%\r", progress_pct); break;
				case 1:
					printf("\\ %2d%%\r", progress_pct); break;
				case 2:
					printf("| %2d%%\r", progress_pct); break;
				case 3:
					printf("/ %2d%%\r", progress_pct); break;
				};
				fflush(stdout);
				progress++;
			}
		}
		#endif
	}
/*
	// renormalize bin reference error for each pixel
	for (signed long irec = 0; irec < num_records_; ++irec)
	{
		pixel_record& rrec = pixel_records_[irec];			
		
		if (rrec.num_bin_references >= 1)
		{
			calc_real e = 0.0f;
			for (signed short bi = 0; bi < rrec.num_bin_references; ++bi)
			{
				const bin_reference& rbref = rrec.bin_references[bi];
				e += rbref.error;				
			}

			if (e > 1e-6)
			{
				calc_real re = 1.0f / e;
				for (signed short bi = 0; bi < rrec.num_bin_references; ++bi)
				{
					bin_reference& rbref = rrec.bin_references[bi];
					rbref.error *= re;				
				}
			}
			else
			{
				for (signed short bi = 0; bi < rrec.num_bin_references; ++bi)
				{
					bin_reference& rbref = rrec.bin_references[bi];
					rbref.error = 1.0f;				
				}
			}
		}
	}
*/	
}

bool ReduceUnit_3_BIN_BALANCER::generate_colours_pass2b_simplify()
{
	// pass #2, step #2 - eliminate references with greatest error

	bool done = true;
	
	int progress = 0;
	int last_progress_pct = -1;

	// iterate all pixel records
	for (signed long irec = 0; irec < num_records_; ++irec)
	{
		pixel_record& rrec = pixel_records_[irec];			
		
		if (rrec.num_bin_references > 1)
		{
			// at least one pixel which hasn't been resolved to one bin
			done = false;
			
			// iterate over all bin references from this pixel
			//unsigned short best_error = 0;
			calc_real best_error = 0;
			unsigned long best_bin_2_pixref_index = ~0;
			unsigned long best_pix_2_binref_index = ~0;
			unsigned long best_bin_index = ~0;
			unsigned int best_bin_size = 0;
			for (signed short bi = 0; bi < rrec.num_bin_references; ++bi)
			{
				// find bin for this reference
				const bin_reference& rbref = rrec.bin_references[bi];
#if defined(__ATARI__)
				signed short ibin = rbref.bin;
#else
				int ibin = rbref.bin;
#endif
				palette_bin& rbin = palette_bins_[ibin];
				// locate reference within bin (slow search, but we're already using a ton of ram...)
				//bool found = false;
				signed short found_bin_2_pixref_index = -1;
				for (signed short ri = 0; ri < rbin.num_pixel_references; ++ri)
				{
					pixel_reference& rref = rbin.pix_references[ri];
					
					unsigned long irecref = rref.pixel_x + (rref.pixel_y * display_width_);					
					if (irec == irecref)
					{
						found_bin_2_pixref_index = ri;
						break;
					}
				}
				
				//if (found_bin_2_pixref_index < 0)
				//	printf("****************** error finding bin->pixel reference *************\n");

				// track error
				pixel_reference& rref = rbin.pix_references[found_bin_2_pixref_index];					
				
				if (rbin.num_pixel_references == 1)
				{
					// there should never be a bin with population==1 and error > 0
					//assert(rref.error < 1e-6f);
				}
					
				//if (rbin.num_pixel_references > 1)
				{
					
					calc_real combined_error = rref.error;//r/* * rbref.error*/;
					if (combined_error >= best_error)
					{
						//if ((combined_error != best_error) ||
						//	((combined_error == best_error) && (rbin.num_pixel_references > best_bin_size)))
						{
							best_error = combined_error;
							best_bin_2_pixref_index = found_bin_2_pixref_index;
							best_pix_2_binref_index = bi;
							best_bin_index = ibin;
							best_bin_size = rbin.num_pixel_references;
						}
					}
				}
			}

			//if (best_bin_index == ~0)
			//	printf("****************** could not find best bin index! *************\n");
				
			
			// remove this link between palette bin and pixel record
			// need:
			//  *palette_bin*.references[*index*]
			//  pixel_record.bin_references[*index*]

			//pixel_reference& rpref = rbin.references[best_bin_2_pixref_index];
			
			palette_bin& rbin = palette_bins_[best_bin_index];

			// remove link from bin->pix (by packing the slot with the last remaining reference)
			//if (best_bin_2_pixref_index >= rbin.num_references)
			//	printf("****************** bad bin_2_pixref_index *************\n");
			rbin.num_pixel_references--;
			if (rbin.num_pixel_references > 0)
				rbin.pix_references[best_bin_2_pixref_index] = rbin.pix_references[rbin.num_pixel_references];
			
			// remove link from pix->bin
			//if (best_pix_2_binref_index >= rrec.num_references)
			//	printf("****************** bad bin_2_pixref_index *************\n");
			
			rrec.num_bin_references--;
			if (rrec.num_bin_references > 0)
				rrec.bin_references[best_pix_2_binref_index] = rrec.bin_references[rrec.num_bin_references];

			// update error on references from modified bin			
			if (quality_ == Quality_HIGH)
				update((signed long)best_bin_index);
		}
		
		
		#ifdef __ATARI__			
		// update progress
		if (!g_allargs.quiet)
		{
			int progress_pct = (irec * 100) / num_records_;
			if (progress_pct != last_progress_pct)
			{
				last_progress_pct = progress_pct;		
				switch (progress & 3)
				{
				case 0:
					printf("- %2d%%\r", progress_pct); break;
				case 1:
					printf("\\ %2d%%\r", progress_pct); break;
				case 2:
					printf("| %2d%%\r", progress_pct); break;
				case 3:
					printf("/ %2d%%\r", progress_pct); break;
				};
				fflush(stdout);
				progress++;
			}
		}
		#endif
	}
	
	return done;
}
	
void ReduceUnit_3_BIN_BALANCER::generate_colours_pass2_reduce()
{
	// pass #2 - iteratively reduce bin population
	
	if (g_allargs.verbose)	
		printf("iteratively reducing bins...\n");
		
	int iter = 1;
	bool firstpass = true;
	bool done = true;
	do 
	{
		if (g_allargs.verbose)
		#ifdef __ATARI__
		{ printf("iteration %2d\n", iter++); fflush(stdout); }
		#else
		{ printf("\riteration %2d", iter++); fflush(stdout); }
		#endif

		// always perform initial scoring on first iteration
		// otherwise, only on subsequent iterations in LOW quality mode
		// since HIGH quality update bin scores immediately on each reference change
		if (firstpass ||
			(quality_ == Quality_LOW))
		{						
			firstpass = false;
			generate_colours_pass2a_score();
		}
/*		
		printf("\n");	
				
		int last_pal = 0;
		int last_scan = 0;
		for (signed long ibin = 0; ibin < num_bins_; ++ibin)
		{
			palette_bin& rbin = palette_bins_[ibin];
			
			int scan = (ibin / palettebuffer_stride_);
			int scani = ibin - (scan * palettebuffer_stride_);
			int pal = scani / g_allargs.colours;
			int col = scani - (pal * g_allargs.colours);
	
			if (scan != last_scan)
			{
				printf("\n");			
			}
			else if (pal != last_pal)
			{
				printf("  ");			
			}
			
			// calculate bin error
			double e = 0;
			for (signed short iref = 0; iref < rbin.num_references; ++iref)
			{				
				pixel_reference& rref = rbin.pix_references[iref];
				assert(!isnan(rref.error));
				e += rref.error;
			}		
			
			printf("%.2f ", e);
			
			last_scan = scan;
			last_pal = pal;
		}		
*/		
		done = generate_colours_pass2b_simplify();
	}
	while (!done);
	
	// final scoring for diagnostics
	generate_colours_pass2a_score();

	#ifndef __ATARI__
	if (g_allargs.verbose)	
		printf("\n");
	#endif
}


void ReduceUnit_3_BIN_BALANCER::generate_colours_pass3_merge()
{
	// calculate error for native colour depth
	// 1/2 of the native colour LSB
	//int n = (8 - colourdepth_) - 1;
	//int colour_offset = 1 << n;//8;
	//int colour_denormalize = 8 - colourdepth_;
	
	// pass #3 - merge bins, generating final colours
	// note: lossy, numeric operations taking place here. also sensitive to colour space used.
	// when operating in RGB space, floating point is not necessary. storing 8-bit triplets per pixel reference will suffice.
	// when operating in CIE or other spaces, floating point (or at least careful use of fixed point) is recommended

	for (signed long ibin = 0; ibin < num_bins_; ++ibin)
	{
		palette_bin& rbin = palette_bins_[ibin];
		
		if (rbin.num_pixel_references > 0)
		{
			// merge bin - sum colour channels
			
			calc_real dcxm = 0.0f;
			calc_real dcym = 0.0f;
			calc_real dczm = 0.0f;
			calc_real n = 0.0f;
			
			for (signed short iref = 0; iref < rbin.num_pixel_references; ++iref)
			{
				pixel_reference& rref = rbin.pix_references[iref];
				signed long irec = rref.pixel_x + (rref.pixel_y * display_width_);
				pixel_record& rrec = pixel_records_[irec];

				// sum channels
				calc_real dcx = rrec.cx;
				calc_real dcy = rrec.cy;
				calc_real dcz = rrec.cz;

				// weight colour contribution according to perception filter for this pixel
				calc_real perception_filter_mask = fmask_[irec];
				dcxm += dcx * perception_filter_mask;
				dcym += dcy * perception_filter_mask;
				dczm += dcz * perception_filter_mask;
				n += perception_filter_mask;
			}
			
			// renormalize
			calc_real rn = (1.0f / n)/* / 255.0f*/;
			dcxm *= rn;
			dcym *= rn;
			dczm *= rn;

			// into palette space for this field
			const calc_real fpn = (calc_real)pal_crange(colourdepth_);
			signed short sr = f2range(dcxm, fpn);
			signed short sg = f2range(dcym, fpn);
			signed short sb = f2range(dczm, fpn);			
		
			word_t palette_entry =  0x8000 | 
									(sr << (r_pos_ - colourdepth_)) | 
									(sg << (g_pos_ - colourdepth_)) | 
									(sb << (b_pos_ - colourdepth_));

			signed long real_palette_index = ibin;
			
			//if (real_palette_index < 32)
			//	printf("****************** bad ibin out of display range! *************\n");
			//if (real_palette_index > (palettebuffer_words_ + 32))
			//	printf("****************** bad ibin out of display range! *************\n");
							
			palette_[real_palette_index] = palette_entry;			
	
			
			// convert back to integer storage
			//cl = (signed short)dcl; 
			//ca = (signed short)dca; 
			//cb = (signed short)dcb; 
			//
			// modify pixels to use merged colour
			//for (signed short iref = 0; iref < rbin.num_references; ++iref)
			//{
			//	pixel_reference& rref = rbin.references[iref];
			//	signed long irec = rref.pixel_x + (rref.pixel_y * display_width_);
			//	pixel_record& rrec = pixel_records[irec];
			//					
			//	// replace colour
			//	rrec.cx = cx;
			//	rrec.cy = cy;
			//	rrec.cz = cz;
			//}			
		}
	}
}

// --------------------------------------------------------------------

void ReduceUnit_3_BIN_BALANCER::allocate_pixels()
{
	//allocate_pixels_remap_();
	//allocate_pixels_stable_();

	if ((g_allargs.errortype == ErrorType_FIELD) &&		
		(g_allargs.erroriters > 1))
	{	
		// field-error feedback benefits from more tightly controlled image remapping...
		// the first field of should always be remapped as a starting reference point
		// all subsequent fields use a stable mapping to push the error around without
		// hysteresis
		
		if ((index_ == 0) && 
			(error_iteration_ == 0))
		{
			// first field of first iteration - force remap for best results. all subsequent fields respond to feedback only
			allocate_pixels_remap_();
		}
		else
		{
			// otherwise use stable index mapping between fields for sensible feedback
			allocate_pixels_stable_();
		}
	}
	else
	{
		// otherwise just remap all fields for best fit
		allocate_pixels_remap_();
	}

	
/*	
	// pass #4 - map image to available colours
	if ((g_allargs.errortype == ErrorType_SPATIAL) || 
		(g_allargs.errortype == ErrorType_MULTIMODE))
	{
		// use this one if we are apply FS dithering and need to remap the image properly, propagating error
		// into the colour search as we go
		allocate_pixels_remap_();
	}
	else
	{
		// use this one if we're only sampling error from the opposite field, as this gets dealt with while generating the palette
		// also use this one if we're not using error diffusion at all
		allocate_pixels_stable_();
	}
*/

#ifdef _DEBUG
	printf("\n");	

	// dump the pixel reference count per bin
	int last_pal = 0;
	int last_scan = 0;
	for (signed long ibin = 0; ibin < num_bins_; ++ibin)
	{
		palette_bin& rbin = palette_bins_[ibin];
		
		int scan = (ibin / palettebuffer_stride_);
		int scani = ibin - (scan * palettebuffer_stride_);
		int pal = scani / g_allargs.colours;
		int col = scani - (pal * g_allargs.colours);

		if (scan != last_scan)
		{
			printf("\n");			
		}
		else if (pal != last_pal)
		{
			printf("  ");			
		}
		
		if (rbin.num_pixel_references > 0)
			printf("%2x ", rbin.num_pixel_references);
		else
			printf("-- ");
			
		
		last_scan = scan;
		last_pal = pal;
	}
			
	printf("\n");
#endif
		
/*			
	last_pal = 0;
	last_scan = 0;
	for (signed long ibin = 0; ibin < num_bins_; ++ibin)
	{
		palette_bin& rbin = palette_bins_[ibin];
		
		int scan = (ibin / palettebuffer_stride_);
		int scani = ibin - (scan * palettebuffer_stride_);
		int pal = scani / g_allargs.colours;
		int col = scani - (pal * g_allargs.colours);

		if (scan != last_scan)
		{
			printf("\n");			
		}
		else if (pal != last_pal)
		{
			printf("  ");			
		}
		
		// calculate bin error
		calc_real e = 0;
		for (signed short iref = 0; iref < rbin.num_pixel_references; ++iref)
		{				
			pixel_reference& rref = rbin.pix_references[iref];
			assert(!isnan(rref.error));
			e += rref.error;
		}		
		
		printf("%.3f ", (double)e);
		
		last_scan = scan;
		last_pal = pal;
	}	
*/	
}

// --------------------------------------------------------------------
// pass #4 - reduce image, quickly using allocated bin colour index
// note: nothing numerical happening here - just dealing with references

void ReduceUnit_3_BIN_BALANCER::allocate_pixels_stable_()
{	
	if (g_allargs.verbose)	
		printf("remapping pixels using preassigned indices...\n");
	
	for (signed long irec = 0; irec < num_records_; ++irec)
	{
		pixel_record& rrec = pixel_records_[irec];			
		
		if (rrec.num_bin_references > 0)
		{	
			//if (rrec.num_references > 1)
			//	printf("****************** pixel left pointing at more than one bin! *************\n");

			// find the bin containing the colour intended to be used by this pixel			
			const bin_reference& rbref = rrec.bin_references[0];
#if defined(__ATARI__)
			signed short ibin = rbref.bin;
#else
			int ibin = rbref.bin;
#endif
			palette_bin& rbin = palette_bins_[ibin];

			// use bin's colour index
			bitmap_[irec] = rbin.index;

			if (errorunit_)
			{
				unsigned short chosen_index = rbin.index;
				
				// get coordinates from record index
				signed short y = irec / display_width_;
				signed short x = irec - (y * display_width_);
	
				//byte_t* sync_mapping = &PCT_[x << 4];
				byte_t* sync_mapping = ADDRESS_PCT_SCAN(PCT_, x, display_width_, y);	// todo: check this!
				word_t* scan_palette = &palette_[y * palettebuffer_stride_];
				
				// get assigned colour
				
				// index correct palette for this x & p
				byte_t upi = sync_mapping[chosen_index];
				signed short spi = (signed short)((signed char)upi);		
				word_t* x_palette = &scan_palette[spi * g_allargs.colours];

				// find colour (0x0RGB)
				word_t pc = x_palette[chosen_index];
										
				// palette colour
				signed short pr = ((pc >> (r_pos_ - colourdepth_)) & c_mask_);
				signed short pg = ((pc >> (g_pos_ - colourdepth_)) & c_mask_);
				signed short pb = ((pc >> (b_pos_ - colourdepth_)) & c_mask_);

				// into unit space
				const calc_real fpn = (calc_real)pal_crange(colourdepth_);
				calc_real fpr = range2f(pr, fpn);
				calc_real fpg = range2f(pg, fpn);
				calc_real fpb = range2f(pb, fpn);			

				calc_real fsx = rrec.ix;
				calc_real fsy = rrec.iy;
				calc_real fsz = rrec.iz;

				feedback_field_error
				(
					x, y, 
					chosen_index, 
					fsx, 
					fsy, 
					fsz
				);
/*				
				// quantize in native virtual space
				//fsx = range2f(f2range(fsx, (float)nvi_crange(colourdepth_, fields_)), (float)nvi_crange(colourdepth_, fields_));
				//fsy = range2f(f2range(fsy, (float)nvi_crange(colourdepth_, fields_)), (float)nvi_crange(colourdepth_, fields_));
				//fsz = range2f(f2range(fsz, (float)nvi_crange(colourdepth_, fields_)), (float)nvi_crange(colourdepth_, fields_));
								
				// calculate error between matched colour and intended field colour
				
				float perception_filter_mask = fmask_[irec];			
				float fedr = (fsx - fpr) * perception_filter_mask;
				float fedg = (fsy - fpg) * perception_filter_mask;
				float fedb = (fsz - fpb) * perception_filter_mask;
							
				// feedback field error				
				errorunit_->feedback_field(x, y, fedr, fedg, fedb);
*/				
			}
		}
	}
}

void ReduceUnit_3_BIN_BALANCER::allocate_pixels_remap_()
{	
	if (g_allargs.verbose)	
		printf("remapping pixels using best match...\n");

	int progress = 0;
	int last_progress_pct = -1;
			
	signed short palette_scan_size = palettebuffer_stride_;
	signed short bitmap_scan_size = display_width_;
		
	for (int scan = 1; scan < display_height_; ++scan)
	{
		word_t* scan_palette = &palette_[scan * palette_scan_size];
		byte_t* scan_bitmap = &bitmap_[scan * bitmap_scan_size];
		
		//byte_t* sync_mapping_scan = &PCT_[(scan-1)*display_width_];
		byte_t* sync_mapping_scan = ADDRESS_PCT_SCAN(PCT_, 0, display_width_, scan);
		
		// alternating x-sweep for FS error propagation
		
		signed short xsweepflag = (scan & 1);
		signed short xsweepdir = (xsweepflag << 1) - 1;
		if (errorunit_)
			errorunit_->set_sweep(xsweepdir);
		
		signed short x = 0;
		signed short xend = bitmap_scan_size;
		if (xsweepflag == 0)
		{
			xend = -1;
			x = bitmap_scan_size - 1;
		}	

		// map one line of image
				
 		while (x != xend)
 		{							
			// find palette mapping for this 'x'			
			byte_t* sync_mapping = &sync_mapping_scan[x << MAX_PALETTE_SHIFT];
	
			pixel_record& rrec = pixel_records_[x + (scan * display_width_)];			
		
			byte_t chosen_index = 0;
			
			if (rrec.num_bin_references > 0)
			{	
				float fexadj = 0.0f;
				float feyadj = 0.0f;
				float fezadj = 0.0f;
				if (errorunit_)
					errorunit_->sample_spatial(x, scan, fexadj, feyadj, fezadj);
	
				// into unit space, combining spatial error adjustment
				float fsx = rrec.ix + fexadj;
				float fsy = rrec.iy + feyadj;
				float fsz = rrec.iz + fezadj;
				//calc_real fsx = rrec.ix;
				//calc_real fsy = rrec.iy;
				//calc_real fsz = rrec.iz;
										
				// into palette space
				const calc_real fpn = (calc_real)pal_crange(colourdepth_);
				signed short sr = f2range(fsx, fpn);
				signed short sg = f2range(fsy, fpn);
				signed short sb = f2range(fsz, fpn);			
	
				chosen_index = match_colour(x, scan, sr, sg, sb);
				
				if (errorunit_)
				{
					feedback_error
					(
						x, scan, 
						chosen_index, 
						fsx, 
						fsy, 
						fsz
					);
				}	
			}
			
			scan_bitmap[x] = chosen_index;
			
			x += xsweepdir;	
		}
		
		#ifdef __ATARI__
		// update progress
		if (!g_allargs.quiet)
		{
			int progress_pct = ((int)scan * 100) / display_height_;
			if (progress_pct != last_progress_pct)
			{
				last_progress_pct = progress_pct;		
				switch (progress & 3)
				{
				case 0:
					printf("- %2d%%\r", progress_pct); break;
				case 1:
					printf("\\ %2d%%\r", progress_pct); break;
				case 2:
					printf("| %2d%%\r", progress_pct); break;
				case 3:
					printf("/ %2d%%\r", progress_pct); break;
				};
				fflush(stdout);
				progress++;
			}
		}
		#endif
	}
}

// --------------------------------------------------------------------

arg_real ReduceUnit_INDIRECT_ANALYTICAL::feedback_error(int x, int scan, byte_t chosen_index, arg_real fsx, arg_real fsy, arg_real fsz)
{
	// get assigned colour
	
	// index correct palette for this x & p
	//byte_t* sync_mapping = &PCT_[x << 4];
	byte_t* sync_mapping = ADDRESS_PCT_SCAN(PCT_, x, display_width_, scan);
	
	byte_t upi = sync_mapping[chosen_index];
	signed short spi = (signed short)((signed char)upi);		
	word_t* scan_palette = &palette_[scan * palettebuffer_stride_];
	word_t* x_palette = &scan_palette[spi * g_allargs.colours];

	// find colour (0x0RGB)
	word_t pc = x_palette[chosen_index];

	// palette colour
	signed short pr = ((pc >> (r_pos_ - colourdepth_)) & c_mask_);
	signed short pg = ((pc >> (g_pos_ - colourdepth_)) & c_mask_);
	signed short pb = ((pc >> (b_pos_ - colourdepth_)) & c_mask_);

	// into unit space
	const calc_real fpn = (calc_real)pal_crange(colourdepth_);
	calc_real fpr = range2f(pr, fpn);
	calc_real fpg = range2f(pg, fpn);
	calc_real fpb = range2f(pb, fpn);	

	// quantize in native virtual space
	//fsx = range2f(f2range(fsx, (float)nvi_crange(colourdepth_, fields_)), (float)nvi_crange(colourdepth_, fields_));
	//fsy = range2f(f2range(fsy, (float)nvi_crange(colourdepth_, fields_)), (float)nvi_crange(colourdepth_, fields_));
	//fsz = range2f(f2range(fsz, (float)nvi_crange(colourdepth_, fields_)), (float)nvi_crange(colourdepth_, fields_));

	// calculate error
	calc_real perception_filter_mask = fmask_[x + (scan * display_width_)];			
	calc_real fedr = (fsx - fpr) * perception_filter_mask;
	calc_real fedg = (fsy - fpg) * perception_filter_mask;
	calc_real fedb = (fsz - fpb) * perception_filter_mask;
								
	// feedback error				
	errorunit_->feedback(x, scan, fedr, fedg, fedb);

	return (fedr * fedr) + (fedg * fedg) + (fedb * fedb);
}

arg_real ReduceUnit_INDIRECT_ANALYTICAL::feedback_field_error(int x, int scan, byte_t chosen_index, arg_real fsx, arg_real fsy, arg_real fsz)
{
	// get assigned colour
	
	// index correct palette for this x & p
	//byte_t* sync_mapping = &PCT_[x << 4];
	byte_t* sync_mapping = ADDRESS_PCT_SCAN(PCT_, x, display_width_, scan);
	
	byte_t upi = sync_mapping[chosen_index];
	signed short spi = (signed short)((signed char)upi);		
	word_t* scan_palette = &palette_[scan * palettebuffer_stride_];
	word_t* x_palette = &scan_palette[spi * g_allargs.colours];

	// find colour (0x0RGB)
	word_t pc = x_palette[chosen_index];

	// palette colour
	signed short pr = ((pc >> (r_pos_ - colourdepth_)) & c_mask_);
	signed short pg = ((pc >> (g_pos_ - colourdepth_)) & c_mask_);
	signed short pb = ((pc >> (b_pos_ - colourdepth_)) & c_mask_);

	// into unit space
	const calc_real fpn = (calc_real)pal_crange(colourdepth_);
	calc_real fpr = range2f(pr, fpn);
	calc_real fpg = range2f(pg, fpn);
	calc_real fpb = range2f(pb, fpn);	

	// quantize in native virtual space
	//fsx = range2f(f2range(fsx, (float)nvi_crange(colourdepth_, fields_)), (float)nvi_crange(colourdepth_, fields_));
	//fsy = range2f(f2range(fsy, (float)nvi_crange(colourdepth_, fields_)), (float)nvi_crange(colourdepth_, fields_));
	//fsz = range2f(f2range(fsz, (float)nvi_crange(colourdepth_, fields_)), (float)nvi_crange(colourdepth_, fields_));

	// calculate error
	calc_real perception_filter_mask = fmask_[x + (scan * display_width_)];			
	calc_real fedr = (fsx - fpr) * perception_filter_mask;
	calc_real fedg = (fsy - fpg) * perception_filter_mask;
	calc_real fedb = (fsz - fpb) * perception_filter_mask;
								
	// feedback error				
	errorunit_->feedback_field(x, scan, fedr, fedg, fedb);

	return (fedr * fedr) + (fedg * fedg) + (fedb * fedb);
}

// --------------------------------------------------------------------
