// ---------------------------------------------------------------------------------------------------------------------
//	PhotoChrome v6.xx - Atari ST interlaced image reduction framework
//	1991-2016 Douglas Little (d.m.l)
// ---------------------------------------------------------------------------------------------------------------------
//	special thanks to:
//		evil@DHS, RA/PDX, Troed/SYNC for help testing and input on improvements
// ---------------------------------------------------------------------------------------------------------------------

// ====================================================================
// 	image colour reduction unit
// ====================================================================

#include "common.h"
#include "config.h"

#if defined(__ATARI__) || defined(__MINGW32__)
#include "FreeImage.3.15.3.h"
#else
#include "FreeImage.3.15.4.h"
#endif

#include "imagepkg.h"
#include "pcsimage.h"

extern Arguments g_allargs;

// --------------------------------------------------------------------
// colour reduction/allocation method 0 (linear column, row order)
// --------------------------------------------------------------------
// this method is not intended to produce good results - it is provided
// only as the simplest possible method of matching colours - helpful
// when verifying rounding bias and range against known testcards, as
// well as a simple reference point for better reduction methods 
// --------------------------------------------------------------------
// allocate pixels left to right, top to bottom using the following
// rules:
// 	* look for an existing match
//	* otherwise look for a free colour slot to claim
//	* otherwise find match with lowest sum-squares error
// --------------------------------------------------------------------

void ReduceUnit_0_SIMPLE::generate_colours()
{
	if (g_allargs.verbose)
		printf("reducing field %d of %d / method 0 (simple)...\n", index_, fields_);
			
	int progress = 0;
	int last_progress_pct = -1;
					
	for (signed short scan = 1; scan < display_height_; ++scan)
	{				
		// linear left-to-right ordering
		for (int x = 0; x < display_width_; ++x)
		{
			sample(x, scan);			
		}
		
		#ifdef __ATARI__
		// update progress
		if (!g_allargs.quiet)
		{
			int progress_pct = ((int)scan * 100) / display_height_;
			if (progress_pct != last_progress_pct)
			{
				last_progress_pct = progress_pct;		
				switch (progress & 3)
				{
				case 0:
					printf("- %2d%%\r", progress_pct); break;
				case 1:
					printf("\\ %2d%%\r", progress_pct); break;
				case 2:
					printf("| %2d%%\r", progress_pct); break;
				case 3:
					printf("/ %2d%%\r", progress_pct); break;
				};
				fflush(stdout);
				progress++;
			}
		}
		#endif
	}	
}

// --------------------------------------------------------------------
// colour reduction/allocation method 1 (reverse carry column, linear row order)
// --------------------------------------------------------------------
// allocate pixels with horizontal reverse-carry (subdivision) pattern
// in scanline order, top to bottom using the following
// rules:
// 	* look for an existing match
//	* otherwise look for a free colour slot to claim
//	* otherwise find match with lowest sum-squares error
// --------------------------------------------------------------------

void ReduceUnit_1_SCAN_SUBDIVISION::generate_colours()
{
	if (g_allargs.verbose)
		printf("reducing field %d of %d / method 1 (scan subdivision)...\n", index_, fields_);
			
	int progress = 0;
	int last_progress_pct = -1;

	// calculate log2(width) and renormalizing bit shift for reverse carry addressing pattern
	int ceil_width = nlpo2(display_width_);
	int msb_width = 31-msb(ceil_width-1);
	int remainder_width = ceil_width - display_width_;
	int offset_width = remainder_width >> 1;
						
	for (signed short scan = 1; scan < display_height_; ++scan)
	{				
		// use reverse-carry addressing of the log2(display_width), ignoring pixels out of bounds
		// this effectively performs a divide-and-conquer ordering on solved pixels and helps reduce
		// the amount of streaking while using this rather simplistic colour allocation algorithm
		// that being: 1) look for a perfect match, or 2) a free slot, or 3) existing with lowest error
		for (int rcx = 0; rcx < ceil_width; ++rcx)
		{
			// reverse the bits and renormalize
			int x = (reverse(rcx) >> msb_width) - offset_width;
			// just shortcut out-of-bounds positions (512-320=192 dots will get skipped - but so what?)
			if (x >= display_width_ || x < 0)
				continue;			
			
			sample(x, scan);			
		}
		
		#ifdef __ATARI__
		// update progress
		if (!g_allargs.quiet)
		{
			int progress_pct = ((int)scan * 100) / display_height_;
			if (progress_pct != last_progress_pct)
			{
				last_progress_pct = progress_pct;		
				switch (progress & 3)
				{
				case 0:
					printf("- %2d%%\r", progress_pct); break;
				case 1:
					printf("\\ %2d%%\r", progress_pct); break;
				case 2:
					printf("| %2d%%\r", progress_pct); break;
				case 3:
					printf("/ %2d%%\r", progress_pct); break;
				};
				fflush(stdout);
				progress++;
			}
		}
		#endif
	}	
}

// --------------------------------------------------------------------
// colour reduction/allocation method 2 (linear congruential index order)
// --------------------------------------------------------------------
// allocate pixels using LCG index, which covers the framebuffer in
// pseudorandom order. this prevents the beginning of each scan from
// relying on colours allocated during the previous scan through 
// linear scan ordering. LCG ensures each pixel is only visited once.
// rules:
// 	* look for an existing match
//	* otherwise look for a free colour slot to claim
//	* otherwise find match with lowest sum-squares error
// --------------------------------------------------------------------

void ReduceUnit_2_LCG::generate_colours()
{
	if (g_allargs.verbose)
		printf("reducing field %d of %d / method 2 (global linear congruential)...\n", index_, fields_);
			
	int progress = 0;
	int last_progress_pct = -1;
	signed short progress_decay = 255;

	datasize_t lcg_input_range = display_width_ * display_height_;
	datasize_t lcg_lpo2_range = nlpo2(lcg_input_range);
	datasize_t lcg_v = (lcg_input_range / 2) | 1;	// start in display centre with odd number

	for (datasize_t lcg_index = 0; lcg_index < lcg_lpo2_range / 1; ++lcg_index)
	{
		// convert LCG value back into x,y coords
		signed short scan = lcg_v / display_width_;
		signed short x = lcg_v - (scan * display_width_);
		lcg_v = (lcg_v * (2048+1) + 4651) % lcg_lpo2_range;		// 261,117
		
		// clip LCG samples outwith framebuffer (since the full cycle is 2^n)
		if (scan == 0 || scan >= display_height_)
			continue;		
			
		sample(x, scan);			
		
		#ifdef __ATARI__
		// update progress
		if (!g_allargs.quiet)
		{
			if (--progress_decay < 0)
			{
				progress_decay = 255;
				int progress_pct = ((int)lcg_index * 100) / lcg_lpo2_range;
				if (progress_pct != last_progress_pct)
				{
					last_progress_pct = progress_pct;		
					switch (progress & 3)
					{
					case 0:
						printf("- %2d%%\r", progress_pct); break;
					case 1:
						printf("\\ %2d%%\r", progress_pct); break;
					case 2:
						printf("| %2d%%\r", progress_pct); break;
					case 3:
						printf("/ %2d%%\r", progress_pct); break;
					};
					fflush(stdout);
					progress++;
				}
			}
		}
		#endif
	}	
}

// --------------------------------------------------------------------
// colour reduction/allocation method 4 (reverse carry column order)
// --------------------------------------------------------------------
// allocate pixels with horizontal reverse-carry (subdivision) pattern
// in column order (whole column at a time), using the following
// rules:
// 	* look for an existing match
//	* otherwise look for a free colour slot to claim
//	* otherwise find match with lowest sum-squares error
// --------------------------------------------------------------------

void ReduceUnit_4_SCAN_SUBDIVISION2::generate_colours()
{
	if (g_allargs.verbose)
		printf("reducing field %d of %d / method 4 (scan subdivision using columns)...\n", index_, fields_);
			
	int progress = 0;
	int last_progress_pct = -1;

	// calculate log2(width) and renormalizing bit shift for reverse carry addressing pattern
	int ceil_width = nlpo2(display_width_);
	int msb_width = 31-msb(ceil_width-1);
	int remainder_width = ceil_width - display_width_;
	int offset_width = remainder_width >> 1;
						
	// use reverse-carry addressing of the log2(display_width), ignoring pixels out of bounds
	// this effectively performs a divide-and-conquer ordering on solved pixels and helps reduce
	// the amount of streaking while using this rather simplistic colour allocation algorithm
	// that being: 1) look for a perfect match, or 2) a free slot, or 3) existing with lowest error
	int column = 0;
	for (int rcx = 0; rcx < ceil_width; ++rcx)
	{
		// reverse the bits and renormalize
		int x = (reverse(rcx) >> msb_width) - offset_width;
		// just shortcut out-of-bounds positions (512-320=192 dots will get skipped - but so what?)
		if (x >= display_width_ || x < 0)
			continue;			

		for (signed short scan = 1; scan < display_height_; ++scan)
		{							
			sample(x, scan);			
		}

		#ifdef __ATARI__
		// update progress
		if (!g_allargs.quiet)
		{
			int progress_pct = ((int)column * 100) / display_width_;
			if (progress_pct != last_progress_pct)
			{
				last_progress_pct = progress_pct;		
				switch (progress & 3)
				{
				case 0:
					printf("- %2d%%\r", progress_pct); break;
				case 1:
					printf("\\ %2d%%\r", progress_pct); break;
				case 2:
					printf("| %2d%%\r", progress_pct); break;
				case 3:
					printf("/ %2d%%\r", progress_pct); break;
				};
				fflush(stdout);
				progress++;
			}
		}
		#endif
		
		++column;			
	}	
}

// --------------------------------------------------------------------
// common image sampler
// --------------------------------------------------------------------

void ReduceUnit_DIRECT_SAMPLING::sample(int x, int scan)
{
	word_t* scan_palette = &palette_[scan * palettebuffer_stride_];
	byte_t* scan_bitmap = &bitmap_[scan * display_width_];
		
	// extract pixel
	byte_t ur, ug, ub;
	signed short sr;
	signed short sg;
	signed short sb;
	float fsx;
	float fsy;
	float fsz;

	if (g_allargs.ccmode == CCMode_NONE)
	{
		imgpkg_.get_rgb(active_source_, (int)x, (int)scan, display_height_, ur, ug, ub);
		signed short srn = ((signed short)(unsigned short)ur);
		signed short sgn = ((signed short)(unsigned short)ug);
		signed short sbn = ((signed short)(unsigned short)ub);

		//printf("rgb: %x,%x,%x   ", srn, sgn, sbn);

		// fetch input adjustments (lace, dither)
		
		float fladj = 0.0f;
		if (laceunit_)
			fladj = laceunit_->fsample(x, scan);

		float fdxadj = 0.0f;
		float fdyadj = 0.0f;
		float fdzadj = 0.0f;
		if (ditherunit_)
			ditherunit_->fsample(x, scan, fdxadj, fdyadj, fdzadj);

		// into unit space, combining lace/dither adjustments - this is the intended colour
		fsx = col2f(srn) + fladj + fdxadj;
		fsy = col2f(sgn) + fladj + fdyadj;
		fsz = col2f(sbn) + fladj + fdzadj;

		// field error adjustment should offset colour being generated, but only a portion (if any)
		// of this error should be fed back into next field error otherwise it gets amplified

		float fexadj = 0.0f;
		float feyadj = 0.0f;
		float fezadj = 0.0f;
		if (errorunit_)
			errorunit_->sample_field(x, scan, fexadj, feyadj, fezadj);

		// into palette space for this field
		const float fpn = (float)pal_crange(colourdepth_);
		sr = f2range(fsx + fexadj, fpn);
		sg = f2range(fsy + feyadj, fpn);
		sb = f2range(fsz + fezadj, fpn);			
	}

#ifdef CRYPTOCHROME
	else
	{
		if (index_ == 0)
			imgpkg_.get_rgb(cc_field0_, (int)x, (int)scan, display_height_, ur, ug, ub);
		else
			imgpkg_.get_rgb(cc_field1_, (int)x, (int)scan, display_height_, ur, ug, ub);
		
		signed short srn = ((signed short)(unsigned short)ur);
		signed short sgn = ((signed short)(unsigned short)ug);
		signed short sbn = ((signed short)(unsigned short)ub);

		fsx = col2f(srn);
		fsy = col2f(sgn);
		fsz = col2f(sbn);
		
		// into palette space for this field
		const float fpn = (float)pal_crange(colourdepth_);
		sr = f2range(fsx, fpn);
		sg = f2range(fsy, fpn);
		sb = f2range(fsz, fpn);
	}
#endif							

	// allocate or match existing colour 
	byte_t chosen_index = allocate_colour(x, scan, sr, sg, sb);
	
	// feed back allocation error if appropriate

	if (errorunit_)
	{		
		feedback_field_error
		(
			x, scan, 
			chosen_index, 
			fsx, 
			fsy, 
			fsz
		);
	}

}

// --------------------------------------------------------------------
// common colour allocator
// --------------------------------------------------------------------

byte_t ReduceUnit_DIRECT_SAMPLING::allocate_colour(int x, int scan, signed short sr, signed short sg, signed short sb)
{
//	byte_t* sync_mapping = &PCT_[x << 4];
	byte_t* sync_mapping = ADDRESS_PCT_SCAN(PCT_, x, display_width_, scan);
	
	word_t* scan_palette = &palette_[scan * palettebuffer_stride_];
				
	// look for match in visible palette(s)
	signed short best_index = 0;
	signed long best_error = 0x7FFFFFFF;
	signed short findex = -1;
	for (signed short p = g_allargs.colours-1; p >= 0; --p)
	{
		word_t mask = 0;
		if (DPT_) 
			mask = DPT_[x + (scan*display_width_)];

		// avoid masked colours for this pixel
		if ((mask & (1 << p)) == 0)
		{
			// index correct palette for this x & p
			byte_t upi = sync_mapping[p];
			signed short spi = (signed short)((signed char)upi);		
			word_t* x_palette = &scan_palette[spi * g_allargs.colours];

			// find colour (0x0RGB)
			word_t pc = x_palette[p];
			
			// check allocated status - top bits signify palette entry has been claimed
			if ((pc & 0x8000) != 0)
			{
				// separate colour channels	
				signed short pr = pc >> (r_pos_ - colourdepth_) & c_mask_;
				signed short pg = pc >> (g_pos_ - colourdepth_) & c_mask_;
				signed short pb = pc >> (b_pos_ - colourdepth_) & c_mask_;
							
				// calculate error 
				signed short rerr = pr - sr;
				signed short gerr = pg - sg;
				signed short berr = pb - sb;
				rerr *= rerr;
				gerr *= gerr;
				berr *= berr;
				signed long errorsum = (signed long)rerr + (signed long)gerr + (signed long)berr;
				
				// track smallest error
				if (errorsum < best_error)
				{
					best_error = errorsum;
					best_index = p;

					// can stop if we have exact match
					if (errorsum == 0)
						break;
				}
			}
			else
			{
				// unallocated - can use if no exact match found
				findex = p;
			}
		}
	}
	
	// decide index to use
	
	byte_t chosen_index = 0;
	if (best_error == 0)
	{
		// if we have exact match, give that priority over free slots
		chosen_index = (byte_t)best_index;
	}
	else if (findex >= 0)
	{
		// otherwise, if we have a free slot, use that			
		chosen_index = (byte_t)findex;
		
		// allocate slot
		word_t palette_entry =  0x8000 | 
								(sr << (r_pos_ - colourdepth_)) | 
								(sg << (g_pos_ - colourdepth_)) | 
								(sb << (b_pos_ - colourdepth_));
	
		// index correct palette for this x & i
		byte_t upi = sync_mapping[chosen_index];
		signed char spi = (signed char)upi;
		word_t* x_palette = &scan_palette[spi * g_allargs.colours];

		// extract colour (0x0RGB)
		x_palette[chosen_index] = palette_entry;
	}
	else
	{
		// fall back to nearest match
		chosen_index = (byte_t)best_index;
	}
	
	return chosen_index;
}

// --------------------------------------------------------------------
// common pixel allocation pass
// --------------------------------------------------------------------

void ReduceUnit_DIRECT_SAMPLING::allocate_pixels()
{	
	int progress = 0;
	int last_progress_pct = -1;
			
	signed short palette_scan_size = palettebuffer_stride_;
	signed short bitmap_scan_size = display_width_;
		
	for (signed short scan = 1; scan < display_height_; ++scan)
	{
		word_t* scan_palette = &palette_[scan * palette_scan_size];
		byte_t* scan_bitmap = &bitmap_[scan * bitmap_scan_size];
		
		byte_t* sync_mapping_scan = ADDRESS_PCT_SCAN(PCT_, 0, display_width_, scan);
//		byte_t* sync_mapping_scan = &PCT_[0];
				
		// alternating x-sweep for FS error propagation
		
		signed short xsweepflag = (scan & 1);
		signed short xsweepdir = (xsweepflag << 1) - 1;
		if (errorunit_)
			errorunit_->set_sweep(xsweepdir);
		
		signed short x = 0;
		signed short xend = bitmap_scan_size;
		if (xsweepflag == 0)
		{
			xend = -1;
			x = bitmap_scan_size - 1;
		}	

		// map one line of image
				
 		while (x != xend)
 		{							
			// find palette mapping for this 'x'			
			byte_t* sync_mapping = &sync_mapping_scan[x << MAX_PALETTE_SHIFT];
	
			// extract pixel
			byte_t ur, ug, ub;
			signed short sr;
			signed short sg;
			signed short sb;
			float fsx;
			float fsy;
			float fsz;

			if (g_allargs.ccmode == CCMode_NONE)
			{
				imgpkg_.get_rgb(active_source_, (int)x, (int)scan, display_height_, ur, ug, ub);
				signed short srn = ((signed short)(unsigned short)ur);
				signed short sgn = ((signed short)(unsigned short)ug);
				signed short sbn = ((signed short)(unsigned short)ub);

				float fladj = 0.0f;
				if (laceunit_)
					fladj = laceunit_->fsample(x, scan);

				float fdxadj = 0.0f;
				float fdyadj = 0.0f;
				float fdzadj = 0.0f;
				if (ditherunit_)
					ditherunit_->fsample(x, scan, fdxadj, fdyadj, fdzadj);

				float fexadj = 0.0f;
				float feyadj = 0.0f;
				float fezadj = 0.0f;
				if (errorunit_)
					errorunit_->sample_spatial(x, scan, fexadj, feyadj, fezadj);

				// into unit space, combining lace/dither/error adjustments
				fsx = col2f(srn) + fladj + fdxadj + fexadj;
				fsy = col2f(sgn) + fladj + fdyadj + feyadj;
				fsz = col2f(sbn) + fladj + fdzadj + fezadj;
										
				// into palette space
				const float fpn = (float)pal_crange(colourdepth_);
				sr = f2range(fsx, fpn);
				sg = f2range(fsy, fpn);
				sb = f2range(fsz, fpn);			
			}
#ifdef CRYPTOCHROME
			else
			{

				if (index_ == 0)
					imgpkg_.get_rgb(cc_field0_, (int)x, (int)scan, display_height_, ur, ug, ub);
				else
					imgpkg_.get_rgb(cc_field1_, (int)x, (int)scan, display_height_, ur, ug, ub);
				
				signed short srn = ((signed short)(unsigned short)ur);
				signed short sgn = ((signed short)(unsigned short)ug);
				signed short sbn = ((signed short)(unsigned short)ub);
			
				fsx = col2f(srn);
				fsy = col2f(sgn);
				fsz = col2f(sbn);
				
				// into palette space for this field
				const float fpn = (float)pal_crange(colourdepth_);
				sr = f2range(fsx, fpn);
				sg = f2range(fsy, fpn);
				sb = f2range(fsz, fpn);
			}
#endif
			byte_t chosen_index = match_colour(x, scan, sr, sg, sb);
			
			if (errorunit_)
			{
				feedback_error
				(
					x, scan, 
					chosen_index, 
					fsx, 
					fsy, 
					fsz
				);			
			}
			
			scan_bitmap[x] = chosen_index;
			
			x += xsweepdir;	
		}
		
		#ifdef __ATARI__
		// update progress
		if (!g_allargs.quiet)
		{
			int progress_pct = ((int)scan * 100) / display_height_;
			if (progress_pct != last_progress_pct)
			{
				last_progress_pct = progress_pct;		
				switch (progress & 3)
				{
				case 0:
					printf("- %2d%%\r", progress_pct); break;
				case 1:
					printf("\\ %2d%%\r", progress_pct); break;
				case 2:
					printf("| %2d%%\r", progress_pct); break;
				case 3:
					printf("/ %2d%%\r", progress_pct); break;
				};
				fflush(stdout);
				progress++;
			}
		}
		#endif
	}
}

// --------------------------------------------------------------------
