// ---------------------------------------------------------------------------------------------------------------------
//	PhotoChrome v6.xx - Atari ST interlaced image reduction framework
//	1991-2016 Douglas Little (d.m.l)
// ---------------------------------------------------------------------------------------------------------------------
//	special thanks to:
//		evil@DHS, RA/PDX, Troed/SYNC for help testing and input on improvements
// ---------------------------------------------------------------------------------------------------------------------

// ====================================================================
// 	image colour reduction unit
// ====================================================================

#include "common.h"
#include "config.h"

#if defined(__ATARI__) || defined(__MINGW32__)
#include "FreeImage.3.15.3.h"
#else
#include "FreeImage.3.15.4.h"
#endif

#include "imagepkg.h"
#include "pcsimage.h"

#include <math.h>

#include <iostream>
#include <map>
#include <assert.h>

extern Arguments g_allargs;

// --------------------------------------------------------------------
// colour reduction/allocation method 5 (self organising network / map)
// --------------------------------------------------------------------
// rules:
// 	1) create link from every pixel to every neuron representing a potential palette entry for that pixel
//	2) train neurons against pixels they can see, over N iterations
//	3) convert each neuron to a single colour
//	4) remap image to nearest matching colour
// notes:
//	* typically 3 palettes per scan, so 320/3 = ~100 pixels visible to each palette (maximum of 172, by measurement)
//	* ~48 palette bins per scan, so 48*100 = ~5000 colour records per scan
// --------------------------------------------------------------------

ReduceUnit_5_SOM::ReduceUnit_5_SOM(PCSImage* _host, int _index)
	: ReduceUnit_INDIRECT_ANALYTICAL(_host, _index)
{			
	// create pixel and colour association buffers
	
	num_neurons_ = palettebuffer_stride_ * display_height_;
	neurons_ = new neuron[num_neurons_];
	if (!neurons_) { printf("error - could not allocate memory for colour neurons!\n"); getchar(); exit(1); }

	num_pixel_records_ = display_width_ * display_height_;
	pixel_records_ = new pixel_record[num_pixel_records_];
	if (!pixel_records_) { printf("error - could not allocate memory for pixel records!\n"); getchar(); exit(1); }	
}	

ReduceUnit_5_SOM::~ReduceUnit_5_SOM()
{
	delete[] pixel_records_;
	delete[] neurons_;
}

// --------------------------------------------------------------------

void ReduceUnit_5_SOM::generate_colours()
{
	generate_colours_init();
	generate_colours_pass1_associate();
	generate_colours_pass2_train();
	generate_colours_pass3_merge();
}

// --------------------------------------------------------------------

void ReduceUnit_5_SOM::generate_colours_init()
{
	if (g_allargs.verbose)
		printf("reducing field %d of %d / method 5 (self organising network / map)...\n", index_, fields_);

	memset(neurons_, 0, sizeof(neuron) * num_neurons_);
	memset(pixel_records_, 0, sizeof(pixel_record) * num_pixel_records_);

	// prepare perception filter mask

	calc_real pfm_power_curve = g_allargs.pfm_gamma;
	calc_real pfm_influence = g_allargs.pfm_influence;

	memset(fmask_, 0, sizeof(float) * num_pixel_records_);

	for (signed long mi = 0; mi < num_pixel_records_; ++mi)
	{
		// look up perception filter mask (range: 1.0 -> 0.0)
		calc_real perception_filter_mask = (calc_real)((unsigned int)(mask_[mi])) / 255.0f;
		
		// apply gamma curve
		perception_filter_mask = xclamp<calc_real>(pow(perception_filter_mask, pfm_power_curve), 0.0, 1.0);
		
		// attenuate influence
		perception_filter_mask = inverse_attenuation<calc_real>(perception_filter_mask, pfm_influence);
		
		fmask_[mi] = (float)perception_filter_mask;
	}	
}

void ReduceUnit_5_SOM::init_exp_solutions(arg_real _radius)
{
	for (int i = 0; i < max_neuron_references; ++i)
	{
		calc_real rho = 1.0f;
		if (i > 0)
			rho = exp(-1.0 * 4.0 * (calc_real(i) / _radius));
		exp_solutions_[i] = rho;
	}
}
	
// --------------------------------------------------------------------

void ReduceUnit_5_SOM::generate_colours_pass1_associate()
{		
	int progress = 0;
	int last_progress_pct = -1;

	signed short track_max_pixrefs = 0;
		
	ImagePkg& imgpkg = ImagePkg::instance();	
		
	// pass #1 - create palette bins with pixel references, and pixel records with bin references
	
	if (g_allargs.verbose)
		printf("generating colour bins...\n");
	
	for (signed short scan = 1; scan < display_height_; ++scan)
	{
		byte_t* sync_mapping_scan = ADDRESS_PCT_SCAN(PCT_, 0, display_width_, scan);
		//byte_t* sync_mapping_scan = &PCT_[0];
	
		// linear left-to-right ordering
		for (int x = 0; x < display_width_; ++x)
		{							
			unsigned long pixel_index = (scan * display_width_) + x;

			pixel_record& record = pixel_records_[pixel_index];	


			// extract pixel
			byte_t ur, ug, ub;
			imgpkg_.get_rgb(active_source_, x, (int)scan, display_height_, ur, ug, ub);
			signed short srn = ((signed short)(unsigned short)ur);
			signed short sgn = ((signed short)(unsigned short)ug);
			signed short sbn = ((signed short)(unsigned short)ub);
		
			// fetch input adjustments (lace, dither)
			
			float fladj = 0.0f;
			if (laceunit_)
				fladj = laceunit_->fsample(x, scan);
		
			float fdxadj = 0.0f;
			float fdyadj = 0.0f;
			float fdzadj = 0.0f;
			if (ditherunit_)
				ditherunit_->fsample(x, scan, fdxadj, fdyadj, fdzadj);
		
			// into unit space, combining lace/dither adjustments - this is the intended colour
			float fsx = col2f(srn) + fladj + fdxadj;
			float fsy = col2f(sgn) + fladj + fdyadj;
			float fsz = col2f(sbn) + fladj + fdzadj;
				
			// intended field colour
			record.ix = fsx;
			record.iy = fsy;
			record.iz = fsz;

			// field error adjustment should offset colour being generated, but only a portion (if any)
			// of this error should be fed back into next field error otherwise it gets amplified
			float fexadj = 0.0f;
			float feyadj = 0.0f;
			float fezadj = 0.0f;
			if (errorunit_)
				errorunit_->sample_field(x, scan, fexadj, feyadj, fezadj);

			//if (--prnt > 0)
			//	printf("fedr: %.4f, fedg: %.4f, fedb: %.4f\n", fexadj, feyadj, fezadj);

			// aim colour (including feedback error offset)
//			record.cx = nvi_quantize(fsx + fexadj, colourdepth_, fields_);
//			record.cy = nvi_quantize(fsy + feyadj, colourdepth_, fields_);
//			record.cz = nvi_quantize(fsz + fezadj, colourdepth_, fields_);
			record.cx = fsx + fexadj;
			record.cy = fsy + feyadj;
			record.cz = fsz + fezadj;
						
			// find palette mapping for this 'x'			
			byte_t* sync_mapping = &sync_mapping_scan[x << MAX_PALETTE_SHIFT];
			
			// insert reference to this pixel into every visible palette index
			
			for (signed short p = g_allargs.colours-1; p >= 0; --p)
			{
				word_t mask = 0;
				if (DPT_) 
					mask = DPT_[x + (scan*display_width_)];

				// avoid masked colours for this pixel
				if ((mask & (1 << p)) == 0)
				{
					// index correct palette for this x & p
					byte_t upi = sync_mapping[p];
					signed short spi = (signed short)((signed char)upi);
	
					// calculate absolute palette bin to fill, from:				
					//  scan = current line
					// 	spi = relative palette offset from this scan
					// 	p = colour index
#if defined (__ATARI__)
					signed short ineuron = (scan * palettebuffer_stride_) + (spi * g_allargs.colours) + p;
#else
					int ineuron = (scan * palettebuffer_stride_) + (spi * g_allargs.colours) + p;
#endif
					neuron& rneuron = neurons_[ineuron];
					// assign bin index/name
					rneuron.index = p;
					
					// init neuron
					rneuron.wx = range2f(p, 15.0);
					rneuron.wy = range2f(p, 15.0);
					rneuron.wz = range2f(p, 15.0);					
					//rneuron.wx = twister_.genFloat();
					//rneuron.wy = twister_.genFloat();
					//rneuron.wz = twister_.genFloat();

					rneuron.vx = 0.0f;
					rneuron.vy = 0.0f;
					rneuron.vz = 0.0f;
					
					if (rneuron.num_pixel_references < max_pixel_references)
					{
						// add pixel reference										
						pixel_reference& rpref = rneuron.pix_references[rneuron.num_pixel_references++];											
						rpref.pixel_x = x;
						rpref.pixel_y = scan;												
						rpref.inref = record.num_neuron_references;

						track_max_pixrefs = xmax(track_max_pixrefs, rneuron.num_pixel_references);
						
						// add bin reference to pixel record, for reduction later
						neuron_reference& rnref = record.neuron_references[record.num_neuron_references++];
						rnref.ineuron = ineuron;
						
						//if (record.num_references > g_allargs.colours)
						//	printf("******************************* error on refcount *****************\n");
					}
					else
					{
						static bool s_bin_warning = false;
						if (!s_bin_warning)
						{
							s_bin_warning = true;
							std::cerr << "error - number of possible pixel references per colour has exceeded fixed limit of (" << max_pixel_references << ")" << std::endl;
							std::cerr << "\tthis will cause some pixel positions to remain unprocessed and will display incorrect colours" << std::endl;
							std::cerr << "\teither uses more colour changes per scan to reduce the pixel->colour refcount" << std::endl;
							std::cerr << "\tor rebuild from source with larger 'max_pixel_references'," << std::endl;
							std::cerr << "\tor use a reduction method which does not use pixel<->colour association (-m 0,1,2,4)" << std::endl;
						}
					}
				}
			}			
		}
		
		#ifdef __ATARI__
		// update progress
		if (!g_allargs.quiet)
		{
			int progress_pct = ((int)scan * 100) / display_height_;
			if (progress_pct != last_progress_pct)
			{
				last_progress_pct = progress_pct;		
				switch (progress & 3)
				{
				case 0:
					printf("- %2d%%\r", progress_pct); break;
				case 1:
					printf("\\ %2d%%\r", progress_pct); break;
				case 2:
					printf("| %2d%%\r", progress_pct); break;
				case 3:
					printf("/ %2d%%\r", progress_pct); break;
				};
				fflush(stdout);
				progress++;
			}
		}
		#endif
	}
	
	if (g_allargs.verbose)
		printf("max pixel refs per bin = %d\n", track_max_pixrefs);
}

inline void ReduceUnit_5_SOM::present(pixel_record& rrec, arg_real alpha, arg_real radius)
{
	// find winning *visible* neuron via references
	
	calc_real win_d = 1e6f;
	int win_ni = 0;
	
	for (signed short ni = 0; ni < rrec.num_neuron_references; ++ni)
	{
		// find neuron for this reference
		const neuron_reference& rnref = rrec.neuron_references[ni];
#if defined (__ATARI__)
		signed short ineuron = rnref.ineuron;
#else
		int ineuron = rnref.ineuron;
#endif
		neuron& rneuron = neurons_[ineuron];
		
		// find biased distance
		calc_real dx = (rrec.cx - rneuron.wx);
		calc_real dy = (rrec.cy - rneuron.wy);
		calc_real dz = (rrec.cz - rneuron.wz);
		calc_real d = (dx * dx) + (dy * dy) + (dz * dz);
		
		if (d < win_d)
		{
			win_d = d;
			win_ni = ni;
		}
	}

	// calculate neighbour influence 
	for (signed short ni = 0; ni < rrec.num_neuron_references; ++ni)
	{
		// find neuron for this reference
		neuron_reference& rnref = rrec.neuron_references[ni];
#if defined (__ATARI__)
		signed short ineuron = rnref.ineuron;
#else
		int ineuron = rnref.ineuron;
#endif
		neuron& rneuron = neurons_[ineuron];

		int i = xabs(ni - win_ni);
		assert(i < max_neuron_references);
		rnref.rho = exp_solutions_[i];
	}	
}

#include <list>


mersenne_twister g_twister;
inline unsigned int grand() { return g_twister.genU32(); }
ptrdiff_t myrandom (ptrdiff_t i) { return grand() % i;}
ptrdiff_t (*p_myrandom)(ptrdiff_t) = myrandom;

void ReduceUnit_5_SOM::generate_colours_pass2_train()
{
	int xxx = 0;
		
	int sampling_breadth = xftoi_floor(float(num_pixel_records_) * g_allargs.nn_breadth);

	std::vector<signed long> sequence;		
	for (int i = 0; i < num_pixel_records_; ++i)
		sequence.push_back(i);
	g_twister.reset(exposure_twister_.genU32());
				
	float learning_rate = g_allargs.nn_rate;
	
	int total_cycles = 1 << g_allargs.nn_depth;
	for (int cycle = 0; cycle < total_cycles; ++cycle)
	{		
		std::random_shuffle(sequence.begin(), sequence.end(), p_myrandom);
			
		// calculate alpha
		calc_real alpha = learning_rate * (0.5 * exp(-1.0 * 1.5 * (calc_real(cycle) / calc_real(total_cycles))));
		calc_real radius = 10.0 * exp(-1.0 * 5.0 * (calc_real(cycle) / calc_real(total_cycles)));
		init_exp_solutions(radius);
		
		if (--xxx <= 0)
		{
			xxx = 100;
			printf("\riteration:%5d/%5d, radius(%.3f), alpha(%.6f), ", 
				cycle, total_cycles, 
				(double)radius, 
				alpha);
				
			generate_colours_pass3_merge();
			calc_real ee = measure_pixels();
			printf("sqerr:%.2f     ", (double)ee);
			fflush(stdout);			
		}
	
		// present each pixel to each relevant neuron
		signed long* psequence = &sequence[0];
		for (int bi = 0; bi < sampling_breadth; ++bi)
		{
			//signed long irec = *(sit++);
			signed long irec = psequence[bi];
			pixel_record& rrec = pixel_records_[irec];
			present(rrec, alpha, radius);	
		}
		
		// update neurons via references
		for (int irec = 0; irec < num_pixel_records_; ++irec)
		{
			pixel_record& rrec = pixel_records_[irec];			
					
			for (signed short ni = 0; ni < rrec.num_neuron_references; ++ni)
			{
				// find neuron for this reference
				neuron_reference& rnref = rrec.neuron_references[ni];
#if defined (__ATARI__)
				signed short ineuron = rnref.ineuron;
#else
				int ineuron = rnref.ineuron;
#endif
				neuron& rneuron = neurons_[ineuron];
								
				// train neuron
				calc_real t = rnref.rho * alpha;

				rneuron.vx += (rrec.cx - rneuron.wx) * t;
				rneuron.vy += (rrec.cy - rneuron.wy) * t;
				rneuron.vz += (rrec.cz - rneuron.wz) * t;

				rneuron.wx += rneuron.vx;
				rneuron.wy += rneuron.vy;
				rneuron.wz += rneuron.vz;
				
				rneuron.vx *= 0.95f;
				rneuron.vy *= 0.95f;
				rneuron.vz *= 0.95f;
			}	
		}
	}
	printf("\n");	
}

void ReduceUnit_5_SOM::generate_colours_pass3_merge()
{
	for (signed long ineuron = 0; ineuron < num_neurons_; ++ineuron)
	{
		neuron& rneuron = neurons_[ineuron];
		
		if (rneuron.num_pixel_references > 0)
		{			
			calc_real dcxm = rneuron.wx;
			calc_real dcym = rneuron.wy;
			calc_real dczm = rneuron.wz;

			// into palette space for this field
			const calc_real fpn = (calc_real)pal_crange(colourdepth_);
			signed short sr = f2range(dcxm, fpn);
			signed short sg = f2range(dcym, fpn);
			signed short sb = f2range(dczm, fpn);			
		
			word_t palette_entry =  0x8000 | 
									(sr << (r_pos_ - colourdepth_)) | 
									(sg << (g_pos_ - colourdepth_)) | 
									(sb << (b_pos_ - colourdepth_));

			palette_[ineuron] = palette_entry;					
		}
	}
}

// --------------------------------------------------------------------

void ReduceUnit_5_SOM::allocate_pixels()
{
	allocate_pixels_remap_();
}

// --------------------------------------------------------------------


void ReduceUnit_5_SOM::allocate_pixels_remap_()
{	
	int progress = 0;
	int last_progress_pct = -1;
			
	signed short palette_scan_size = palettebuffer_stride_;
	signed short bitmap_scan_size = display_width_;
	
	for (int scan = 1; scan < display_height_; ++scan)
	{
		word_t* scan_palette = &palette_[scan * palette_scan_size];
		byte_t* scan_bitmap = &bitmap_[scan * bitmap_scan_size];

		byte_t* sync_mapping_scan = ADDRESS_PCT_SCAN(PCT_, 0, display_width_, scan);
//		byte_t* sync_mapping_scan = &PCT_[0];
				
		// alternating x-sweep for FS error propagation
		
		signed short xsweepflag = (scan & 1);
		signed short xsweepdir = (xsweepflag << 1) - 1;
		if (errorunit_)
			errorunit_->set_sweep(xsweepdir);
		
		signed short x = 0;
		signed short xend = bitmap_scan_size;
		if (xsweepflag == 0)
		{
			xend = -1;
			x = bitmap_scan_size - 1;
		}	

		// map one line of image
				
 		while (x != xend)
 		{							
			// find palette mapping for this 'x'			
			byte_t* sync_mapping = &sync_mapping_scan[x << MAX_PALETTE_SHIFT];
	
			pixel_record& rrec = pixel_records_[x + (scan * display_width_)];			
		
			byte_t chosen_index = 0;

			if (rrec.num_neuron_references > 0)
			{	
				float fexadj = 0.0f;
				float feyadj = 0.0f;
				float fezadj = 0.0f;
				if (errorunit_)
					errorunit_->sample_spatial(x, scan, fexadj, feyadj, fezadj);

				// into unit space, combining spatial error adjustment
				float fsx = rrec.ix + fexadj;
				float fsy = rrec.iy + feyadj;
				float fsz = rrec.iz + fezadj;
				//calc_real fsx = rrec.ix;
				//calc_real fsy = rrec.iy;
				//calc_real fsz = rrec.iz;
										
				// into palette space
				const calc_real fpn = (calc_real)pal_crange(colourdepth_);
				signed short sr = f2range(fsx, fpn);
				signed short sg = f2range(fsy, fpn);
				signed short sb = f2range(fsz, fpn);			

				chosen_index = match_colour(x, scan, sr, sg, sb);
				
				if (errorunit_)
				{
					feedback_error
					(
						x, scan, 
						chosen_index, 
						fsx, 
						fsy, 
						fsz
					);
				}
			}
			
			scan_bitmap[x] = chosen_index;
			
			x += xsweepdir;	
		}
		
		#ifdef __ATARI__
		// update progress
		if (!g_allargs.quiet)
		{
			int progress_pct = ((int)scan * 100) / display_height_;
			if (progress_pct != last_progress_pct)
			{
				last_progress_pct = progress_pct;		
				switch (progress & 3)
				{
				case 0:
					printf("- %2d%%\r", progress_pct); break;
				case 1:
					printf("\\ %2d%%\r", progress_pct); break;
				case 2:
					printf("| %2d%%\r", progress_pct); break;
				case 3:
					printf("/ %2d%%\r", progress_pct); break;
				};
				fflush(stdout);
				progress++;
			}
		}
		#endif
		
	}
}


calc_real ReduceUnit_5_SOM::measure_pixels()
{	
	int progress = 0;
	int last_progress_pct = -1;
			
	signed short palette_scan_size = palettebuffer_stride_;
	signed short bitmap_scan_size = display_width_;

	calc_real ee = 0.0f;
	
	for (int scan = 1; scan < display_height_; ++scan)
	{
		word_t* scan_palette = &palette_[scan * palette_scan_size];
		byte_t* scan_bitmap = &bitmap_[scan * bitmap_scan_size];

		byte_t* sync_mapping_scan = ADDRESS_PCT_SCAN(PCT_, 0, display_width_, scan);
//		byte_t* sync_mapping_scan = &PCT_[0];
								
		// map one line of image				
 		for (int x = 0; x < bitmap_scan_size; ++x)
 		{							
			// find palette mapping for this 'x'			
			byte_t* sync_mapping = &sync_mapping_scan[x << MAX_PALETTE_SHIFT];
	
			pixel_record& rrec = pixel_records_[x + (scan * display_width_)];			
		
			if (rrec.num_neuron_references <= 0)
				continue;
	
			// into unit space, combining spatial error adjustment
			calc_real fsx = rrec.ix;
			calc_real fsy = rrec.iy;
			calc_real fsz = rrec.iz;
									
			// into palette space
			const calc_real fpn = (calc_real)pal_crange(colourdepth_);
			signed short sr = f2range(fsx, fpn);
			signed short sg = f2range(fsy, fpn);
			signed short sb = f2range(fsz, fpn);			

			byte_t chosen_index = match_colour(x, scan, sr, sg, sb);
	
			{
				// index correct palette for this x & p
				byte_t* sync_mapping = &sync_mapping_scan[x << MAX_PALETTE_SHIFT];
				byte_t upi = sync_mapping[chosen_index];
				signed short spi = (signed short)((signed char)upi);		
				word_t* scan_palette = &palette_[scan * palettebuffer_stride_];
				word_t* x_palette = &scan_palette[spi * g_allargs.colours];
			
				// find colour (0x0RGB)
				word_t pc = x_palette[chosen_index];
			
				// palette colour
				signed short pr = ((pc >> (r_pos_ - colourdepth_)) & c_mask_);
				signed short pg = ((pc >> (g_pos_ - colourdepth_)) & c_mask_);
				signed short pb = ((pc >> (b_pos_ - colourdepth_)) & c_mask_);
			
				// into unit space
				const calc_real fpn = (calc_real)pal_crange(colourdepth_);
				calc_real fpr = range2f(pr, fpn);
				calc_real fpg = range2f(pg, fpn);
				calc_real fpb = range2f(pb, fpn);	
			
				// quantize in native virtual space
				//fsx = range2f(f2range(fsx, (float)nvi_crange(colourdepth_, fields_)), (float)nvi_crange(colourdepth_, fields_));
				//fsy = range2f(f2range(fsy, (float)nvi_crange(colourdepth_, fields_)), (float)nvi_crange(colourdepth_, fields_));
				//fsz = range2f(f2range(fsz, (float)nvi_crange(colourdepth_, fields_)), (float)nvi_crange(colourdepth_, fields_));
			
				// calculate error
				calc_real perception_filter_mask = fmask_[x + (scan * display_width_)];			
				calc_real fedr = (fsx - fpr) * perception_filter_mask;
				calc_real fedg = (fsy - fpg) * perception_filter_mask;
				calc_real fedb = (fsz - fpb) * perception_filter_mask;
														
				ee += (fedr * fedr) + (fedg * fedg) + (fedb * fedb);				
			}		
		}		
	}

	return ee;
}

// --------------------------------------------------------------------
